--  _    _          ___           __ _     (R)
-- | |  (_)_ _____ / __|___ _ _  / _(_)__ _
-- | |__| \ V / -_) (__/ _ \ ' \|  _| / _` |
-- |____|_|\_/\___|\___\___/_||_|_| |_\__, |
--                                    |___/
-- Copyright (c) 2009-2019 LiveConfig GmbH.
-- ---------------------------------------------------------------------------
-- common/lua/backup.lua
-- Backup service management
-- $Id$
-- ---------------------------------------------------------------------------

-- Imported functions
local LC = LC
local io = io
local os = os
local type = type
local ipairs = ipairs

-- Module declaration
module("backup")

-- Exported functions
--   install()
--   uninstall()

-- ---------------------------------------------------------------------------
-- install()
--
-- Install & enable the backup service (lcbackup)
-- ---------------------------------------------------------------------------
function install()

  LC.log.print(LC.log.INFO, "Installing backup service (lcbackup)")

  -- acquire mutex
  LC.mutex.lock("backup.install")

  local ret, stdout, stderr

  -- install service (systemd)
  if LC.fs.is_dir("/lib/systemd/system") and not LC.fs.is_file("/lib/systemd/system/lcbackup.service") then
    LC.exec("cp --preserve=all " .. LC.liveconfig.libexecdir .. "/lcbackup.service /lib/systemd/system/")
    if LC.fs.is_file("/bin/systemctl") then
      ret, stdout, stderr = LC.exec("/bin/systemctl enable lcbackup")
      if ret ~= 0 then
        LC.mutex.unlock("backup.install")
        return false, stderr
      end
    end
  end

  -- lcbackup: install init script
  if not LC.fs.is_file("/etc/init.d/lcbackup") then
    LC.exec("cp " .. LC.liveconfig.libexecdir .. "/lcbackup.init /etc/init.d/lcbackup")
    LC.fs.setperm("/etc/init.d/lcbackup", "0700", "root", "root")

    if LC.fs.is_file("/usr/sbin/update-rc.d") then
      ret, stdout, stderr = LC.exec("/usr/sbin/update-rc.d lcbackup defaults")
    elseif LC.fs.is_file("/sbin/chkconfig") then
      ret, stdout, stderr = LC.exec("/sbin/chkconfig --add lcbackup")
    elseif LC.fs.is_file("/sbin/rc-update") then
      ret, stdout, stderr = LC.exec("/sbin/rc-update add lcbackup default")
    else
      ret = 1
      stderr = "Can't install init script 'lcbackup' - no supported tool found"
    end

    if ret ~= 0 then
      LC.mutex.unlock("backup.install")
      return false, stderr
    end

  end

  -- start lcbackup
  if LC.fs.is_file("/sbin/service") then
    ret, stdout, stderr = LC.exec("/sbin/service lcbackup start")
  else
    ret, stdout, stderr = LC.exec("/etc/init.d/lcbackup start")
  end

  if ret ~= 0 then
    LC.mutex.unlock("backup.install")
    return false, stderr
  end

  -- release mutex
  LC.mutex.unlock("backup.install")

  return true

end

-- ---------------------------------------------------------------------------
-- uninstall()
--
-- Uninstall backup service (lcbackup)
-- ---------------------------------------------------------------------------
function uninstall()

  LC.log.print(LC.log.INFO, "Uninstalling backup service (lcbackup)")

  -- acquire mutex
  LC.mutex.lock("backup.install")

  -- quit lcbackup
  if LC.fs.is_file("/sbin/service") then
    LC.exec("/sbin/service lcbackup stop")
  else
    LC.exec("/etc/init.d/lcbackup stop")
  end

  -- disable lcbackup
  if LC.fs.is_file("/lib/systemd/system/lcbackup") then
    -- disable service (systemd)
    if LC.fs.is_file("/bin/systemctl") then
      LC.exec("/bin/systemctl disable lcbackup")
    end
    -- remove service file
    os.remove("/lib/systemd/system/lcbackup")
  end

  if LC.fs.is_file("/etc/init.d/lcbackup") then
    if LC.fs.is_file("/usr/sbin/update-rc.d") then
      LC.exec("/usr/sbin/update-rc.d -f lcbackup remove")
    elseif LC.fs.is_file("/sbin/chkconfig") then
      LC.exec("/sbin/chkconfig --del lcbackup")
    elseif LC.fs.is_file("/sbin/rc-update") then
      LC.exec("/sbin/rc-update del lcbackup default")
    else
      LC.log.print(LC.log.ERR, "Can't uninstall init script 'lcbackup' - no supported tool found")
    end
    os.remove("/etc/init.d/lcbackup")
  end

  -- release mutex
  LC.mutex.unlock("backup.install")

  return true

end

-- <EOF>----------------------------------------------------------------------
