--  _    _          ___           __ _     (R)
-- | |  (_)_ _____ / __|___ _ _  / _(_)__ _
-- | |__| \ V / -_) (__/ _ \ ' \|  _| / _` |
-- |____|_|\_/\___|\___\___/_||_|_| |_\__, |
--                                    |___/
-- Copyright (c) 2009-2024 LiveConfig GmbH.
-- ---------------------------------------------------------------------------
-- common/lua/dovecot.lua
-- Lua module to manage dovecot POP3/IMAP server
-- ---------------------------------------------------------------------------
-- This is the "driver" module for dovecot pop/imap server.
-- It must be loaded by the command
--   LC.popimap.load("dovecot")
-- Usually, this should happen at liveconfig.lua (or, if you have a customized
-- module, at custom.lua)
-- ---------------------------------------------------------------------------

-- Imported functions
local LC = LC
local LCS = LCS -- import liveconfig global storage
local type = type
local io = io
local os = os
local string = string
local pairs = pairs
local ipairs = ipairs
local tonumber = tonumber

-- Module declaration
module("dovecot")

-- Exported functions
--   detect()
--   install()
--   uninstall()
--   updateSSL()
--   configure()
--   addMailbox()
--   editMailbox()
--   deleteMailbox()
--   editlock()

-- Optionally disable updating dovecot.conf by LiveConfig
if NOUPDATE == nil then
  NOUPDATE = false
end

-- optionally disallow authentication via CRAM-MD5
-- (This is *ONLY* required if you have imported hashed password with
-- MD5-CRYPT scheme! Enable this option using your "custom.lua" only if you
-- really know what you're doing!
if DISABLE_CRAM == nil then
  DISABLE_CRAM = false
end

-- pattern for per-mailbox home directory in /etc/dovecot/passwd:
if HOME == nil then
  HOME = "/var/mail/%C/%I"
end

-- pattern for "userdb_mail" setting in /etc/dovecot/passwd
if USERDB_MAIL == nil then
  USERDB_MAIL = "userdb_mail=maildir:/var/mail/%C/%I/"
end

-- pattern for "userdb_sieve" setting in /etc/dovecot/passwd
if USERDB_SIEVE == nil then
  USERDB_SIEVE = "userdb_sieve=/var/mail/%C/%I/dovecot.sieve"
end

-- SSL ciphers
if SSL_CIPHERS == nil then
  SSL_CIPHERS = LC.liveconfig.DEFAULT_SSL_CIPHERS
end
if SSL_PCI_CIPHERS ~= nil then
  LC.log.print(LC.log.WARNING, "dovecot.SSL_PCI_CIPHERS is deprecated - use dovecot.SSL_STRONG_CIPHERS instead")
  if SSL_STRONG_CIPHERS == nil then
    SSL_STRONG_CIPHERS = SSL_PCI_CIPHERS
  end
end
if SSL_STRONG_CIPHERS == nil then
  SSL_STRONG_CIPHERS = LC.liveconfig.DEFAULT_SSL_STRONG_CIPHERS
end

-- ---------------------------------------------------------------------------
-- getBinaryVersion()
--
-- Return the binary version number from dovecot
-- ---------------------------------------------------------------------------
local function getBinaryVersion(bin)
  local handle = io.popen(bin .. " --version", "r")
  local ret = handle:read("*a")
  handle:close()
  local v = string.match(ret, "(%d+%.%d+[%.%d+]*)")
  return v
end

-- ---------------------------------------------------------------------------
-- checkDenyFile()
--
-- Check if "deny.*" files are existing - if not, create them accordingly
-- ---------------------------------------------------------------------------
local function checkDenyFile(denyfile)
  local fh, msg
  if not LC.fs.is_file(denyfile .. ".imap") then
    fh, msg = io.open(denyfile .. ".imap", "a")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", denyfile, ".imap' for appending: ", msg)
      return false, "Can't open '" .. denyfile .. ".imap' for appending: " .. msg
    end
    -- adjust owner & permissions - only Dovecot needs access to this file, so set mod 600 !!!
    LC.fs.setperm(denyfile .. ".imap", "0600", "dovecot", "root")
    fh:close()
  end
  if not LC.fs.is_file(denyfile .. ".pop3") then
    -- create symlink from deny.pop3 to deny.imap:
    os.execute("ln -s " .. denyfile .. ".imap " .. denyfile .. ".pop3")
  end
  if not LC.fs.is_file(denyfile .. ".sieve") then
    -- create symlink from deny.sieve to deny.imap:
    os.execute("ln -s " .. denyfile .. ".imap " .. denyfile .. ".sieve")
  end

  if not LC.fs.is_file(denyfile .. ".smtp") then
    fh, msg = io.open(denyfile .. ".smtp", "a")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", denyfile, ".smtp' for appending: ", msg)
      return false, "Can't open '" .. denyfile .. ".smtp' for appending: " .. msg
    end
    -- adjust owner & permissions - only Dovecot needs access to this file, so set mod 600 !!!
    LC.fs.setperm(denyfile .. ".smtp", "0600", "dovecot", "root")
    fh:close()
  end
  if not LC.fs.is_file(denyfile .. ".lda") then
    -- create symlink from deny.smtp to deny.lda:
    os.execute("ln -s " .. denyfile .. ".smtp " .. denyfile .. ".lda")
  end
  return true
end

-- ---------------------------------------------------------------------------
-- detect()
--
-- Detect if an dovecote pop/imap server is installed. If yes, return a table containing
-- all relevant informations:
--   type          => "dovecot"
--   binname       => file name of the dovecot binary (eg. "/usr/sbin/dovecot")
--   binversion    => binary version (eg. "2.5.5")
--   pkgname       => name of the package (see below)
--   pkgversion    => package version number (eg. "2.2.3-4+etch11")
--   configpath    => Main configuration path for dovecot
--   configfile    => LiveConfigs' configuration file (snippet)
--   statusfile    => LiveConfigs' status file (containing config version etc.)
--   defaultlog    => Default log file (for default and unknown/unconfigured vhosts)
--   start_cmd     => command to start dovecot
--   stop_cmd      => command to stop dovecot
--   reload_cmd    => command to re-load configuration
--   restart_cmd   => command to re-start dovecot process
--
-- If pkgversion is 'nil', then dovecot Server was most propably not installed
-- as native distribution package, but compiled/installed manually.
-- ---------------------------------------------------------------------------
function detect()

  local pkg, v, bin

  -- first check for distribution-specific packages
  if LC.distribution.family == "Debian" or LC.distribution.family == "Gentoo" or LC.distribution.family == "RedHat" or LC.distribution.family == "BSD" or LC.distribution.family == "SUSE" then
    -- Debian/Ubuntu Paketsearch
    if LC.distribution.family == "Debian" then
    -- Debian (all releases)
      pkg, v = LC.distribution.hasPackage('dovecot-common', 'dovecot-pop3d', 'dovecot-imapd')
    elseif LC.distribution.family == "SUSE" then
        -- OpenSUSE: check for both Dovecot 1.2 or 2.0:
        pkg, v = LC.distribution.hasPackage('dovecot21', 'dovecot20', 'dovecot12', 'dovecot')
    else
        -- Gentoo, CentOS
        pkg, v = LC.distribution.hasPackage('dovecot')
    end
    if pkg ~= nil then
      LC.log.print(LC.log.DEBUG, "Found package '", pkg, "' (Version ", v, ")")
      if LC.distribution.family == "BSD" then
        bin = "/usr/local/sbin/dovecot"
      else
        bin = "/usr/sbin/dovecot"
      end
      -- get binary version
      local bv = getBinaryVersion(bin)
      if bv ~= nil then
        -- ok, we have all informations. return data table:
        if LC.distribution.family == "RedHat" then
          local data = {
              ["type"]          = "dovecot",
              ["binname"]       = bin,
              ["binversion"]    = bv,
              ["pkgname"]       = pkg,
              ["pkgversion"]    = v,
              ["configpath"]    = "/etc/dovecot",
              ["configfile"]    = "/etc/dovecot.conf",
              ["userfile"]      = "/etc/dovecot/passwd",
              ["statusfile"]    = "/etc/dovecot/liveconfig.status",
              ["crtfile"]       = "/etc/pki/tls/certs/dovecot.crt",
              ["keyfile"]       = "/etc/pki/tls/private/dovecot.key",
              ["defaultlog"]    = "/var/log/maillog",
              ["start_cmd"]     = "/sbin/service dovecot start",
              ["stop_cmd"]      = "/sbin/service dovecot stop",
              ["reload_cmd"]    = "/sbin/service dovecot reload",
              ["restart_cmd"]   = "/sbin/service dovecot restart",
              ["disable_cram"]  = DISABLE_CRAM,
              ["units"]         = { ["dovecot"] = "dovecot" },
            }
          if string.match(bv, "^(%d+)") == "2" then
            -- Dovecot 2.x configuration
            data.configfile = "/etc/dovecot/dovecot.conf"
          end
          return data
        else
          local data = {
              ["type"]          = "dovecot",
              ["binname"]       = bin,
              ["binversion"]    = bv,
              ["pkgname"]       = pkg,
              ["pkgversion"]    = v,
              ["configpath"]    = "/etc/dovecot",
              ["configfile"]    = "/etc/dovecot/dovecot.conf",
              ["userfile"]      = "/etc/dovecot/passwd",
              ["statusfile"]    = "/etc/dovecot/liveconfig.status",
              ["crtfile"]       = "/etc/dovecot/ssl-cert.pem",
              ["keyfile"]       = "/etc/dovecot/ssl-key.pem",
              ["defaultlog"]    = "/var/log/mail.log",
              ["disable_cram"]  = DISABLE_CRAM,
              ["units"]         = { ["dovecot"] = "dovecot" },
            }
          if LC.fs.is_file("/etc/init.d/dovecot") then
            data["start_cmd"]   = "/etc/init.d/dovecot start"
            data["stop_cmd"]    = "/etc/init.d/dovecot stop"
            data["reload_cmd"]  = "/etc/init.d/dovecot reload"
            data["restart_cmd"] = "/etc/init.d/dovecot restart"
          else
            data["start_cmd"]   = "/usr/sbin/service dovecot start"
            data["stop_cmd"]    = "/usr/sbin/service dovecot stop"
            data["reload_cmd"]  = "/usr/sbin/service dovecot reload"
            data["restart_cmd"] = "/usr/sbin/service dovecot restart"
          end
          if LC.distribution.family == "Debian" then
            data["crtfile"] = "/etc/ssl/certs/dovecot.crt"
            data["keyfile"] = "/etc/ssl/private/dovecot.key"
          end
          return data
        end
      end
      -- else: fall trough, to check for custom dovecot installation
      LC.log.print(LC.log.DEBUG, "LC.dovecot.detect(): Found Dovecot package '", pkg, "', but no binary at ", bin)
    end
  end
end

-- ---------------------------------------------------------------------------
-- install(cfg, opts)
--
-- Install LiveConfig configuration files for Dovecot server
-- ---------------------------------------------------------------------------
function install(cfg, opts)
  LC.log.print(LC.log.INFO, "Activating management of Dovecot server")

  local fh, status, msg
  local configpath  = cfg["configpath"]
  local configfile  = cfg["configfile"]
  local statusfile  = cfg["statusfile"]
  local userfile    = cfg["userfile"]

  if opts and opts.prefix then
    -- use prefix (for testing etc.)
    configpath = opts.prefix .. configpath
    configfile = opts.prefix .. configfile
    statusfile = opts.prefix .. statusfile
    userfile   = opts.prefix .. userfile
  end

  -- for CentOS5 does not exists configpath...create
  if not LC.fs.is_dir(configpath) then
    LC.fs.mkdir(configpath)
  end
  -- back up old default configfile (if existing)
  if LC.fs.is_file(configfile) and not LC.fs.is_file( configfile .. ".lcbak") then
    LC.fs.rename( configfile,  configfile .. ".lcbak")
  end

  -- back up old default userfile file (if existing)
  if LC.fs.is_file(userfile) and not LC.fs.is_file( userfile .. ".lcbak") then
    LC.fs.rename( userfile,  userfile .. ".lcbak")
  end

  -- check whether mail user and group exists or not
  -- if not create user and group
  if not LC.sys.user_exists("mail") then
    if not LC.sys.group_exists("mail") then
      LC.users.addGroup("mail")
    end
    LC.users.addUser("mail", "mail", "/var/mail", "nologin")
  end

  -- create (empty) passwd file (if not existing)
  if not LC.fs.is_file(userfile) then
    local fh, msg = io.open(userfile, "a")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", userfile, "' for writing: ", msg)
      return false, "Can't open '" .. userfile .. "' for writing: " .. msg
    end
    -- adjust owner & permissions - only Dovecot needs access to this file, so set mod 600 !!!
    LC.fs.setperm(userfile, "0600", "dovecot", "root")
    fh:close()
  end

  -- write default config file dovecot.conf:
  configure(cfg, opts)

  -- create userfile (/etc/dovecot/passwd)
  os.execute("touch " .. userfile)
  -- adjust permissions
  LC.fs.setperm(userfile, "0600", "dovecot", "root")

  return true
end

-- ---------------------------------------------------------------------------
-- uninstall(cfg, opts)
--
-- Remove LiveConfig configuration files from Dovecot server
-- ---------------------------------------------------------------------------
function uninstall(cfg, opts)
  LC.log.print(LC.log.INFO, "Disabling management of Dovecot server")

  local status, msg
  local configpath = cfg["configpath"]
  local configfile = cfg["configfile"]
  local statusfile = cfg["statusfile"]
  local userfile   = cfg["userfile"]

  if opts and opts.prefix then
    -- use prefix (for testing etc.)
    configpath = opts.prefix .. configpath
    configfile = opts.prefix .. configfile
    statusfile = opts.prefix .. statusfile
    userfile   = opts.prefix .. userfile
  end

  -- restore original "dovecot.conf" (if existing...)
  if LC.fs.is_file( configfile .. ".lcbak") then
    LC.fs.rename( configfile .. ".lcbak", configfile)
  end
  -- restore original "/etc/dovecot/passwd" (if existing...)
  if LC.fs.is_file( userfile .. ".lcbak") then
    LC.fs.rename( userfile .. ".lcbak", userfile)
  end
  
  -- Currently do not delete (because dependencies with postfix)
  -- Delete user and group vmail
  -- local data  = { }
  -- data.user   = "vmail"
  -- data.group  = "vmail"
  -- data.home   = "vmail"
  -- data.shell  = "nologin"
  -- data.passwd = ""
  -- local uid = LC.sys.user_exists("vmail")
  -- if uid ~= false then 
  --   LC.users.delUser(data)
  -- end
  -- local gid = LC.sys.group_exists("vmail")
  -- if gid ~= false then
  --   LC.users.delGroup(data)
  -- end

  -- remove status file
  if LC.fs.is_file(statusfile) then
    os.remove(statusfile)
  end

  -- restart dovecot
  LC.exec(cfg["restart_cmd"])

  return true
end

-- ---------------------------------------------------------------------------
-- updateSSL(cfg, opts)
--
-- Update SSL certificate
-- ---------------------------------------------------------------------------
function updateSSL(cfg, opts)

  -- create/update SSL certificate file:
  if opts.ssl then
    local sfh, msg
    sfh, msg = io.open(cfg.crtfile, "w")
    LC.fs.setperm(cfg.crtfile, "0600", "root", "root")
    if sfh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", cfg.crtfile, "' for writing: ", msg)
      return false, "Can't open '" .. cfg.crtfile .. "' for writing: " .. msg
    end
    sfh:write(opts.ssl_crt)
    if opts.ssl_ca ~= nil then
      -- append chained certificate:
      sfh:write(opts.ssl_ca)
    end
    sfh:close()

    -- create/update SSL key file:
    sfh, msg = io.open(cfg.keyfile, "w")
    LC.fs.setperm(cfg.keyfile, "0600", "root", "root")
    if sfh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", cfg.keyfile, "' for writing: ", msg)
      return false, "Can't open '" .. cfg.keyfile .. "' for writing: " .. msg
    end
    sfh:write(opts.ssl_key)
    sfh:close()
  end

  return true

end

-- ---------------------------------------------------------------------------
-- configure(cfg, opts)
--
-- Configure dovecot Server
-- ---------------------------------------------------------------------------
function configure(cfg, opts)

  local fh, status, msg
  local statusfile = cfg["statusfile"]
  local configpath  = cfg["configpath"]
  local configfile  = cfg["configfile"]
  local v_major, v_minor = string.match(cfg.binversion, "^(%d+)%.(%d+)")
  v_major = tonumber(v_major)
  v_minor = tonumber(v_minor)

  if opts and opts.prefix then
    -- use prefix (for testing etc.)
    statusfile  = opts.prefix .. statusfile
    configfile  = opts.prefix .. configfile
    configpath  = opts.prefix .. configpath
  end

  -- create/update SSL certificate file:
  res, msg = updateSSL(cfg, opts)
  if not res then return false, msg end

  if NOUPDATE then
    LC.log.print(LC.log.INFO, "Won't update Dovecot configuration (dovecot.NOUPDATE=true)")
  else
    -- create configuration file
    fh, msg = io.open(configfile .. ".tmp", "w")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", configfile, ".tmp' for writing: ", msg)
      return false, "Can't open '" .. configfile.. ".tmp' for writing: " .. msg
    end
    -- adjust owner & permissions - both Postfix and Dovecot need access to this file, so set mod 644 !!!
    LC.fs.setperm(configfile .. ".tmp", "0644", "dovecot", "root")

    -- get uid and gid for user "mail"
    local uid = LC.sys.user_exists("mail")
    if uid == false then
      return false, "System user 'mail' does not exist, please create user"
    end
    local gid = LC.sys.group_exists("mail")
    if gid == false then
      return false, "System group 'mail' does not exist, please create group"
    end

    -- create /etc/dovecot.conf  
    LC.liveconfig.writeHeader(fh)

    fh:write("# To include custom configuration options, write them into\n")
    fh:write("# /etc/dovecot/dovecot.local.conf - then save the Dovecot settings\n")
    fh:write("# again in LiveConfig (to update the dovecot.conf and include your file)\n\n")

    -- check version
    if v_major == 1 then
      -- Dovcot version 1.x

      LC.log.print(LC.log.DEBUG, "Dovecot version 1.x")

      if opts.ssl then
        if opts.sslports then
          fh:write("protocols = imap imaps pop3 pop3s\n")
        else
          fh:write("protocols = imap pop3\n")
        end
        if v_minor < 2 then
          fh:write("ssl_disable = no\n")
        else
          if opts.sslonly then
            fh:write("ssl = required\n")
          else
            fh:write("ssl = yes\n")
          end
        end
        if opts.sslstrong then
          -- modern/strong ciphers (actually, defending BEAST is not necessary on POP3/IMAP, but most dumb PCI scans don't care :(
          fh:write("ssl_cipher_list = ", SSL_STRONG_CIPHERS, "\n")
        else
          fh:write("ssl_cipher_list = ", SSL_CIPHERS, "\n")
        end
        fh:write("ssl_cert_file = ", cfg.crtfile, "\n")
        fh:write("ssl_key_file = ", cfg.keyfile, "\n")
      else
        fh:write("protocols = imap pop3\n")
      end

      fh:write([[

login_greeting = server ready
disable_plaintext_auth = no
#log_path = /var/log/dovecot
log_timestamp = "%Y-%m-%d %H:%M:%S "
mail_privileged_group = mail
first_valid_uid = ]], uid, "\n", [[
protocol imap {
  mail_plugins = quota imap_quota
]])

      if opts.imapconn and v_minor > 0 then
        -- only available since v1.1 - bad for CentOS 5.x shipping with Dovecot 1.0.7 :(
        fh:write("  mail_max_userip_connections = ", opts.imapconn, "\n")
      end

      fh:write([[
}
protocol pop3 {
  pop3_uidl_format = %08Xu%08Xv
  mail_plugins = quota
]])

      if opts.pop3conn and v_minor > 0 then
        -- only available since v1.1 - bad for CentOS 5.x shipping with Dovecot 1.0.7 :(
        fh:write("  mail_max_userip_connections = ", opts.pop3conn, "\n")
      end

      fh:write([[
}
protocol lda {
  postmaster_address = postmaster@localhost
  hostname = localhost
  mail_plugins = quota sieve
  mail_plugin_dir = /usr/lib/dovecot/modules/lda
  auth_socket_path = /var/run/dovecot/auth-master
}
auth default {
]])

      -- optionally disable CRAM-MD5 authentication (see top of this file)
      if DISABLE_CRAM then
        fh:write("  mechanisms = plain login\n")
      else
        fh:write("  mechanisms = plain login cram-md5\n")
      end

      fh:write([[
  passdb passwd-file {
    args = /etc/dovecot/deny.%s
    deny = yes
  }
  passdb passwd-file {
    args = /etc/dovecot/passwd
  }
  userdb passwd-file {
    args = /etc/dovecot/passwd
  }
  socket listen {
    client {
      path = /var/spool/postfix/private/auth
      mode = 0660
      user = postfix
      group = mail
    }
    master {
      path = /var/run/dovecot/auth-master
      mode = 0600
      user = mail
      group = mail
    }
  }
  user = root
}
plugin {
  quota = maildir:User quota
  quota_rule = *:storage=0
  quota_rule2 = Trash:storage=50M
]])
      if opts.quotawarning then
        fh:write("  quota_warning = storage=95%% " .. LC.liveconfig.libexecdir .. "/mailquota.sh 95\n")
        fh:write("  quota_warning2 = storage=80%% " .. LC.liveconfig.libexecdir .. "/mailquota.sh 80\n")
      end
      fh:write([[
}
]])

    else
      -- Dovecot version 2.x
      LC.log.print(LC.log.DEBUG, "Dovcot version 2.x")

      fh:write([[

login_greeting = server ready
]])
      -- optionally disable CRAM-MD5 authentication (see top of this file)
      if DISABLE_CRAM then
        fh:write("auth_mechanisms = plain login\n")
      else
        fh:write("auth_mechanisms = plain login cram-md5\n")
      end
      fh:write([[
disable_plaintext_auth = no
first_valid_uid = ]], uid, "\n", [[
postmaster_address = postmaster@localhost
log_timestamp = "%Y-%m-%d %H:%M:%S "
mail_privileged_group = mail
mail_plugins = $mail_plugins quota
passdb {
  args = /etc/dovecot/deny.%s
  driver = passwd-file
  deny = yes
}
passdb {
  args = /etc/dovecot/passwd
  driver = passwd-file
}
plugin {
  quota = maildir:User quota
  quota_rule = *:storage=0
  quota_rule2 = Trash:storage=+50M
}
]])
      if opts.quotawarning then
        fh:write([[
plugin {
  quota_warning = storage=95%% quota-warning 95 %u
  quota_warning2 = storage=80%% quota-warning 80 %u
}
service quota-warning {
  executable = script ]], LC.liveconfig.libexecdir, [[/mailquota.sh
  user = mail
  unix_listener quota-warning {
    user = mail
    group = mail
    mode = 0600
  }
}
]])

      end

      fh:write([[

protocols = imap pop3
protocol imap {
  mail_plugins = $mail_plugins imap_quota
]])
      if opts.imapconn then
        fh:write("  mail_max_userip_connections = ", opts.imapconn, "\n")
      end
      fh:write([[
}
protocol pop3 {
  pop3_client_workarounds = outlook-no-nuls oe-ns-eoh
]])
      if opts.pop3conn then
        fh:write("  mail_max_userip_connections = ", opts.pop3conn, "\n")
      end
      fh:write([[
}
protocol lda {
  mail_plugins = $mail_plugins sieve
}
]])

      -- SSL configuration
      if opts.ssl then
        -- SSL enabled
        if opts.sslonly then
          fh:write("ssl = required\n")
        else
          fh:write("ssl = yes\n")
        end
        if v_major > 2 or (v_major == 2 and v_minor > 0) then
          -- ssl_protocols: supported since v2.1
          local v = string.match(LC.distribution.version, "^%d+")
          local protocols = "!SSLv2 !SSLv3"
          if (LC.distribution.name == "Debian" and tonumber(v) >= 9)
             or (LC.distribution.name == "Ubuntu" and tonumber(v) >= 20)
             or (LC.distribution.family == "RedHat" and tonumber(v) >= 8) then
            -- Debian 9+, Ubuntu20+, CentOS8+ ship with OpenSSL 1.1, which doesn't know about SSLv2 any more!
            protocols = "!SSLv3"
          end
          if v_major > 2 or (v_major == 2 and v_minor >= 3) then
            -- with Dovecot 2.3, ssl_protocols was replaced by ssl_min_protocol, specifying the minimum supported
            -- SSL/TLS version. Default is TLSv1 (no more SSLv2/SSLv3), so we don't need any explicit setting here.
            -- But: Dovecot v2.3+ requires an explicit DH parameters file. Copy parameters from LiveConfig unless
            -- local DH parameters have been generated:
            if not LC.fs.is_file(configpath .. "/dh.pem") then
              LC.fs.copy("/etc/liveconfig/dhparam.dist.pem", configpath .. "/dh.pem")
              LC.fs.setperm(configpath .. "/dh.pem", "0644", "root", "root")
            end
            fh:write("ssl_dh = <", configpath, "/dh.pem\n")
          else
            fh:write("ssl_protocols = ", protocols, "\n")
          end
        end
        if opts.sslstrong then
          -- modern/strong ciphers (actually, defending BEAST is not necessary on POP3/IMAP, but most dumb PCI scans don't care :(
          fh:write("ssl_cipher_list = ", LC.liveconfig.DEFAULT_SSL_STRONG_CIPHERS, "\n")
        else
          fh:write("ssl_cipher_list = ", LC.liveconfig.DEFAULT_SSL_CIPHERS, "\n")
        end
        fh:write("ssl_cert = <", cfg.crtfile, "\n")
        fh:write("ssl_key = <", cfg.keyfile, "\n")

        if not opts.sslports then
      fh:write([[
service imap-login {
  inet_listener imaps {
    port = 0
  }
}
service pop3-login {
  inet_listener pop3s {
    port = 0
  }
}
]])
        end
      else
        -- no SSL
        fh:write("ssl = no\n")
      end

      local group, user
      if v_major > 2 or (v_major == 2 and v_minor >= 3) then
        group = ""
        user = ""
      else
        group = " mail"
        user = " mail"
      end

      fh:write([[
service auth {
  unix_listener auth-userdb {
    group = mail
    mode = 0600
    user = mail
  }
  unix_listener /var/spool/postfix/private/auth {
    group = mail
    mode = 0660
    user = postfix
  }
  unix_listener auth-master {
    group =]], group, "\n", [[
    mode = 0600
    user =]], user, "\n", [[
  }
  user = root
}
userdb {
  args = /etc/dovecot/passwd
  driver = passwd-file
}

]])

      if v_major > 2 or (v_major == 2 and v_minor >= 3) then
        fh:write([[
service stats {
  unix_listener stats-writer {
     user = mail
     group = mail
  }
}

]])
      end

      if opts.autofolders then
        -- let Dovecot automatically create some default IMAP folders
        -- usually this is done by the MUA, but at least recent (brain-dead) Outlook doesn't do...
        fh:write([[
namespace inbox {
  inbox = yes
  mailbox Drafts {
    auto = subscribe
    special_use = \Drafts
  }
  mailbox Junk {
    special_use = \Junk
  }
  mailbox Trash {
    auto = subscribe
    special_use = \Trash
  }
  mailbox Sent {
    auto = subscribe
    special_use = \Sent
  }
  mailbox "Sent Messages" {
    auto = no
    special_use = \Sent
  }
}
]])
      end

    end

    -- include "dovecot.local.conf" (if existing)
    if LC.fs.is_file(configpath .. "/dovecot.local.conf") then
      fh:write("# local configuration options\n")
      fh:write("!include_try " .. configpath .. "/dovecot.local.conf\n")
    end

    LC.liveconfig.writeFooter(fh)
    fh:close()

    if v_major == 1 then
      -- Dovcot version 1.x
      if opts.quotawarning then
        -- create copy of config file, remove quota warnings to avoid loops:
        os.execute("sed -e 's/^[ \\t]*quota_warning[0-9]*[ \\t]*=.*$//' " .. configfile .. ".tmp >" .. configfile .. ".noquota")
        LC.fs.setperm(configfile .. ".noquota", "0644", "dovecot", "root")
      elseif LC.fs.is_file(configfile .. ".noquota") then
        -- remove old "noquota" config file
        os.execute("rm " .. configfile .. ".noquota")
      end
    end

    local denyfile  = cfg["configpath"] .. "/deny"
    status, msg = checkDenyFile(denyfile)
    if not status then return status, msg end

    -- enable configuration
    LC.fs.rename(configfile .. ".tmp", configfile)

  end -- if not NOUPDATE

  -- update status file
  LC.liveconfig.writeStatus(statusfile, "dovecot", opts.revision, os.date("%Y-%m-%d %H:%M:%S %Z"))

  return true
end

-- ---------------------------------------------------------------------------
-- replace placeholder with configuration values
-- ---------------------------------------------------------------------------
local function replace_pattern(var, data)
  var = string.gsub(var, "%%C", data.contract)
  var = string.gsub(var, "%%I", data.id)
  return var
end

-- ---------------------------------------------------------------------------
-- addMailbox
--
-- Add a mailbox to dovecot server
-- ---------------------------------------------------------------------------
function addMailbox(cfg, opts, data)

  local userfile  = cfg["userfile"]
  if opts and opts.prefix then
    -- use prefix (for testing etc.)
    userfile = opts.prefix .. userfile
  end

  LC.log.print(LC.log.INFO, "Adding/updating user account " .. data.name .. "@" .. data.domain .. " at dovecot config file: " .. userfile)

  -- get uid and gid for user "mail"
  local uid = LC.sys.user_exists("mail")
  if uid == false then
    return false, "System user 'mail' does not exist, please create user"
  end
  local gid = LC.sys.group_exists("mail")
  if gid == false then
    return false, "System group 'mail' does not exist, please create group"
  end

  -- check if subscription directory already exists, otherwise create it:
  if not LC.fs.is_dir('/var/mail/' .. data.contract) then
    if LC.fs.is_file('/var/mail/' .. data.contract) then
      os.remove('/var/mail/' .. data.contract)
    end
    LC.fs.mkdir_rec('/var/mail/' .. data.contract)
    LC.fs.setperm('/var/mail/' .. data.contract, "2700", "mail", "mail")
  end

  -- check if mailbox directory already exists, otherwise create it:
  if not LC.fs.is_dir('/var/mail/' .. data.contract .. '/' .. data.id) then
    LC.fs.mkdir_rec('/var/mail/' .. data.contract .. '/' .. data.id)
    LC.fs.setperm('/var/mail/' .. data.contract .. '/' .. data.id, "2700", "mail", "mail")
  end

  -- add entry to user file
  local fhr, fhw, msg
  if not LC.fs.is_file(userfile) then
    fhr, msg = io.open(userfile, "a")
    if fhr == nil then
      LC.log.print(LC.log.ERR, "Can't open '", userfile, "' for appending: ", msg)
      return false, "Can't open '" .. userfile .. "' for appending: " .. msg
    end
    -- adjust owner & permissions - only Dovecot needs access to this file, so set mod 600 !!!
    LC.fs.setperm(userfile, "0600", "dovecot", "root")
    fhr:close()
  end

  fhr, msg = io.open(userfile, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", userfile, "' for reading: ", msg)
    return false, "Can't open '" .. userfile .. "' for reading: " .. msg
  end

  fhw, msg = io.open(userfile .. ".tmp", "w")
  if fhw == nil then
    LC.log.print(LC.log.ERR, "Can't open '", userfile .. ".tmp", "' for writing: ", msg)
    fhr:close()
    return false, "Can't open '" .. userfile .. ".tmp" .. "' for writing: " .. msg
  end
  -- adjust owner & permissions - only Dovecot needs to read this file:
  LC.fs.setperm(userfile .. ".tmp", "0600", "dovecot", "root")

  -- build new/updated entry:
  local new_line = nil
  if data.mailbox == true or data.autoresponder == true then
    local pwd, algo
    if data.mailbox == false or data.password == nil then
      pwd = "!"
      algo = ""
    elseif string.match(data.password, "^{[%u1256%-]+}[%w./]+") then
      -- pre-hashed password in Dovecot format - use without modification
      pwd = data.password
      algo = ""
    elseif string.len(data.password) == 34 and string.match(data.password, "^$1$[%w./][%w./][%w./][%w./][%w./][%w./][%w./][%w./]$[%w./]+$") then
      -- MD5-CRYPT password (propably imported via SOAP interface)
      pwd = data.password
      algo = "{MD5-CRYPT}"
    else
      -- create CRAM-MD5 hash
      pwd = LC.crypt.cram_md5(data.password)
      algo = "{CRAM-MD5}"
    end
    local p_home = replace_pattern(HOME, data)
    local p_mail = replace_pattern(USERDB_MAIL, data)
    local p_sieve = replace_pattern(USERDB_SIEVE, data)
    new_line = data.name .. "@" .. data.domain .. ":" .. algo .. pwd .. ":" .. uid .. ":" .. gid .. "::" .. p_home .. "::"
    if p_mail ~= "" then
      -- userdb_mail:
      new_line = new_line .. p_mail
    end
    if data.quota ~= nil and data.quota > 0 then
      -- userdb_quota_rule:
      new_line = new_line .. " userdb_quota_rule=*:storage=" .. data.quota .. "MB"
    end
    if p_sieve ~= "" then
      -- userdb_sieve:
      new_line = new_line .. " " .. p_sieve
    end
  end

  -- park this mailbox?
  if data.parking then new_line = '#' .. new_line end

  -- search/replace existing entry
  local search
  if data.old_addr == nil then
    search = "^#?" .. data.name .. "@" .. data.domain .. ":"
  else
    -- rename existing mailbox
    search = "^#?" .. data.old_addr .. ":"
  end
  search = string.gsub(search, "%%", "%%%%")
  search = string.gsub(search, "%.", "%%.")
  search = string.gsub(search, "%+", "%%+")
  search = string.gsub(search, "%-", "%%-")
  search = string.gsub(search, "%*", "%%*")
  local line
  local found = false
  while true do
    line = fhr:read()
    if line == nil then break end
    if string.find(line, search) ~= nil then
      found = true
      if new_line ~= nil then
        fhw:write(new_line, "\n")
      end
    elseif line ~= "" then
      fhw:write(line, "\n")
    end
  end

  fhr:close()

  if found == false and new_line ~= nil then
    -- append new entry
    fhw:write(new_line, "\n")
  end

  fhw:close()

  -- move temporary file to new password file
  LC.fs.rename(userfile .. ".tmp", userfile)

  -- check for autoresponder
  local sievepath = "/var/mail/" .. data.contract .. "/" .. data.id
  if data.autoresponder == true then
    -- create autoresponder
    fhw, msg = io.open(sievepath .. "/dovecot.sieve.tmp", "w")
    if fhw == nil then
      LC.log.print(LC.log.ERR, "Can't open '", sievepath .. "/dovecot.sieve.tmp", "' for writing: ", msg)
      return false, "Can't open '" .. sievepath .. "/dovecot.sieve.tmp" .. "' for writing: " .. msg
    end
    -- adjust owner & permissions - only Dovecot (mail) needs to read this file:
    LC.fs.setperm(sievepath .. "/dovecot.sieve.tmp", "0640", "mail", "mail")
    -- escape quotes
    local as = string.gsub(data.autosubject, "\\", "\\\\")
    as = string.gsub(as, "\"", "\\\"")
    local am = string.gsub(data.automessage, "\\", "\\\\")
    am = string.gsub(am, "\"", "\\\"")
    local useVars = string.match(as, "${subject}") or string.match(am, "${subject}")
    fhw:write("# Created by LiveConfig\n")
    if useVars then
      fhw:write("require [\"date\", \"relational\", \"vacation\", \"variables\"];\n")
      fhw:write("if header :matches \"subject\" \"*\" {\n")
      fhw:write("  set \"subject\" \"${1}\";\n")
      fhw:write("}\n")
    else
      fhw:write("require [\"date\", \"relational\", \"vacation\"];\n")
    end
    -- only send auto-reply if message is NOT tagges as spam:
    if data.autoreplyend ~= nil then
      fhw:write("if allof(not header :contains \"X-Spam-Flag\" \"YES\",\n")
      fhw:write("         currentdate :value \"lt\" \"date\" \"", data.autoreplyend, "\") {\n")
    else
      fhw:write("if not header :contains \"X-Spam-Flag\" \"YES\" {\n")
    end
    fhw:write("  vacation\n")
    fhw:write("    :days 1\n")
    fhw:write("    :subject \"", as, "\"\n")
    fhw:write("    :addresses [\"", data.name, "@", data.domain, "\"")
    if type(data.aliases) == "table" then
      local i,s
      for i, s in ipairs(data.aliases) do
        fhw:write(", \"", s, "@", data.domain, "\"")
      end
    end
    fhw:write("]\n")
    fhw:write("    \"", am, "\";\n")
    fhw:write("}\n")
    if not data.mailbox then
      -- no real mailbox (forward-only), so discard message:
      fhw:write("discard;\n")
    end
    fhw:close()
    LC.fs.rename(sievepath .. "/dovecot.sieve.tmp", sievepath .. "/dovecot.sieve")
  else
    -- delete autoresponder if still existing...
    if LC.fs.is_file(sievepath .. "/dovecot.sieve") and not LC.fs.is_symlink(sievepath .. "/dovecot.sieve") then
      os.remove(sievepath .. "/dovecot.sieve")
      os.remove(sievepath .. "/dovecot.svbin")
    end
  end

  return true

end

-- ---------------------------------------------------------------------------
-- editMailbox
--
-- Edit mailbox configuration
-- ---------------------------------------------------------------------------
function editMailbox(cfg, opts, data)
  -- same as addMailbox():
  return addMailbox(cfg, opts, data)
end

-- ---------------------------------------------------------------------------
-- deleteMailbox
--
-- Delete a mailbox
-- ---------------------------------------------------------------------------
function deleteMailbox(cfg, opts, data)

  local userfile  = cfg["userfile"]
  if opts and opts.prefix then
    -- use prefix (for testing etc.)
    userfile = opts.prefix .. userfile
  end

  LC.log.print(LC.log.INFO, "Deleting mailbox " .. data.name .. "@" .. data.domain .. " from dovecot config file: " .. userfile)

  -- remove entry from user file
  local fhr, fhw, msg
  if not LC.fs.is_file(userfile) then
    fhr, msg = io.open(userfile, "a")
    if fhr == nil then
      LC.log.print(LC.log.ERR, "Can't open '", userfile, "' for appending: ", msg)
      return false, "Can't open '" .. userfile .. "' for appending: " .. msg
    end
    -- adjust owner & permissions - only Dovecot needs access to this file, so set mod 600 !!!
    LC.fs.setperm(userfile, "0600", "dovecot", "root")
    fhr:close()
  end

  fhr, msg = io.open(userfile, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", userfile, "' for reading: ", msg)
    return false, "Can't open '" .. userfile .. "' for reading: " .. msg
  end

  fhw, msg = io.open(userfile .. ".tmp", "w")
  if fhw == nil then
    LC.log.print(LC.log.ERR, "Can't open '", userfile .. ".tmp", "' for writing: ", msg)
    fhr:close()
    return false, "Can't open '" .. userfile .. ".tmp" .. "' for writing: " .. msg
  end
  -- adjust owner & permissions - only Dovecot needs to read this file:
  LC.fs.setperm(userfile .. ".tmp", "0600", "dovecot", "root")

  -- search/remove existing entry
  local search = "^#?" .. data.name .. "@" .. data.domain .. ":"
  search = string.gsub(search, "%%", "%%%%")
  search = string.gsub(search, "%.", "%%.")
  search = string.gsub(search, "%+", "%%+")
  search = string.gsub(search, "%-", "%%-")
  search = string.gsub(search, "%*", "%%*")
  local line
  local found = false
  while true do
    line = fhr:read()
    if line == nil then break end
--    LC.log.print(LC.log.DEBUG, "Line: '" .. line .. "'")
    if string.find(line, search) ~= nil then
      found = true
--      LC.log.print(LC.log.DEBUG, "  FOUND: '" .. line .. "'")
    else
--      LC.log.print(LC.log.DEBUG, "  not found")
      fhw:write(line, "\n")
    end
  end

  fhr:close()

  fhw:close()

  -- move temporary file to new password file
  LC.fs.rename(userfile .. ".tmp", userfile)

  -- remove mailbox directory
  if data.id and LC.fs.is_dir("/var/mail/" .. data.contract .. "/" .. data.id) then
    os.execute("rm -rf /var/mail/" .. data.contract .. "/" .. data.id)
  end

  if LC.fs.is_dir("/var/mail/" .. data.contract) and LC.fs.filecount("/var/mail/" .. data.contract) == 0 then
    os.execute("rmdir /var/mail/" .. data.contract)
  end

  -- remove from deny.imap file:
  editlock(cfg, { ["mode"] = 0, ["accounts"] = { data.name .. "@" .. data.domain } } )

  return true

end

-- ---------------------------------------------------------------------------
-- editlock()
-- ---------------------------------------------------------------------------
function editlock(cfg, data)
  -- lock mode = 0: unlock
  --             1: lock (allow receiving, but not sending)
  --             2: disable (neither receiving nor sending allowed)

  local fhr, fhw, status, msg

  if data.accounts == nil then return true end

  -- edit /etc/dovecot/passwd (enable/disable SASL accounts)
  local denyfile  = cfg["configpath"] .. "/deny"
  status, msg = checkDenyFile(denyfile)
  if not status then return status, msg end

  fhr, msg = io.open(denyfile .. ".imap", "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", denyfile, ".imap' for reading: ", msg)
    return false, "Can't open '" .. denyfile .. ".imap' for reading: " .. msg
  end

  fhw, msg = io.open(denyfile .. ".imap.tmp", "w")
  if fhw == nil then
    fhr:close()
    LC.log.print(LC.log.ERR, "Can't open '", denyfile, ".imap.tmp' for writing: ", msg)
    return false, "Can't open '" .. denyfile .. ".imap.tmp' for writing: " .. msg
  end
  -- adjust owner & permissions - only Dovecot needs to read this file:
  LC.fs.setperm(denyfile .. ".imap.tmp", "0600", "dovecot", "root")

  while true do
    line = fhr:read()
    if line == nil then break end
    if not LC.liveconfig.inTable(line, data.accounts) then fhw:write(line, "\n") end
  end

  fhr:close()

  if data.mode == 2 then
    -- add accounts to "deny" file to forbid e-mail access
    for _,line in pairs(data.accounts) do
      fhw:write(line, "\n")
    end
  end

  fhw:close()

  LC.fs.rename(denyfile .. ".imap.tmp", denyfile .. ".imap")

  return true
end

-- <EOF>----------------------------------------------------------------------
