--  _    _          ___           __ _     (R)
-- | |  (_)_ _____ / __|___ _ _  / _(_)__ _
-- | |__| \ V / -_) (__/ _ \ ' \|  _| / _` |
-- |____|_|\_/\___|\___\___/_||_|_| |_\__, |
--                                    |___/
-- Copyright (c) 2009-2021 LiveConfig GmbH.
-- ---------------------------------------------------------------------------
-- common/lua/postfix.lua
-- Lua module to manage postfix smtp server
-- ---------------------------------------------------------------------------
-- This is the "driver" module for postfix smtp server.
-- It must be loaded by the command
--   LC.smtp.load("postfix")
-- Usually, this should happen at liveconfig.lua (or, if you have a customized
-- module, at custom.lua)
-- ---------------------------------------------------------------------------

-- Imported functions
local LC = LC
local LCS = LCS -- import liveconfig global storage
local type = type
local io = io
local os = os
local string = string
local ipairs = ipairs
local pairs = pairs
local table = table
local tonumber = tonumber

-- Module declaration
module("postfix")

-- Exported functions
--   detect()
--   install()
--   uninstall()
--   updateSSL()
--   configure()
--   addMailbox()
--   editMailbox()
--   deleteMailbox()
--   update_lclogparse()
--   editlock()
--   updateDomain()

-- Optionally disable updating main.cf/master.cf by LiveConfig
if NOUPDATE == nil then
  NOUPDATE = false
end

-- local configuration overrides
if LOCALCONFIG == nil then
  LOCALCONFIG = { }
end
if LOCALMASTER == nil then
  LOCALMASTER = { }
end

if TXT_SASL_REJECT == nil then
  TXT_SASL_REJECT = "Sending e-mails not allowed with this account."
end

if TXT_DOMAIN_DISABLED == nil then
  TXT_DOMAIN_DISABLED = "Domain disabled"
end

if TXT_DOMAIN_EXT_REJECTED == nil then
  TXT_DOMAIN_EXT_REJECTED = "External e-mails from this domain not allowed."
end

-- ---------------------------------------------------------------------------
-- removeLine
--
-- Remove a line from given map file
-- ---------------------------------------------------------------------------
local function removeLine(filename, domain)
  local fhr, fhw, msg, line
  fhr, msg = io.open(filename, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", filename, "' for reading: ", msg)
    return false
  end

  fhw, msg = io.open(filename .. ".tmp", "w")
  if fhw == nil then
    LC.log.print(LC.log.ERR, "Can't open '", filename .. ".tmp", "' for writing: ", msg)
    fhr:close()
    return false
  end
  -- adjust owner & permissions - only Postfix needs to read this file:
  LC.fs.setperm(filename .. ".tmp", 640, "root", "postfix")

  local found = false
  local search = "^" .. domain
  search = string.gsub(search, "%.", "%%.")
  search = string.gsub(search, "%-", "%%-")

  while true do
    line = fhr:read()
    if line == nil then break end
    if string.find(line, search .. "$") ~= nil or string.find(line, search .. "%s") ~= nil then
      found = true
    else
      -- only write line if not matched
      fhw:write(line)
      fhw:write("\n")
    end
  end

  fhr:close()

  fhw:close()

  if found then
    LC.fs.rename(filename .. ".tmp", filename)
    return true
  else
    os.remove(filename .. ".tmp")
    return false
  end

end

-- ---------------------------------------------------------------------------
-- updateMap
--
-- Remove a list of search expressions from given map file, add new data
-- ---------------------------------------------------------------------------
local function updateMap(filename, search, replace)
  local fhr, fhw, msg, line
  LC.mutex.lock(filename)
  fhr, msg = io.open(filename, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", filename, "' for reading: ", msg)
    LC.mutex.unlock(filename)
    return false
  end

  fhw, msg = io.open(filename .. ".tmp", "w")
  if fhw == nil then
    LC.log.print(LC.log.ERR, "Can't open '", filename .. ".tmp", "' for writing: ", msg)
    fhr:close()
    LC.mutex.unlock(filename)
    return false
  end
  -- adjust owner & permissions - only Postfix needs to read this file:
  LC.fs.setperm(filename .. ".tmp", 640, "root", "postfix")

  local found = false
  local modified = false

  while true do
    line = fhr:read()
    if line == nil then break end
    found = false
    for i, s in ipairs(search) do
      if string.find(line, s) ~= nil then
        found = true
        modified = true
        break
      end
    end
    if found == false then
      -- only write line if not matched
      fhw:write(line, "\n")
    end
  end

  fhr:close()

  if replace ~= nil and replace ~= "" then
    fhw:write(replace)
    modified = true
  end

  fhw:close()

  if modified then
    LC.fs.rename(filename .. ".tmp", filename)
    LC.mutex.unlock(filename)
    return true
  else
    os.remove(filename .. ".tmp")
    LC.mutex.unlock(filename)
    return false
  end

end

-- ---------------------------------------------------------------------------
-- getBinaryVersion()
--
-- Return the binary version number from postfix
-- ---------------------------------------------------------------------------
local function getBinaryVersion(bin)
  local handle = io.popen(postconfpath .. "postconf mail_version", "r")
  local ret = handle:read("*a")
  handle:close()
  local v = string.match(ret, "(%d+%.%d+[%.%d+]*)")
  return v
end

-- ---------------------------------------------------------------------------
-- detect()
--
-- Detect if an postfix smtp server is installed. If yes, return a table containing
-- all relevant informations:
--   type          => "postfix"
--   binname       => file name of the postfix binary (eg. "/usr/sbin/postfix")
--   binversion    => binary version (eg. "2.5.5")
--   pkgname       => name of the package (see below)
--   pkgversion    => package version number (eg. "2.2.3-4+etch11")
--   configpath    => Main configuration path for postfix
--   configfile    => LiveConfigs' configuration file (snippet)
--   statusfile    => LiveConfigs' status file (containing config version etc.)
--   defaultlog    => Default log file (for default and unknown/unconfigured vhosts)
--   start_cmd     => command to start postfix
--   stop_cmd      => command to stop postfix
--   reload_cmd    => command to re-load configuration
--   restart_cmd   => command to re-start postfix process
--
-- If pkgversion is 'nil', then Postfix Server was most propably not installed
-- as native distribution package, but compiled/installed manually.
-- ---------------------------------------------------------------------------
function detect()

  local bin
  -- first check for distribution-specific packages
  if LC.distribution.family == "Debian" or LC.distribution.family == "Gentoo" or LC.distribution.family == "RedHat" or LC.distribution.family == "BSD" or LC.distribution.family == "SUSE"  then
    -- Debian/Ubuntu and Gentoo
    local pkg, v = LC.distribution.hasPackage('postfix', 'postfix3')
    if pkg ~= nil then
      LC.log.print(LC.log.DEBUG, "Found package '", pkg, "' (Version ", v, ")")
      if LC.distribution.family == "BSD" then
        postconfpath = "/usr/local/sbin/"
        bin = "/usr/local/sbin/postfix"
      else
        postconfpath = "/usr/sbin/"
        bin = "/usr/sbin/postfix"
      end
      -- get binary version
      local bv = getBinaryVersion(bin)
      if bv ~= nil then
        local data
        if LC.distribution.family == "RedHat" then
          data = {
            ["type"]          = "postfix",
            ["binname"]       = bin,
            ["binversion"]    = bv,
            ["pkgname"]       = pkg,
            ["pkgversion"]    = v,
            ["configpath"]    = "/etc/postfix",
            ["configfile"]    = "/etc/postfix/main.cf",
            ["statusfile"]    = "/etc/postfix/liveconfig.status",
            ["domainsfile"]   = "/etc/postfix/virtual_domains",
            ["aliasfile"]     = "/etc/postfix/virtual_alias",
            ["mailboxfile"]   = "/etc/postfix/virtual_mailbox",
            ["defaultlog"]    = "/var/log/maillog",
            ["crtfile"]       = "/etc/pki/tls/certs/postfix.crt",
            ["keyfile"]       = "/etc/pki/tls/private/postfix.key",
            ["start_cmd"]     = "/sbin/service postfix start",
            ["stop_cmd"]      = "/sbin/service postfix stop",
            ["reload_cmd"]    = "/sbin/service postfix reload",
            ["restart_cmd"]   = "/sbin/service postfix restart",
            ["logfile"]       = "/var/log/maillog",
            ["units"]         = {
              ["postfix"]       = "postfix",
              ["spamassassin"]  = "spamassassin",
              ["lcsam"]         = "lcsam",
              ["dkim"]          = "opendkim",
              ["postgrey"]      = "postgrey",
              ["lclogparse"]    = "lclogparse",
              ["lcpolicyd"]     = "lcpolicyd",
              ["clamav"]        = "clamd@scan",
              ["clamav-milter"] = "clamav-milter",
            }
          }
        else
          data = {
            ["type"]          = "postfix",
            ["binname"]       = bin,
            ["binversion"]    = bv,
            ["pkgname"]       = pkg,
            ["pkgversion"]    = v,
            ["configpath"]    = "/etc/postfix",
            ["configfile"]    = "/etc/postfix/main.cf",
            ["statusfile"]    = "/etc/postfix/liveconfig.status",
            ["domainsfile"]   = "/etc/postfix/virtual_domains",
            ["aliasfile"]     = "/etc/postfix/virtual_alias",
            ["mailboxfile"]   = "/etc/postfix/virtual_mailbox",
            ["defaultlog"]    = "/var/log/mail.log",
            ["crtfile"]       = "/etc/postfix/ssl-cert.pem",
            ["keyfile"]       = "/etc/postfix/ssl-key.pem",
            ["start_cmd"]     = "service postfix start",
            ["stop_cmd"]      = "service postfix stop",
            ["reload_cmd"]    = "service postfix reload",
            ["restart_cmd"]   = "service postfix restart",
            ["logfile"]       = "/var/log/mail.log",
            ["units"]         = {
              ["postfix"]       = "postfix",
              ["spamassassin"]  = "spamassassin",
              ["lcsam"]         = "lcsam",
              ["dkim"]          = "opendkim",
              ["postgrey"]      = "postgrey",
              ["lclogparse"]    = "lclogparse",
              ["lcpolicyd"]     = "lcpolicyd",
              ["clamav"]        = "clamav-daemon",
              ["clamav-milter"] = "clamav-milter",
            },
          }
          if LC.distribution.family == "SUSE" then
            data.logfile = "/var/log/mail"
          end
          if LC.distribution.family == "Debian" then
            data.crtfile = "/etc/ssl/certs/postfix.crt"
            data.keyfile = "/etc/ssl/private/postfix.key"
            if (not LC.fs.is_file(data.logfile) and LC.fs.is_dir("/var/log/journal")) then
              data.logfile = "journald:postfix@-.service"
            end
            if LC.fs.is_file("/etc/default/spamd") then
              -- Debian 12 has "spamd" instead of "spamassassin" service
              data.units.spamassassin = "spamd"
            end
          end
        end
        -- SpamAssassin (via lcsam):
        data["lcsam_milter_socket"] = "unix:/var/run/lcsam.sock"
        -- now check for clamav-milter:
        if LC.distribution.family == "Debian" then
          pkg, v = LC.distribution.hasPackage("clamav-milter")
          if pkg ~= nil then
            data["clamav"]    = v;
            data["clamav_milter_cfg"] = "/etc/clamav/clamav-milter.conf";
            data["clamav_milter_socket"] = "unix:/var/run/clamav/clamav-milter.ctl";
          end
        elseif LC.distribution.family == "RedHat" then
          pkg, v = LC.distribution.hasPackage("clamav-milter")
          if pkg ~= nil then
            data["clamav"]    = v;
            -- default settings (RedHat 5.x)
            data["clamav_milter_cfg"] = "/etc/clamav-milter.conf";
            data["clamav_milter_socket"] = "unix:/var/clamav/clmilter.socket";
            if LC.fs.is_file("/etc/clamav/clamav-milter.conf") then
              -- RedHat 6.x
              data["clamav_milter_cfg"] = "/etc/clamav/clamav-milter.conf";
              data["clamav_milter_socket"] = "unix:/var/run/clamav/clamav-milter.ctl";
            elseif LC.fs.is_file("/etc/mail/clamav-milter.conf") then
              -- RedHat 7.x
              data["clamav_milter_cfg"] = "/etc/mail/clamav-milter.conf";
              data["clamav_milter_socket"] = "unix:/var/run/clamav-milter/clamav-milter.socket";
            end
            if LC.fs.is_file("/etc/clamav-milter.conf")
               and LC.exec("grep -i '^MilterSocket /var/run/clamav/clamav-milter.sock' /etc/clamav-milter.conf") == 0 then
              -- ClamAV via EPEL @ CentOS 6:
              data["clamav_milter_socket"] = "unix:/var/run/clamav/clamav-milter.sock"
            end
          end
        elseif LC.distribution.family == "SUSE" then
          pkg, v = LC.distribution.hasPackage("clamav")
          if pkg ~= nil and LC.fs.is_file("/usr/sbin/clamav-milter") then
            data["clamav"]    = v;
            data["clamav_milter_cfg"] = "/etc/clamav-milter.conf";
            data["clamav_milter_socket"] = "unix:/var/run/clamav/clamav-milter-socket";
          end
        elseif LC.distribution.family == "Gentoo" then
          pkg, v = LC.distribution.hasPackage("clamav")
          if pkg ~= nil and LC.fs.is_file("/usr/sbin/clamav-milter") then
            data["clamav"]    = v;
            data["clamav_milter_cfg"] = "/etc/clamav-milter.conf";
            data["clamav_milter_socket"] = "unix:/var/run/clamav/clamav-milter.sock";
          end
        end
        -- check for Postgrey:
        pkg, v = LC.distribution.hasPackage("postgrey")
        if pkg ~= nil then
          data["postgrey"] = v
        end
        -- check for SpamAssassin
        local sa = LC.spamassassin.detect()
        if sa ~= nil then
          data["spamassassin"] = sa.pkgversion
        end
        -- check for DKIM
        local dkim = LC.dkim.detect()
        if dkim ~= nil then
          data["dkim"] = dkim.pkgversion
        end
        -- check for DANE/TLSA support
        local v_major, v_minor = string.match(data["binversion"], "^(%d+)%.(%d+)")
        v_major = tonumber(v_major)
        v_minor = tonumber(v_minor)
        if v_major > 2 or (v_major == 2 and v_minor >= 11) then
          data["dane"] = true
        end
        return data
      end
      -- else: fall trough, to check for custom postfix installation
      LC.log.print(LC.log.DEBUG, "LC.postfix.detect(): Found Postfix package '", pkg, "', but no binary at ", bin)
    end
  end
end

-- ---------------------------------------------------------------------------
-- install(cfg, opts)
--
-- Install LiveConfig configuration files for Postfix Server
-- ---------------------------------------------------------------------------
function install(cfg, opts)
  LC.log.print(LC.log.INFO, "Activating management of Postfix Server")

  local fh, status, msg
  local configpath  = cfg["configpath"]
  local configfile  = cfg["configfile"]
  local masterfile  = cfg["configpath"] .. "/master.cf"
  local statusfile  = cfg["statusfile"]
  local domainsfile = cfg["domainsfile"]
  local aliasfile   = cfg["aliasfile"]

  if opts and opts.prefix then
    -- use prefix (for testing etc.)
    configpath = opts.prefix .. configpath
    configfile = opts.prefix .. configfile
    masterfile = opts.prefix .. masterfile
    statusfile = opts.prefix .. statusfile
    domainsfile = opts.prefix .. domainsfile
    aliasfile   = opts.prefix .. aliasfile
  end

  -- back up old default main.cf file (if existing)
  if LC.fs.is_file(configfile) and not LC.fs.is_file(configfile .. ".lcbak") then
    LC.fs.rename(configfile, configfile .. ".lcbak")
  end
  -- back up old default master.cf file (if existing)
  if LC.fs.is_file(masterfile) and not LC.fs.is_file(masterfile .. ".lcbak") then
    LC.fs.rename(masterfile, masterfile .. ".lcbak")
  end
  -- back up old domainsfile (if existing)
  if LC.fs.is_file(domainsfile) and not LC.fs.is_file(domainsfile .. ".lcbak") then
    LC.fs.rename(domainsfile, domainsfile .. ".lcbak")
  end
  -- back up old domainsfile database file (if existing)
  if LC.fs.is_file(domainsfile .. ".db") and not LC.fs.is_file(domainsfile .. ".db.lcbak") then
    LC.fs.rename(domainsfile .. ".db", domainsfile .. ".db.lcbak")
  end
  -- back up old aliasfile (if existing)
  if LC.fs.is_file(aliasfile) and not LC.fs.is_file(aliasfile .. ".lcbak") then
    LC.fs.rename(aliasfile, aliasfile .. ".lcbak")
  end
  -- back up old aliasfile database file (if existing)
  if LC.fs.is_file(aliasfile .. ".db") and not LC.fs.is_file(aliasfile .. ".db.lcbak") then
    LC.fs.rename(aliasfile .. ".db", aliasfile .. ".db.lcbak")
  end

  -- check whether mail user and group exists or not
  -- if not create user and group
  if not LC.sys.user_exists("mail") then
    if not LC.sys.group_exists("mail") then
      LC.users.addGroup("mail")
    end
    LC.users.addUser("mail", "mail", "/var/mail", "nologin")
  end

  -- write default config file main.cf:
  configure(cfg, opts)

  -- create lclogparse.conf:
  update_lclogparse(cfg)

  return true
end

-- ---------------------------------------------------------------------------
-- uninstall(cfg, opts)
--
-- Remove LiveConfig configuration files from Postfix Server
-- ---------------------------------------------------------------------------
function uninstall(cfg, opts)
  LC.log.print(LC.log.INFO, "Disabling management of Postfix Server")

  local status, msg
  local configpath  = cfg["configpath"]
  local configfile  = cfg["configfile"]
  local masterfile  = cfg["configpath"] .. "/master.cf"
  local statusfile  = cfg["statusfile"]
  local domainsfile = cfg["domainsfile"]
  local aliasfile   = cfg["aliasfile"]

  if opts and opts.prefix then
    -- use prefix (for testing etc.)
    configpath  = opts.prefix .. configpath
    configfile  = opts.prefix .. configfile
    masterfile  = opts.prefix .. masterfile
    statusfile  = opts.prefix .. statusfile
    domainsfile = opts.prefix .. domainsfile
    aliasfile   = opts.prefix .. aliasfile
  end
  local sender_access_file = cfg["configpath"] .. "/sender_access"
  local recipient_access_file = cfg["configpath"] .. "/recipient_access"
  local greylist_file = cfg["configpath"] .. "/greylist_addrs"
  local sender_local_file = cfg["configpath"] .. "/sender_local"

  -- restore original "main.cf" (if existing...)
  if LC.fs.is_file(configfile .. ".lcbak") then
    LC.fs.rename(configfile .. ".lcbak", configfile)
  end
  -- restore original "master.cf" (if existing...)
  if LC.fs.is_file(masterfile .. ".lcbak") then
    LC.fs.rename(masterfile .. ".lcbak", masterfile)
  end
  -- restore original domainsfile (if existing...)
  if LC.fs.is_file(domainsfile .. ".lcbak") then
    LC.fs.rename(domainsfile .. ".lcbak", domainsfile)
  end
  -- restore original domainsfile database (if existing...)
  if LC.fs.is_file(domainsfile .. ".db.lcbak") then
    LC.fs.rename(domainsfile .. ".db.lcbak", domainsfile .. ".db")
  end
  -- restore original aliasfile (if existing...)
  if LC.fs.is_file(aliasfile .. ".lcbak") then
    LC.fs.rename(aliasfile .. ".lcbak", aliasfile)
  end
  -- restore original aliasfile database (if existing...)
  if LC.fs.is_file(aliasfile .. ".db.lcbak") then
    LC.fs.rename(aliasfile .. ".db.lcbak", aliasfile .. ".db")
  end

  -- Currently do not delete (because dependencies with dovecot)
  -- Delete user and group vmail
  -- local data  = { }
  -- data.user   = "vmail"
  -- data.group  = "vmail"
  -- data.home   = "vmail"
  -- data.shell  = "nologin"
  -- data.passwd = ""
  -- local uid = LC.sys.user_exists("vmail")
  -- if uid ~= false then 
  --   LC.users.delUser(data)
  -- end
  -- local gid = LC.sys.group_exists("vmail")
  -- if gid ~= false then
  --   LC.users.delGroup(data)
  -- end

  -- remove sender_access file
  if LC.fs.is_file(sender_access_file) then
    os.remove(sender_access_file)
  end
  if LC.fs.is_file(sender_access_file .. ".db") then
    os.remove(sender_access_file .. ".db")
  end

  -- remove recipient_access file
  if LC.fs.is_file(recipient_access_file) then
    os.remove(recipient_access_file)
  end
  if LC.fs.is_file(recipient_access_file .. ".db") then
    os.remove(recipient_access_file .. ".db")
  end

  -- remove greylist_addrs file
  if LC.fs.is_file(greylist_file) then
    os.remove(greylist_file)
  end
  if LC.fs.is_file(greylist_file .. ".db") then
    os.remove(greylist_file .. ".db")
  end

  -- remove sender_local_file file
  if LC.fs.is_file(sender_local_file) then
    os.remove(sender_local_file)
  end
  if LC.fs.is_file(sender_local_file .. ".db") then
    os.remove(sender_local_file .. ".db")
  end

  -- remove filter file
  if LC.fs.is_file(cfg["configpath"] .. "/cleanup_filter.regex") then
    os.remove(cfg["configpath"] .. "/cleanup_filter.regex")
  end

  -- remove status file
  if LC.fs.is_file(statusfile) then
    os.remove(statusfile)
  end

  -- restart postfix
  LC.exec(cfg["restart_cmd"])

  return true
end


-- ---------------------------------------------------------------------------
-- configure(cfg, opts)
--
-- Configure Postfix Server
-- ---------------------------------------------------------------------------
function configure(cfg, opts)

  local fh, status, res, msg
  local statusfile = cfg["statusfile"]
  local configpath = cfg["configpath"]
  local configfile = cfg["configfile"]
  local masterfile = cfg["configpath"] .. "/master.cf"
  local domainsfile = cfg["domainsfile"]
  local aliasfile   = cfg["aliasfile"]
  local v_major, v_minor = string.match(cfg.binversion, "^(%d+)%.(%d+)")
  v_major = tonumber(v_major)
  v_minor = tonumber(v_minor)

  if opts and opts.prefix then
    -- use prefix (for testing etc.)
    statusfile = opts.prefix .. statusfile
    configpath = opts.prefix .. configpath
    configfile = opts.prefix .. configfile
    masterfile = opts.prefix .. masterfile
    domainsfile = opts.prefix .. domainsfile
    aliasfile   = opts.prefix .. aliasfile
  end
  local sasl_access_file = cfg["configpath"] .. "/sasl_access"
  local sender_access_file = cfg["configpath"] .. "/sender_access"
  local recipient_access_file = cfg["configpath"] .. "/recipient_access"
  local greylist_file = cfg["configpath"] .. "/greylist_addrs"
  local sender_local_file = cfg["configpath"] .. "/sender_local"

  -- get uid and gid for 'mail' user
  local uid = LC.sys.user_exists("mail")
  if uid == false then
    return false, "System user 'mail' does not exist, please create user"
  end
  local gid = LC.sys.group_exists("mail")
  if gid == false then
    return false, "System group 'mail' does not exist, please create group"
  end

  -- configure SSL certificate
  res, msg = updateSSL(cfg, opts)
  if not res then return false, msg end

  if NOUPDATE then
    LC.log.print(LC.log.INFO, "Won't update Postfix configuration (postfix.NOUPDATE=true)")
  else
    -- create configuration file
    fh, msg = io.open(configfile .. ".tmp", "w")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", configfile, ".tmp' for writing: ", msg)
      return false, "Can't open '" .. configfile.. ".tmp' for writing: " .. msg
    end
    -- adjust owner & permissions - only Postfix needs to read this file:
    LC.fs.setperm(configfile .. ".tmp", 644, "root", "postfix")

    LC.liveconfig.writeHeader(fh)

    -- basic configuration
    local CONF = {
      ["smtpd_banner"]            = "$myhostname ESMTP",
      ["biff"]                    = "no",
      ["append_dot_mydomain"]     = "no",
      ["readme_directory"]        = "no",
      ["authorized_flush_users"]  = "root, $mail_owner",
      ["authorized_mailq_users"]  = "root, $mail_owner",
      ["disable_vrfy_command"]    = "yes",
    }

    if v_major >= 3 then
      CONF["compatibility_level"] = "2"
      -- disable SMTPUTF8, because all maps files currently use punycode (so all lookups would fail)
      CONF["smtputf8_enable"] = "no"
    end

    -- IP addresses
    local ipv4count = 0
    local ipv6count = 0
    if type(opts.ips) == "table" then
      -- bound to specific IPs
      local addrs = ""
      local i, s
      for i, s in ipairs(opts.ips) do
        if ipv4count + ipv6count > 0 then addrs = addrs .. ", " end
        addrs = addrs .. s
        if string.find(s, "%.") then
          ipv4count = ipv4count + 1
        else
          ipv6count = ipv6count + 1
        end
      end
      if ipv4count > 0 then
        addrs = addrs .. ", 127.0.0.1"
      end
      if ipv6count > 0 then
        addrs = addrs .. ", ::1"
      end
      CONF["inet_interfaces"] = addrs
    elseif type(opts.ips) == "nil" then
      -- bind only to localhost
      CONF["inet_interfaces"] = "127.0.0.1"
      ipv4count = 1
    elseif opts.ips == 0 then
      -- all protocols
      ipv4count = 1
      ipv6count = 1
    elseif opts.ips == 1 then
      -- all IPv4 interfaces
      ipv4count = 1
    elseif opts.ips == 2 then
      -- all IPv6 interfaces
      ipv6count = 1
    end

    if opts.out_ipv4 then
      ipv4count = 1
    end
    if opts.out_ipv6 and opts.out_ipv6 ~= "" then
      ipv6count = 1
    end

    if ipv4count > 0 and ipv6count == 0 then
      CONF["inet_protocols"] = "ipv4"
    elseif ipv4count == 0 and ipv6count > 0 then
      CONF["inet_protocols"] = "ipv6"
    else
      CONF["inet_protocols"] = "all"
    end

    -- outbound IPs
    if opts.out_ipv4 then
      CONF["smtp_bind_address"] = opts.out_ipv4
    end
    if opts.out_ipv6 and opts.out_ipv6 ~= "" then
      CONF["smtp_bind_address6"] = opts.out_ipv6
    end

    -- TLS parameters
    if cfg.dane and opts.dane then
      -- DANE/TLSA (outbound) available and enabled
      CONF["smtp_tls_security_level"] = "dane"
      CONF["smtp_dns_support_level"] = "dnssec"
    else
      CONF["smtp_tls_security_level"] = "may"
    end
    CONF["smtp_tls_session_cache_database"] = "btree:${data_directory}/smtp_scache"

    if v_major > 2 or (v_major == 2 and v_minor >= 9) then
      CONF["smtp_tls_loglevel"] = "1"
    end

    if (LC.distribution.family == "Debian" or LC.distribution.family == "Gentoo") and LC.fs.is_file("/etc/ssl/certs/ca-certificates.crt") then
      CONF["smtp_tls_CAfile"] = "/etc/ssl/certs/ca-certificates.crt"
    elseif LC.distribution.family == "SUSE" and LC.fs.is_file("/etc/ssl/ca-bundle.pem") then
      CONF["smtp_tls_CAfile"] = "/etc/ssl/ca-bundle.pem"
    elseif LC.distribution.family == "RedHat" and LC.fs.is_file("/etc/pki/tls/certs/ca-bundle.crt") then
      CONF["smtp_tls_CAfile"] = "/etc/pki/tls/certs/ca-bundle.crt"
    end

    if v_major > 2 or (v_major == 2 and v_minor >= 6) then
      CONF["smtp_tls_protocols"] = "!SSLv2, !SSLv3"
    end
    if v_major > 2 or (v_major == 2 and v_minor >= 5) then
      CONF["smtp_tls_mandatory_protocols"] = "!SSLv2, !SSLv3"
    else
      CONF["smtp_tls_mandatory_protocols"] = "TLSv1"
    end

    if opts.ssl then
      if LC.bits.band(opts.sslmode, 4) > 0 then
        -- force encrypted connections for user authentication
        CONF["smtpd_tls_auth_only"] = "yes"
      end

      CONF["smtpd_tls_security_level"] = "may"
      CONF["smtpd_tls_cert_file"] = cfg.crtfile
      CONF["smtpd_tls_key_file"] = cfg.keyfile
      CONF["smtpd_tls_session_cache_database"] = "btree:${data_directory}/smtpd_scache"

      if v_major > 2 or (v_major == 2 and v_minor >= 9) then
        CONF["smtpd_tls_loglevel"] = "1"
      end
      
      if (LC.distribution.family == "Debian" or LC.distribution.family == "Gentoo") and LC.fs.is_file("/etc/ssl/certs/ca-certificates.crt") then
        CONF["smtpd_tls_CAfile"] = "/etc/ssl/certs/ca-certificates.crt"
      elseif LC.distribution.family == "SUSE" and LC.fs.is_file("/etc/ssl/ca-bundle.pem") then
        CONF["smtpd_tls_CAfile"] = "/etc/ssl/ca-bundle.pem"
      elseif LC.distribution.family == "RedHat" and LC.fs.is_file("/etc/pki/tls/certs/ca-bundle.crt") then
        CONF["smtpd_tls_CAfile"] = "/etc/pki/tls/certs/ca-bundle.crt"
      end

      -- Diffie-Hellman parameters
      -- Postfix contains compiled-in default DH parameters since v2.2, so there's no need to create "own" DH parameters.
      -- However, this is considered to be more secure, so we do this. :)
      if not LC.fs.is_file(configpath .. '/dh512.pem') then
        -- generate 512 bit Diffie-Hellman parameters:
        local dh = LC.crypt.gen_dh(512)
        local dhf, msg = io.open(configpath .. '/dh512.pem', "w")
        if dhf == nil then
          LC.log.print(LC.log.ERR, "Can't open '", configpath .. "/dh512.pem' for writing: ", msg)
          return false, "Can't open '" .. configpath .. "/dh512.pem' for writing: " .. msg
        end
        LC.fs.setperm(configpath .. '/dh512.pem', "0640", "root", "postfix")
        LC.liveconfig.writeHeader(dhf)
        dhf:write("512 bit Diffie-Hellman parameters, created automatically by LiveConfig\n")
        dhf:write(dh)
        LC.liveconfig.writeFooter(dhf)
        dhf:close()
      end
      CONF["smtpd_tls_dh512_param_file"] = configpath .. "/dh512.pem"

      -- the 1024 bit DH parameters are only used for submission.
      -- Due to the incompatible Debian patch of GnuTLS we use 2048 parameters for communication with other SMTP servers
      -- (see http://marc.info/?l=postfix-users&m=137824308215202)
      if not LC.fs.is_file(configpath .. '/dh1024.pem') then
        -- generate 1024 bit Diffie-Hellman parameters:
        local dh = LC.crypt.gen_dh(1024)
        local dhf, msg = io.open(configpath .. '/dh1024.pem', "w")
        if dhf == nil then
          LC.log.print(LC.log.ERR, "Can't open '", configpath .. "/dh1024.pem' for writing: ", msg)
          return false, "Can't open '" .. configpath .. "/dh1024.pem' for writing: " .. msg
        end
        LC.fs.setperm(configpath .. '/dh1024.pem', "0640", "root", "postfix")
        LC.liveconfig.writeHeader(dhf)
        dhf:write("1024 bit Diffie-Hellman parameters, created automatically by LiveConfig\n")
        dhf:write(dh)
        LC.liveconfig.writeFooter(dhf)
        dhf:close()
      end

      if not LC.fs.is_file(configpath .. '/dh2048.pem') then
        -- generate 2048 bit Diffie-Hellman parameters:
        local dh = LC.crypt.gen_dh(2048)
        local dhf, msg = io.open(configpath .. '/dh2048.pem', "w")
        if dhf == nil then
          LC.log.print(LC.log.ERR, "Can't open '", configpath .. "/dh2048.pem' for writing: ", msg)
          return false, "Can't open '" .. configpath .. "/dh2048.pem' for writing: " .. msg
        end
        LC.fs.setperm(configpath .. '/dh2048.pem', "0640", "root", "postfix")
        LC.liveconfig.writeHeader(dhf)
        dhf:write("2048 bit Diffie-Hellman parameters, created automatically by LiveConfig\n")
        dhf:write(dh)
        LC.liveconfig.writeFooter(dhf)
        dhf:close()
      end
      CONF["smtpd_tls_dh1024_param_file"] = configpath .. "/dh2048.pem"

      if v_major == 2 and (v_minor == 6 or v_minor == 7) then
        -- only set for Postfix 2.6 and 2.7, defaults to "strong" since v2.8+, see http://www.postfix.org/postconf.5.html#smtpd_tls_eecdh_grade)
        CONF["smtpd_tls_eecdh_grade"] = "strong"
      end

      if v_major > 2 or (v_major == 2 and v_minor >= 6) then
        CONF["smtpd_tls_protocols"] = "!SSLv2, !SSLv3"
        CONF["smtpd_tls_ciphers"] = "medium"
        CONF["smtpd_tls_exclude_ciphers"] = "aNULL, eNULL, ADH"
      end
      if v_major > 2 or (v_major == 2 and v_minor >= 5) then
        CONF["smtpd_tls_mandatory_protocols"] = "!SSLv2, !SSLv3"
      else
        CONF["smtpd_tls_mandatory_protocols"] = "TLSv1"
      end
      if v_major > 2 or (v_major == 2 and v_minor >= 11) then
        CONF["tls_ssl_options"] = "NO_COMPRESSION"
      end
      if LC.bits.band(opts.sslmode, 8) > 0 then
        -- modern ciphers (actually, defending BEAST is not necessary on SMTP, but most dumb PCI scans don't care :(
        CONF["smtpd_tls_mandatory_ciphers"] = "high"
        CONF["tls_high_cipherlist"] = LC.liveconfig.DEFAULT_SSL_STRONG_CIPHERS
      else
        CONF["smtpd_tls_mandatory_ciphers"] = "medium"
        CONF["smtpd_tls_mandatory_exclude_ciphers"] = "aNULL, eNULL, ADH"
      end
    end

    -- 'aliases' file:
    CONF["alias_maps"] = "hash:/etc/aliases"
    CONF["alias_database"] = "hash:/etc/aliases"

    -- if /etc/aliases doesn't exist yet, create it:
    if not LC.fs.is_file("/etc/aliases") then
      local fha, msg = io.open("/etc/aliases", "w")
      if fha == nil then
        LC.log.print(LC.log.ERR, "Can't create '/etc/aliases' for writing: ", msg)
      else
        fha:write([[# /etc/aliases
mailer-daemon: postmaster
postmaster: root
nobody: root
hostmaster: root
usenet: root
news: root
webmaster: root
www: root
ftp: root
abuse: root
noc: root
security: root
]])
        fha:close()
        -- adjust permissions
        LC.fs.setperm("/etc/aliases", 644, "root", "root")
      end
    end
    -- check if hash file exists:
    if not LC.fs.is_file("/etc/aliases.db") then
      LC.exec("/usr/bin/newaliases")
    end

    local hostname = opts.hostname
    if hostname == nil then hostname = LC.sys.get_fqdn() end
    if hostname == nil then hostname = "localhost" end
    CONF["myhostname"] = hostname

    if LC.fs.is_file("/etc/mailname") then
      CONF["myorigin"] = "/etc/mailname"
    end

    CONF["mydestination"] = "localhost.localdomain, localhost, $myhostname"
    CONF["relayhost"] = ""
    CONF["mynetworks"] = "127.0.0.0/8 [::ffff:127.0.0.0]/104 [::1]/128"
    CONF["local_recipient_maps"] = "$alias_maps"

    if LC.distribution.family == "SUSE" then
      -- some OpenSUSE specific settings:
      CONF["setgid_group"] = "maildrop"
      CONF["queue_directory"] = "/var/spool/postfix"
      CONF["command_directory"] = "/usr/sbin"
      CONF["daemon_directory"] = "/usr/lib/postfix"
      CONF["data_directory"] = "/var/lib/postfix"
      CONF["mail_owner"] = "postfix"
      if LC.fs.is_dir("/usr/lib/postfix/bin") then
        -- OpenSUSE >=15
        CONF["daemon_directory"] = "/usr/lib/postfix/bin"
      end
    end

    CONF["virtual_mailbox_domains"] = "hash:/etc/postfix/virtual_domains"
    CONF["virtual_mailbox_base"] = "/var/mail"
    CONF["virtual_alias_maps"] = "hash:/etc/postfix/virtual_alias"
    CONF["virtual_mailbox_maps"] = "hash:/etc/postfix/virtual_alias"
    CONF["show_user_unknown_table_name"] = "no"

    CONF["virtual_minimum_uid"] = uid
    CONF["virtual_uid_maps"] = "static:" .. uid
    CONF["virtual_gid_maps"] = "static:" .. gid
    CONF["virtual_transport"] = "dovecot"
    CONF["dovecot_destination_recipient_limit"] = "1"
    CONF["mailbox_size_limit"] = "0"
    CONF["smtpd_sasl_auth_enable"] = "yes"
    CONF["smtpd_sasl_type"] = "dovecot"
    CONF["smtpd_sasl_path"] = "private/auth"
    CONF["smtpd_sasl_security_options"] = "noanonymous"
    CONF["smtpd_sasl_local_domain"] = "$mydomain"
    CONF["broken_sasl_auth_clients"] = "yes"
    CONF["smtpd_helo_required"] = "yes"
    CONF["smtpd_helo_restrictions"] = "reject_invalid_helo_hostname"
    -- workaround against SMTP smuggling, see https://www.postfix.org/smtp-smuggling.html
    CONF["smtpd_data_restrictions"] = "reject_unauth_pipelining"

    local smtpd_client_restrictions = "\n    permit_mynetworks"

    if v_major > 2 or (v_major == 2 and v_minor >= 11) then
      smtpd_client_restrictions = smtpd_client_restrictions .. ",\n    check_sasl_access hash:" .. sasl_access_file
    end

    smtpd_client_restrictions = smtpd_client_restrictions .. ",\n" .. [[
    permit_sasl_authenticated,
    reject_invalid_hostname,
    reject_unknown_reverse_client_hostname]]
    -- table with list of DNS Whitelists (Postfix 2.8+)
    if (v_major > 2 or (v_major == 2 and v_minor >= 8)) and type(opts.dnswl) == "table" then
      local i, s
      for i, s in ipairs(opts.dnswl) do
        smtpd_client_restrictions = smtpd_client_restrictions .. ",\n    permit_dnswl_client " .. s
      end
    end
    -- table with list of RBLs to check
    if type(opts.dnsbl) == "table" then
      local i, s
      for i, s in ipairs(opts.dnsbl) do
        smtpd_client_restrictions = smtpd_client_restrictions .. ",\n    reject_rbl_client " .. s
      end
    end
    CONF["smtpd_client_restrictions"] = smtpd_client_restrictions

    CONF["smtpd_sender_restrictions"] = [[
    permit_mynetworks,
    reject_unknown_address,
    reject_unknown_sender_domain,
    reject_non_fqdn_sender,
    check_sender_access hash:/etc/postfix/sender_access]]

    local smtpd_recipient_restrictions = "\n    reject_unknown_recipient_domain"
    smtpd_recipient_restrictions = smtpd_recipient_restrictions .. ",\n    check_recipient_access hash:" .. recipient_access_file

    if opts.limit then
      smtpd_recipient_restrictions = smtpd_recipient_restrictions .. ",\n    check_policy_service unix:private/lcpolicyd"
      if v_major > 2 or (v_major == 2 and v_minor >= 10) then
        CONF["authorized_submit_users"] = "socketmap:unix:/var/run/lcpolicyd-lookup.sock:user, static:anyone"
      end
      -- install/enable lcpolicyd
      lcpolicyd_enable(cfg)
    else
      -- uninstall/disable lcpolicyd
      lcpolicyd_disable(cfg)
    end

    smtpd_recipient_restrictions = smtpd_recipient_restrictions .. ",\n" .. [[
    permit_mynetworks,
    permit_sasl_authenticated,
    reject_unauth_destination,
    check_sender_access hash:]] .. sender_local_file
    -- Greylisting enabled?
    if opts.greylist then
      -- table with list of DNS Whitelists (Postfix 2.8+)
      if (v_major > 2 or (v_major == 2 and v_minor >= 8)) and type(opts.dnswl) == "table" then
        local i, s
        for i, s in ipairs(opts.dnswl) do
          smtpd_recipient_restrictions = smtpd_recipient_restrictions .. ",\n    permit_dnswl_client " .. s
        end
      end
      smtpd_recipient_restrictions = smtpd_recipient_restrictions .. ",\n    check_recipient_access hash:" .. greylist_file
    end
    CONF["smtpd_recipient_restrictions"] = smtpd_recipient_restrictions

    CONF["smtpd_discard_ehlo_keywords"] = "silent-discard, dsn"
    CONF["smtpd_recipient_limit"] = "250"

    if v_major > 2 or (v_major == 2 and v_minor >= 10) then
      CONF["smtpd_relay_restrictions"] = "\n" .. [[
    permit_mynetworks,
    permit_sasl_authenticated,
    reject_unauth_destination
]]
    end

    if opts.msgsize ~= nil then
      CONF["message_size_limit"] = opts.msgsize * 1024 * 1024
    else
      -- default maximum message size: 25 MB
      CONF["message_size_limit"] = "26214400"
    end

    -- check if antivirus software (via milter) is to be configured:
    if opts.antivirus or opts.spamfilter then
      local smtpd_milters = ""
      if opts.antivirus then
        smtpd_milters = smtpd_milters .. cfg["clamav_milter_socket"]
        -- adjust clamav-milter.conf
        LC.exec("sed -i -e 's/^#\\?OnInfected .*$/OnInfected Reject/i' " .. cfg["clamav_milter_cfg"])
        LC.exec("grep -q -i '^RejectMsg' " .. cfg["clamav_milter_cfg"] .. " || echo 'RejectMsg Rejecting harmful email: %v found.' >> " .. cfg["clamav_milter_cfg"])
        LC.exec("grep -q -i '^AddHeader' " .. cfg["clamav_milter_cfg"] .. " || echo 'AddHeader add' >> " .. cfg["clamav_milter_cfg"])
        if LC.distribution.family == "RedHat" or LC.distribution.family == "SUSE" then
          -- openSUSE/CentOS/RHEL: tune some ClamAV settings
          LC.exec("sed -i -e 's/^#\\?MilterSocketGroup virusgroup$/MilterSocketGroup postfix/i' -e 's/^#MilterSocketMode 660/MilterSocketMode 660/i' " .. cfg["clamav_milter_cfg"])
          if LC.fs.is_dir("/var/run/clamav-milter") then
            LC.exec("chmod o+x /var/run/clamav-milter")
          end
        end
        LC.exec("service clamav-milter restart")
      end

      if opts.spamfilter then
        local st, msg = LC.spamassassin.install()
        if not st then
          -- error while installing/configuring lcsam
          LC.log.print(LC.log.ERR, "Error while configuring lcsam: ", msg)
        else
          if opts.antivirus then
            smtpd_milters = smtpd_milters .. ", "
          end
          smtpd_milters = smtpd_milters .. cfg["lcsam_milter_socket"]
        end
      end

      CONF["smtpd_milters"] = smtpd_milters
    end

    if opts.spamfilter and v_major == 2 and (v_minor == 6 or v_minor == 7) then
      -- workaround against incompatible Milter protocol version with Postfix 2.3 - 2.7:
      CONF["milter_protocol"] = "2"
    end

    -- default action if a milter is not available: accept
    CONF["milter_default_action"] = "accept"

    -- check if DKIM is enabled
    if opts.dkim then
      local st, msg = LC.dkim.install()
      if not st then
        -- error while installing/configuring DKIM
        LC.log.print(LC.log.ERR, "Error while configuring DKIM", msg)
      else
        if CONF["smtpd_milters"] == nil then
          CONF["smtpd_milters"] = "unix:/var/spool/postfix/opendkim/opendkim.sock"
        else
          CONF["smtpd_milters"] = CONF["smtpd_milters"] .. ", unix:/var/spool/postfix/opendkim/opendkim.sock"
        end
        -- non_smtpd_milters run within a chroot jail! (/var/spool/postfix)
        if CONF["non_smtpd_milters"] == nil then
          CONF["non_smtpd_milters"] = "unix:/opendkim/opendkim.sock"
        else
          CONF["non_smtpd_milters"] = CONF["non_smtpd_milters"] .. ", unix:/opendkim/opendkim.sock"
        end
      end
    end

    -- check if greylisting is enabled:
    if opts.greylist then
      CONF["smtpd_restriction_classes"] = "greylist"
      if LC.distribution.family == "Debian" then
        CONF["greylist"] = "check_policy_service inet:127.0.0.1:10023"
      elseif LC.distribution.family == "SUSE" or LC.distribution.family == "RedHat" then
        CONF["greylist"] = "check_policy_service unix:/var/spool/postfix/postgrey/socket"
      elseif LC.distribution.family == "Gentoo" then
        CONF["greylist"] = "check_policy_service inet:127.0.0.1:10030"
      end

      if not LC.fs.is_file(greylist_file .. ".db") then
        if not LC.fs.is_file(greylist_file) then
          local fha, msg = io.open(greylist_file, "w")
          if fha == nil then
            LC.log.print(LC.log.ERR, "Can't create '" .. greylist_file .. "' for writing: ", msg)
          else
            LC.liveconfig.writeHeader(fha)
          end
          fha:close()
          -- adjust permissions
          LC.fs.setperm(greylist_file, 640, "root", "postfix")
        end

        -- update map file
        LC.exec("/usr/sbin/postmap " .. greylist_file)
      end
    end

    -- check if sub-addressing (VERP-style) is allowed:
    if opts.subaddr then
      CONF["recipient_delimiter"] = "+"
    else
      CONF["recipient_delimiter"] = ""
    end

    -- check if sender_bcc / recipient_bcc is enabled
    if opts.bccIn then
      local bcc_file = configpath .. "/recipient_bcc"
      CONF["recipient_bcc_maps"] = "hash:" .. bcc_file

      if not LC.fs.is_file(bcc_file .. ".db") then
        if not LC.fs.is_file(bcc_file) then
          local fha, msg = io.open(bcc_file, "w")
          if fha == nil then
            LC.log.print(LC.log.ERR, "Can't create '" .. bcc_file .. "' for writing: ", msg)
          else
            LC.liveconfig.writeHeader(fha)
            fha:close()
            -- adjust permissions
            LC.fs.setperm(bcc_file, 640, "root", "postfix")
          end
        end

        -- update map file
        LC.exec("/usr/sbin/postmap " .. bcc_file)
      end
    end

    if opts.bccOut then
      local bcc_file = configpath .. "/sender_bcc"
      CONF["sender_bcc_maps"] = "hash:" .. bcc_file

      if not LC.fs.is_file(bcc_file .. ".db") then
        if not LC.fs.is_file(bcc_file) then
          local fha, msg = io.open(bcc_file, "w")
          if fha == nil then
            LC.log.print(LC.log.ERR, "Can't create '" .. bcc_file .. "' for writing: ", msg)
          else
            LC.liveconfig.writeHeader(fha)
            fha:close()
            -- adjust permissions
            LC.fs.setperm(bcc_file, 640, "root", "postfix")
          end
        end

        -- update map file
        LC.exec("/usr/sbin/postmap " .. bcc_file)
      end
    end

    if not LC.fs.is_file(sender_local_file .. ".db") then
      if not LC.fs.is_file(sender_local_file) then
        local fha, msg = io.open(sender_local_file, "w")
        if fha == nil then
          LC.log.print(LC.log.ERR, "Can't create '" .. sender_local_file .. "' for writing: ", msg)
        else
          LC.liveconfig.writeHeader(fha)
        end
        fha:close()
        -- adjust permissions
        LC.fs.setperm(sender_local_file, 640, "root", "postfix")
      end

      -- update map file
      LC.exec("/usr/sbin/postmap " .. sender_local_file)
    end

    -- merge with local (override) configuration settings
    local i, j, s
    for i,s in pairs(LOCALCONFIG) do CONF[i]=s end

    -- sort configuration table (for human readers)
    local confkeys = {}
    for i in pairs(CONF) do table.insert(confkeys, i) end
    table.sort(confkeys)

    -- write configuration options:
    for i, s in ipairs(confkeys) do
      fh:write(s, " = ", CONF[s], "\n")
    end

    LC.liveconfig.writeFooter(fh)
    fh:close()

    -- enable configuration
    LC.fs.rename(configfile .. ".tmp", configfile)

    -- create "master.cf" configuration file
    fh, msg = io.open(masterfile .. ".tmp", "w")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", masterfile, ".tmp' for writing: ", msg)
      return false, "Can't open '" .. masterfile .. ".tmp' for writing: " .. msg
    end
    -- adjust owner & permissions - only Postfix needs to read this file:
    LC.fs.setperm(masterfile .. ".tmp", 640, "root", "postfix")
    LC.liveconfig.writeHeader(fh)

    -- service / type / private / unpriv / chroot / wakeup / maxproc / command / args
    local MASTER = {
      ["smtp"]        = { ["inet"] = { "n", "-", "n", "-", "-", "smtpd", {} },
                          ["unix"] = { "-", "-", "n", "-", "-", "smtp", {} }
                        },
      ["submission"]  = { ["inet"] = { "n", "-", "n", "-", "-", "smtpd", {
        "-o syslog_name=postfix/submission",
        "-o smtpd_sasl_auth_enable=yes",
        "-o smtpd_client_restrictions=permit_sasl_authenticated,reject",
        "-o smtpd_helo_required=no",
        "-o smtpd_helo_restrictions=",
      }}},
      ["pickup"]      = { ["fifo"] = { "n", "-", "y", "60", "1", "pickup" }},
      ["cleanup"]     = { ["unix"] = { "n", "-", "y", "-", "0", "cleanup" }},
      ["qmgr"]        = { ["fifo"] = { "n", "-", "n", "300", "1", "qmgr" }},
      ["tlsmgr"]      = { ["unix"] = { "-", "-", "y", "1000?", "1", "tlsmgr" }},
      ["rewrite"]     = { ["unix"] = { "-", "-", "y", "-", "-", "trivial-rewrite" }},
      ["bounce"]      = { ["unix"] = { "-", "-", "y", "-", "0", "bounce" }},
      ["defer"]       = { ["unix"] = { "-", "-", "y", "-", "0", "bounce" }},
      ["trace"]       = { ["unix"] = { "-", "-", "y", "-", "0", "bounce" }},
      ["verify"]      = { ["unix"] = { "-", "-", "y", "-", "1", "verify" }},
      ["flush"]       = { ["unix"] = { "n", "-", "y", "1000?", "0", "flush" }},
      ["proxymap"]    = { ["unix"] = { "-", "-", "n", "-", "-", "proxymap" }},
      ["proxywrite"]  = { ["unix"] = { "-", "-", "n", "-", "1", "proxymap" }},
      ["relay"]       = { ["unix"] = { "-", "-", "y", "-", "-", "smtp", {
        "-o smtp_fallback_relay="
      }}},
      ["showq"]       = { ["unix"] = { "n", "-", "y", "-", "-", "showq" }},
      ["error"]       = { ["unix"] = { "-", "-", "y", "-", "-", "error" }},
      ["retry"]       = { ["unix"] = { "-", "-", "y", "-", "-", "error" }},
      ["discard"]     = { ["unix"] = { "-", "-", "y", "-", "-", "discard" }},
      ["local"]       = { ["unix"] = { "-", "n", "n", "-", "-", "local" }},
      ["virtual"]     = { ["unix"] = { "-", "n", "n", "-", "-", "virtual" }},
      ["lmtp"]        = { ["unix"] = { "-", "-", "y", "-", "-", "lmtp" }},
      ["anvil"]       = { ["unix"] = { "-", "-", "y", "-", "1", "anvil" }},
      ["scache"]      = { ["unix"] = { "-", "-", "y", "-", "1", "scache" }},
      ["maildrop"]    = { ["unix"] = { "-", "n", "n", "-", "-", "pipe", {
        "flags=DROhu user=vmail argv=/usr/bin/maildrop -d ${recipient}"
      }}},
      ["dovecot"]     = { ["unix"] = { "-", "n", "n", "-", "-", "pipe", {} }}
    }

    if LC.fs.is_file(configpath .. '/dh1024.pem') then
      -- downgrade to 1024 bit DH parameters for submission:
      table.insert(MASTER["submission"]["inet"][7], "-o smtpd_tls_dh1024_param_file=" .. configpath .. "/dh1024.pem")
    end

    if (opts.hideIp) then
      table.insert(MASTER["submission"]["inet"][7], "-o cleanup_service_name=cleanupfilter")
      MASTER["cleanupfilter"] = { ["unix"] = { "n", "-", "y", "-", "0", "cleanup", {
        "-o header_checks=regexp:" .. configpath .. "/cleanup_filter.regex"
      } }}
      if not LC.fs.is_file(configpath .. "/cleanup_filter.regex") then
        local fh, msg = io.open(configpath .. "/cleanup_filter.regex.tmp", "w")
        if fh == nil then
          LC.log.print(LC.log.ERR, "Can't open '", configpath .. "/cleanup_filter.regex.tmp", "' for writing: ", msg)
          return false, "Can't open '" .. configpath .. "/cleanup_filter.regex.tmp" .. "' for writing: " .. msg
        end
        -- adjust owner & permissions - only Postfix needs to read this file:
        LC.fs.setperm(configpath .. "/cleanup_filter.regex.tmp", 640, "root", "postfix")
        fh:write("/^Received: from .*\\s+(by .*\\s+with ESMTPS?A.*)$/  REPLACE Received: ${1}\n")
        fh:close()
        LC.fs.rename(configpath .. "/cleanup_filter.regex.tmp", configpath .. "/cleanup_filter.regex")
      end
    end

    if opts.ssl and opts.sslmode ~= nil and LC.bits.band(opts.sslmode, 2) > 0 then
      -- enable (non-standard) SMTPS port
      -- we use the explicit port number here (465), because many distributions
      -- have already removed "smtps" from /etc/services
      MASTER["465"] = { ["inet"] = { "n", "-", "n", "-", "-", "smtpd", {
        "-o syslog_name=postfix/smtps",
        "-o smtpd_tls_wrappermode=yes",
        "-o smtpd_sasl_auth_enable=yes",
        "-o smtpd_client_restrictions=permit_sasl_authenticated,reject",
        "-o smtpd_helo_required=no",
        "-o smtpd_helo_restrictions="
      }}}

      if LC.fs.is_file(configpath .. '/dh1024.pem') then
        -- downgrade to 1024 bit DH parameters for submission:
        table.insert(MASTER["465"]["inet"][7], "-o smtpd_tls_dh1024_param_file=" .. configpath .. "/dh1024.pem")
      end

      if (opts.hideIp) then
        table.insert(MASTER["465"]["inet"][7], "-o cleanup_service_name=cleanupfilter")
      end

    end

    if opts.out_ipv6 and opts.out_ipv6 == "" then
      -- disable IPv6 for outbound connections
      table.insert(MASTER["smtp"]["unix"][7], "-o inet_protocols=ipv4")
      table.insert(MASTER["smtp"]["unix"][7], "-o inet_interfaces=all")
    end

    -- Dovecot: adjust location of "deliver" command
    local dovecot_deliver = "/usr/lib/dovecot/deliver"
    if LC.fs.is_file("/usr/libexec/dovecot/deliver") then
      dovecot_deliver = "/usr/libexec/dovecot/deliver"
    end

    -- check if sub-addressing (VERP-style) is allowed.
    -- null_sender is anyway set to empty address, as required by RFC 5321 (sec. 6.1)
    if opts.subaddr then
      table.insert(MASTER["dovecot"]["unix"][7], "flags=DROhu user=mail:mail null_sender= argv=" .. dovecot_deliver .. " -f ${sender} -a ${recipient} -d ${user}@${nexthop} -e")
    else
      table.insert(MASTER["dovecot"]["unix"][7], "flags=DROhu user=mail:mail null_sender= argv=" .. dovecot_deliver .. " -f ${sender} -d ${recipient} -e")
    end

    -- merge with local (override) configuration settings
    for i in pairs(LOCALMASTER) do
      if type(LOCALMASTER[i]) == "table" then
        -- add/update service
        for s in pairs(LOCALMASTER[i]) do
          if type(LOCALMASTER[i][s]) == "table" then
            -- add/update type
            if type(MASTER[i]) ~= "table" then MASTER[i] = {} end
            MASTER[i][s] = LOCALMASTER[i][s]
          else
            -- remove type
            if type(MASTER[i]) == "table" and type(MASTER[i][s]) == "table" then MASTER[i][s] = nil end
          end
        end
      else
        -- remove service
        if type(MASTER[i]) == "table" then MASTER[i] = nil end
      end
    end

    -- sort configuration table (for human readers)
    confkeys = {}
    for i in pairs(MASTER) do table.insert(confkeys, i) end
    table.sort(confkeys)

    -- write configuration options:
    fh:write([[

# ==========================================================================
# service  type private unpriv  chroot  wakeup  maxproc command + args
#               (yes)   (yes)   (yes)   (never) (100)
# ==========================================================================
]])
    for i, s in ipairs(confkeys) do
      for j in pairs(MASTER[s]) do
        fh:write(string.format("%-10s %-4s %-7s %-7s %-7s %-7s %-7s %s", s, j, MASTER[s][j][1], MASTER[s][j][2], MASTER[s][j][3], MASTER[s][j][4], MASTER[s][j][5], MASTER[s][j][6]), "\n")
        if type(MASTER[s][j][7]) == "table" then
          local args
          for args in pairs(MASTER[s][j][7]) do
            fh:write("  ", MASTER[s][j][7][args], "\n")
          end
        end
      end
    end

    LC.liveconfig.writeFooter(fh)
    fh:close()

    -- enable master.cf configuration
    LC.fs.rename(masterfile .. ".tmp", masterfile)
  end -- if not NOUPDATE

  -- create virtual_domains file (if not existing)
  if not LC.fs.is_file(domainsfile) then
    local fh, msg = io.open(domainsfile .. ".tmp", "w")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", domainsfile .. ".tmp", "' for writing: ", msg)
      return false, "Can't open '" .. domainsfile .. ".tmp" .. "' for writing: " .. msg
    end
    -- adjust owner & permissions - only Postfix needs to read this file:
    LC.fs.setperm(domainsfile .. ".tmp", 640, "root", "postfix")
    LC.liveconfig.writeHeader(fh)
    fh:close()
    LC.fs.rename(domainsfile .. ".tmp", domainsfile)
  end
  LC.exec("/usr/sbin/postmap " .. domainsfile)

  -- create virtual_alias file (if not existing)
  if not LC.fs.is_file(aliasfile) then
    local fh, msg = io.open(aliasfile .. ".tmp", "w")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", aliasfile .. ".tmp", "' for writing: ", msg)
      return false, "Can't open '" .. aliasfile .. ".tmp" .. "' for writing: " .. msg
    end
    -- adjust owner & permissions - only Postfix needs to read this file:
    LC.fs.setperm(aliasfile .. ".tmp", 640, "root", "postfix")
    LC.liveconfig.writeHeader(fh)
    fh:close()
    LC.fs.rename(aliasfile .. ".tmp", aliasfile)
  end
  LC.exec("/usr/sbin/postmap " .. aliasfile)

  -- create sasl_access file (if not existing)
  if not LC.fs.is_file(sasl_access_file) then
    local fh, msg = io.open(sasl_access_file .. ".tmp", "w")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", sasl_access_file .. ".tmp", "' for writing: ", msg)
      return false, "Can't open '" .. sasl_access_file .. ".tmp" .. "' for writing: " .. msg
    end
    -- adjust owner & permissions - only Postfix needs to read this file:
    LC.fs.setperm(sasl_access_file .. ".tmp", 640, "root", "postfix")
    LC.liveconfig.writeHeader(fh)
    fh:close()
    LC.fs.rename(sasl_access_file .. ".tmp", sasl_access_file)
  end
  LC.exec("/usr/sbin/postmap " .. sasl_access_file)

  -- create sender_access file (if not existing)
  if not LC.fs.is_file(sender_access_file) then
    local fh, msg = io.open(sender_access_file .. ".tmp", "w")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", sender_access_file .. ".tmp", "' for writing: ", msg)
      return false, "Can't open '" .. sender_access_file .. ".tmp" .. "' for writing: " .. msg
    end
    -- adjust owner & permissions - only Postfix needs to read this file:
    LC.fs.setperm(sender_access_file .. ".tmp", 640, "root", "postfix")
    LC.liveconfig.writeHeader(fh)
    fh:close()
    LC.fs.rename(sender_access_file .. ".tmp", sender_access_file)
  end
  LC.exec("/usr/sbin/postmap " .. sender_access_file)

  -- create recipient_access file (if not existing)
  if not LC.fs.is_file(recipient_access_file) then
    local fh, msg = io.open(recipient_access_file .. ".tmp", "w")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", recipient_access_file .. ".tmp", "' for writing: ", msg)
      return false, "Can't open '" .. recipient_access_file .. ".tmp" .. "' for writing: " .. msg
    end
    -- adjust owner & permissions - only Postfix needs to read this file:
    LC.fs.setperm(recipient_access_file .. ".tmp", 640, "root", "postfix")
    LC.liveconfig.writeHeader(fh)
    fh:close()
    LC.fs.rename(recipient_access_file .. ".tmp", recipient_access_file)
  end
  LC.exec("/usr/sbin/postmap " .. recipient_access_file)

  -- update status file
  LC.liveconfig.writeStatus(statusfile, "postfix", opts.revision, os.date("%Y-%m-%d %H:%M:%S %Z"))

  return true
end

-- ---------------------------------------------------------------------------
-- updateSSL(cfg, opts)
--
-- Update SSL certificate
-- ---------------------------------------------------------------------------
function updateSSL(cfg, opts)

  -- create/update SSL certificate file:
  local msg
  if opts.ssl then
    local sfh
    sfh, msg = io.open(cfg.crtfile, "w")
    LC.fs.setperm(cfg.crtfile, "0640", "root", "postfix")
    if sfh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", cfg.crtfile, "' for writing: ", msg)
      return false, "Can't open '" .. cfg.crtfile .. "' for writing: " .. msg
    end
    sfh:write(opts.ssl_crt)
    if opts.ssl_ca ~= nil then
      -- append chained certificate:
      sfh:write(opts.ssl_ca)
    end
    sfh:close()

    -- create/update SSL key file:
    sfh, msg = io.open(cfg.keyfile, "w")
    LC.fs.setperm(cfg.keyfile, "0640", "root", "postfix")
    if sfh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", cfg.keyfile, "' for writing: ", msg)
      return false, "Can't open '" .. cfg.keyfile .. "' for writing: " .. msg
    end
    sfh:write(opts.ssl_key)
    sfh:close()
  end

  return true

end

-- ---------------------------------------------------------------------------
-- convert an e-mail address into a search pattern:
local function search_addr(addr)
  s = "^" .. addr .. "%s"
  s = string.gsub(s, "%.", "%%.")
  s = string.gsub(s, "%-", "%%-")
  s = string.gsub(s, "%+", "%%+")
  s = string.gsub(s, "%*", "%%*")
  return s
end

-- ---------------------------------------------------------------------------
-- addMailbox
--
-- Add a mailbox to postfix server
-- ---------------------------------------------------------------------------
function addMailbox(cfg, opts, data)

  -- parking a mailbox: delete from Postfix configuration
  if data.parking then return deleteMailbox(cfg, opts, data) end

  local domainsfile = cfg["domainsfile"]
  local aliasfile   = cfg["aliasfile"]
  if opts and opts.prefix then
    domainsfile   = opts.prefix .. domainsfile
    aliasfile     = opts.prefix .. aliasfile
    configpath  = opts.prefix .. configpath
  end
  local greylist_file = cfg["configpath"] .. "/greylist_addrs"
  local name = data.name
  -- if a catch-all-address is configured, replace '*' with '':
  if name == '*' then name = '' end

  -- add domain to 'domains' file
  local fhr, fhw, msg
  fhr, msg = io.open(domainsfile, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", domainsfile, "' for reading: ", msg)
    return false, "Can't open '" .. domainsfile .. "' for reading: " .. msg
  end

  fhw, msg = io.open(domainsfile .. ".tmp", "w")
  if fhw == nil then
    LC.log.print(LC.log.ERR, "Can't open '", domainsfile .. ".tmp", "' for writing: ", msg)
    fhr:close()
    return false, "Can't open '" .. domainsfile .. ".tmp" .. "' for writing: " .. msg
  end
  -- adjust owner & permissions - only Postfix needs to read this file:
  LC.fs.setperm(domainsfile .. ".tmp", 640, "root", "postfix")

  local search = "^" .. data.domain
  search = string.gsub(search, "%.", "%%.")
  search = string.gsub(search, "%-", "%%-")

  local line
  local found = false
  while true do
    line = fhr:read()
    if line == nil then break end
    if string.find(line, search .. "$") ~= nil or string.find(line, search .. "%s") ~= nil then
      found = true
      break
    end
    fhw:write(line)
    fhw:write("\n")
  end

  fhr:close()

  if found == true then
    -- entry already in domains file:
    fhw:close()
    os.remove(domainsfile .. ".tmp")
  else
    -- entry not found, so add new one:
    fhw:write(data.domain, "\t", data.domain, "\n")
    fhw:close()
    LC.fs.rename(domainsfile .. ".tmp", domainsfile)
    LC.timeout.set('postfix.update_virtual_domains', 10, 60)
  end

  -- modify "virtual_alias" file:
  -- - add aliases
  -- - add forwards
  local fhr, fhw, msg
  fhr, msg = io.open(aliasfile, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", aliasfile, "' for reading: ", msg)
    return false, "Can't open '" .. aliasfile .. "' for reading: " .. msg
  end

  fhw, msg = io.open(aliasfile .. ".tmp", "w")
  if fhw == nil then
    LC.log.print(LC.log.ERR, "Can't open '", aliasfile .. ".tmp", "' for writing: ", msg)
    fhr:close()
    return false, "Can't open '" .. aliasfile .. ".tmp" .. "' for writing: " .. msg
  end
  -- adjust owner & permissions - only Postfix needs to read this file:
  LC.fs.setperm(aliasfile .. ".tmp", 640, "root", "postfix")

  -- create list with search patterns
  search = { }
  search[1] = search_addr(name .. "@" .. data.domain)
  local i, s
  
  -- add aliases to search list
  if type(data.aliases) == "table" then
    for i, s in ipairs(data.aliases) do
      search[#search + 1] = search_addr(s .. "@" .. data.domain)
    end
  end
  -- add aliases to be deleted to search list
  if type(data.delaliases) == "table" then
    for i, s in ipairs(data.delaliases) do
      search[#search + 1] = search_addr(s .. "@" .. data.domain)
    end
  end

  -- add "old" mailbox name to search list (if mailbox is renamed)
  local old_domain
  local old_domain_count = 0    -- number of mailboxes/addresses left with "old" domain
  if data.old_addr ~= nil then
    search[#search + 1] = search_addr(data.old_addr)
    local at_pos = data.old_addr:find("@")
    if at_pos ~= nil then
      old_domain = data.old_addr:sub(at_pos+1)
      -- add aliases to search list
      if type(data.aliases) == "table" then
        for i, s in ipairs(data.aliases) do
          search[#search + 1] = search_addr(s .. "@" .. old_domain)
        end
      end
      -- add aliases to be deleted to search list
      if type(data.delaliases) == "table" then
        for i, s in ipairs(data.delaliases) do
          search[#search + 1] = search_addr(s .. "@" .. old_domain)
        end
      end
    end
  end

  if old_domain ~= nil then
    if old_domain == data.domain then old_domain_count = old_domain_count + 1 end
    -- modify to use as search pattern:
    old_domain = string.gsub(old_domain, "%.", "%%.")
    old_domain = string.gsub(old_domain, "%-", "%%-")
    old_domain = "^%S*@" .. old_domain .. "%s"
  end

  while true do
    line = fhr:read()
    if line == nil then break end
--    LC.log.print(LC.log.DEBUG, "Line: '" .. line .. "'")
    found = false
    for i, s in ipairs(search) do
      if string.find(line, s) ~= nil then
        found = true
--        LC.log.print(LC.log.DEBUG, "FOUND: " .. line .. "'")
        break
      end
    end
    if found == false then
      -- only write line if not matched
      fhw:write(line, "\n")
      if old_domain ~= nil and line:find(old_domain) ~= nil then
        old_domain_count = old_domain_count + 1
      end
    end
  end

  fhr:close()

  -- if last mailbox with "old" domain got removed, then remove that domain
  -- also from virtual_domains file:
  if old_domain ~= nil and old_domain_count == 0 then
    if removeLine(domainsfile, data.old_addr:sub(data.old_addr:find("@") + 1)) then
      LC.timeout.set('postfix.update_virtual_domains', 10, 60)
    end
  end

  -- write forwards (if defined)
  if type(data.forwards) == "table" then
    fhw:write(name, "@", data.domain, "\t")
    if (data.mailbox == true or data.autoresponder == true) and name ~= '' then
      -- save copy to local mailbox (or deliver to Dovecot Sieve for autoresponder)
      fhw:write(name, "@", data.domain, ", ")
    end
    for i, s in ipairs(data.forwards) do
      if i > 1 then
        fhw:write(", ", s)
      else
        fhw:write(s)
      end
    end
    fhw:write("\n")
  elseif data.mailbox == true and name ~= '' then
    -- if no forwards are defined and this is a real mailbox, write also an alias
    -- entry pointing to itself (to allow local mailboxes while catch-all addresses may exist)
    fhw:write(name, "@", data.domain, "\t", name, "@", data.domain, "\n")
  end

  -- now add all aliases:
  if type(data.aliases) == "table" and name ~= '' then
    for i, s in ipairs(data.aliases) do
      fhw:write(s, "@", data.domain, "\t", name, "@", data.domain, "\n")
    end
  end

  fhw:close()
  LC.fs.rename(aliasfile .. ".tmp", aliasfile)

  -- update map file
  LC.timeout.set('postfix.update_virtual_alias', 10, 60)

  -- greylisting
  if LC.fs.is_file(greylist_file) then

    local replace

    -- if greylisting is enabled, add all addresses/aliases:
    if data.greylist then
      replace = name .. "@" .. data.domain .. "\tgreylist\n"
      -- now add all aliases:
      if type(data.aliases) == "table" and name ~= '' then
        for i, s in ipairs(data.aliases) do
          replace = replace .. s ..  "@" .. data.domain .. "\tgreylist\n"
        end
      end
    end

    -- update greylisting file
    if updateMap(greylist_file, search, replace) then
      LC.timeout.set('postfix.update_greylist_addrs', 10, 60)
    end

  end

  -- inbound BCC
  local bcc_file = cfg["configpath"] .. "/recipient_bcc"
  if LC.fs.is_file(bcc_file) then

    local replace

    -- if inbound BCC is enabled, add all addresses/aliases:
    if data.bccIn and data.bccIn ~= "" then
      replace = name .. "@" .. data.domain .. "\t" .. data.bccIn .. "\n"
      -- now add all aliases:
      if type(data.aliases) == "table" and name ~= '' then
        for i, s in ipairs(data.aliases) do
          replace = replace .. s ..  "@" .. data.domain .. "\t" .. data.bccIn .. "\n"
        end
      end
    end

    -- update inbound BCC map file
    if updateMap(bcc_file, search, replace) then
      LC.timeout.set('postfix.update_recipient_bcc', 10, 60)
    end
  end

  -- outbound BCC
  local bcc_file = cfg["configpath"] .. "/sender_bcc"
  if LC.fs.is_file(bcc_file) then

    local replace

    -- if outbound BCC is enabled, add all addresses/aliases:
    if data.bccOut and data.bccOut ~= ""  then
      replace = name .. "@" .. data.domain .. "\t" .. data.bccOut .. "\n"
      -- now add all aliases:
      if type(data.aliases) == "table" and name ~= '' then
        for i, s in ipairs(data.aliases) do
          replace = replace .. s ..  "@" .. data.domain .. "\t" .. data.bccOut .. "\n"
        end
      end
    end

    -- update outbound BCC map file
    if updateMap(bcc_file, search, replace) then
      LC.timeout.set('postfix.update_sender_bcc', 10, 60)
    end
  end

  -- spam filtering
  -- create list with all addresses to add/update:
  local spamaddrs = { }
  if type(data.spamfilter) == "table" then
    spamaddrs[1] = name .. "@" .. data.domain
    data.spamfilter.userdir = data.contract .. "_" .. data.id
    -- add aliases
    if type(data.aliases) == "table" then
      for i, s in ipairs(data.aliases) do
        spamaddrs[#spamaddrs + 1] = s .. "@" .. data.domain
      end
    end
  else
    -- no spam filter enabled.
    -- However, set userdir to delete possibly existing configuration file
    data.spamfilter = {
      ["userdir"] = data.contract .. "_" .. data.id
    }
  end

  LC.spamassassin.editMailbox(search, spamaddrs, data.spamfilter)

  return true

end

-- ---------------------------------------------------------------------------
-- editMailbox
--
-- Edit mailbox configuration
-- ---------------------------------------------------------------------------
function editMailbox(cfg, opts, data)
  -- same as addMailbox():
  return addMailbox(cfg, opts, data)
end

-- ---------------------------------------------------------------------------
-- deleteMailbox
--
-- Delete mailbox
-- ---------------------------------------------------------------------------
function deleteMailbox(cfg, opts, data)

  local domainsfile = cfg["domainsfile"]
  local aliasfile   = cfg["aliasfile"]
  local configpath  = cfg["configpath"]
  if opts and opts.prefix then
    domainsfile = opts.prefix .. domainsfile
    aliasfile   = opts.prefix .. aliasfile
    configpath  = opts.prefix .. configpath
  end
  local sender_access_file = cfg["configpath"] .. "/sender_access"
  local recipient_access_file = cfg["configpath"] .. "/recipient_access"
  local greylist_file = cfg["configpath"] .. "/greylist_addrs"

  -- if a catch-all-address is configured, replace '*' with '':
  local name = data.name
  if name == '*' then name = '' end

  -- modify "virtual_alias" file:
  -- - remove aliases
  -- - remove forwards
  local fhr, fhw, msg, line
  fhr, msg = io.open(aliasfile, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", aliasfile, "' for reading: ", msg)
    return false, "Can't open '" .. aliasfile .. "' for reading: " .. msg
  end

  fhw, msg = io.open(aliasfile .. ".tmp", "w")
  if fhw == nil then
    LC.log.print(LC.log.ERR, "Can't open '", aliasfile .. ".tmp", "' for writing: ", msg)
    fhr:close()
    return false, "Can't open '" .. aliasfile .. ".tmp" .. "' for writing: " .. msg
  end
  -- adjust owner & permissions - only Postfix needs to read this file:
  LC.fs.setperm(aliasfile .. ".tmp", 640, "root", "postfix")

  -- create list with search patterns
  search = { }
  search[1] = search_addr(name .. "@" .. data.domain)
  local i, s
  
  -- add aliases to be deleted to search list
  if type(data.delaliases) == "table" then
    for i, s in ipairs(data.delaliases) do
      search[#search + 1] = search_addr(s .. "@" .. data.domain)
    end
  end

  -- count how many mailboxes with this domain are still configured.
  -- if "last" mailbox is deleted, remove also from virtual_domains file:
  local domaincount = 0
  local found = false
  local ds = data.domain
  ds = string.gsub(ds, "%.", "%%.")
  ds = string.gsub(ds, "%-", "%%-")
  ds = "^%S*@" .. ds .. "%s"

  while true do
    line = fhr:read()
    if line == nil then break end
--    LC.log.print(LC.log.DEBUG, "Line: '" .. line .. "'")
    found = false
    for i, s in ipairs(search) do
      if string.find(line, s) ~= nil then
        found = true
--        LC.log.print(LC.log.DEBUG, "FOUND: " .. line .. "'")
        break
      end
    end
    if found == false then
      if string.find(line, ds) ~= nil then
        domaincount = domaincount + 1
      end
      -- only write line if not matched
      fhw:write(line, "\n")
    end
  end

  fhr:close()

  fhw:close()
  LC.fs.rename(aliasfile .. ".tmp", aliasfile)

  -- update map file
  LC.timeout.set('postfix.update_virtual_alias', 10, 60)

  if domaincount == 0 then
    -- deleted last mailbox with this domain, so also remove domain from virtual_aliases, sender_access and recipient_access
    if removeLine(domainsfile, data.domain) then
      LC.timeout.set('postfix.update_virtual_domains', 10, 60)
    end
    if LC.fs.is_file(sender_access_file) and removeLine(sender_access_file, '@' .. data.domain) then
      LC.timeout.set('postfix.update_sender_access', 10, 60)
    end
    if LC.fs.is_file(recipient_access_file) and removeLine(recipient_access_file, data.domain) then
      LC.timeout.set('postfix.update_recipient_access', 10, 60)
    end
  end

  if LC.fs.is_file(greylist_file) and updateMap(greylist_file, search) then
    LC.timeout.set('postfix.update_greylist_addrs', 10, 60)
  end

  local recipient_bcc_file = cfg["configpath"] .. "/recipient_bcc"
  if LC.fs.is_file(recipient_bcc_file) and updateMap(recipient_bcc_file, search) then
    LC.timeout.set('postfix.update_recipient_bcc', 10, 60)
  end

  local sender_bcc_file = cfg["configpath"] .. "/sender_bcc"
  if LC.fs.is_file(sender_bcc_file) and updateMap(sender_bcc_file, search) then
    LC.timeout.set('postfix.update_sender_bcc', 10, 60)
  end

  LC.spamassassin.deleteMailbox(search, data.contract .. "_" .. data.id)

  return true

end

-- ---------------------------------------------------------------------------
-- update_virtual_alias()
--
-- Update "virtual_alias" map file
-- ---------------------------------------------------------------------------
function update_virtual_alias()
  LC.log.print(LC.log.DEBUG, "postfix.update_virtual_alias() called")
  -- get configuration
  local cfg = LC.smtp.getConfig('postfix')
  if cfg == nil then
    LC.log.print(LC.log.ERR, "postfix.update_virtual_alias(): no configuration for 'postfix' available!?")
    return
  end
  -- get filename
  if cfg.aliasfile == nil then
    LC.log.print(LC.log.ERR, "postfix.update_virtual_alias(): no aliasfile for 'postfix' available!?")
    return
  end

  -- update map file
  LC.mutex.lock(cfg.aliasfile)
  LC.exec("/usr/sbin/postmap " .. cfg.aliasfile)
  LC.mutex.unlock(cfg.aliasfile)

end

-- ---------------------------------------------------------------------------
-- update_virtual_domains()
--
-- Update "virtual_domains" map file
-- ---------------------------------------------------------------------------
function update_virtual_domains()
  LC.log.print(LC.log.DEBUG, "postfix.update_virtual_domains() called")
  -- get configuration
  local cfg = LC.smtp.getConfig('postfix')
  if cfg == nil then
    LC.log.print(LC.log.ERR, "postfix.update_virtual_domains(): no configuration for 'postfix' available!?")
    return
  end
  -- get filename
  if cfg.domainsfile == nil then
    LC.log.print(LC.log.ERR, "postfix.update_virtual_domains(): no domainsfile for 'postfix' available!?")
    return
  end

  -- update map file
  LC.mutex.lock(cfg.domainsfile)
  LC.exec("/usr/sbin/postmap " .. cfg.domainsfile)
  LC.mutex.unlock(cfg.domainsfile)

end

-- ---------------------------------------------------------------------------
-- update_sender_access()
--
-- Update "sender_access" map file
-- ---------------------------------------------------------------------------
function update_sender_access()
  LC.log.print(LC.log.DEBUG, "postfix.update_sender_access() called")
  -- get configuration
  local cfg = LC.smtp.getConfig('postfix')
  if cfg == nil then
    LC.log.print(LC.log.ERR, "postfix.update_sender_access(): no configuration for 'postfix' available!?")
    return
  end

  -- update map file
  LC.mutex.lock(cfg.configpath .. "/sender_access")
  LC.exec("/usr/sbin/postmap " .. cfg.configpath .. "/sender_access")
  LC.mutex.unlock(cfg.configpath .. "/sender_access")

end

-- ---------------------------------------------------------------------------
-- update_recipient_access()
--
-- Update "recipient_access" map file
-- ---------------------------------------------------------------------------
function update_recipient_access()
  LC.log.print(LC.log.DEBUG, "postfix.update_recipient_access() called")
  -- get configuration
  local cfg = LC.smtp.getConfig('postfix')
  if cfg == nil then
    LC.log.print(LC.log.ERR, "postfix.update_recipient_access(): no configuration for 'postfix' available!?")
    return
  end

  -- update map file
  LC.mutex.lock(cfg.configpath .. "/recipient_access")
  LC.exec("/usr/sbin/postmap " .. cfg.configpath .. "/recipient_access")
  LC.mutex.unlock(cfg.configpath .. "/recipient_access")

end

-- ---------------------------------------------------------------------------
-- update_greylist_addrs()
--
-- Update "greylist_addrs" map file
-- ---------------------------------------------------------------------------
function update_greylist_addrs()
  LC.log.print(LC.log.DEBUG, "postfix.update_greylist_addrs() called")
  -- get configuration
  local cfg = LC.smtp.getConfig('postfix')
  if cfg == nil then
    LC.log.print(LC.log.ERR, "postfix.update_greylist_addrs(): no configuration for 'postfix' available!?")
    return
  end

  -- update map file
  LC.mutex.lock(cfg.configpath .. "/greylist_addrs")
  LC.exec("/usr/sbin/postmap " .. cfg.configpath .. "/greylist_addrs")
  LC.mutex.unlock(cfg.configpath .. "/greylist_addrs")

end

-- ---------------------------------------------------------------------------
-- update_recipient_bcc()
--
-- Update "recipient_bcc_maps" map file
-- ---------------------------------------------------------------------------
function update_recipient_bcc()
  LC.log.print(LC.log.DEBUG, "postfix.update_recipient_bcc() called")
  -- get configuration
  local cfg = LC.smtp.getConfig('postfix')
  if cfg == nil then
    LC.log.print(LC.log.ERR, "postfix.update_recipient_bcc(): no configuration for 'postfix' available!?")
    return
  end

  -- update map file
  LC.mutex.lock(cfg.configpath .. "/recipient_bcc")
  LC.exec("/usr/sbin/postmap " .. cfg.configpath .. "/recipient_bcc")
  LC.mutex.unlock(cfg.configpath .. "/recipient_bcc")

end

-- ---------------------------------------------------------------------------
-- update_sender_bcc()
--
-- Update "sender_bcc_maps" map file
-- ---------------------------------------------------------------------------
function update_sender_bcc()
  LC.log.print(LC.log.DEBUG, "postfix.update_sender_bcc() called")
  -- get configuration
  local cfg = LC.smtp.getConfig('postfix')
  if cfg == nil then
    LC.log.print(LC.log.ERR, "postfix.update_sender_bcc(): no configuration for 'postfix' available!?")
    return
  end

  -- update map file
  LC.mutex.lock(cfg.configpath .. "/sender_bcc")
  LC.exec("/usr/sbin/postmap " .. cfg.configpath .. "/sender_bcc")
  LC.mutex.unlock(cfg.configpath .. "/sender_bcc")

end

-- ---------------------------------------------------------------------------
-- lockDomains()
--
-- Lock domains (forbid sending of e-mails)
-- ---------------------------------------------------------------------------
function lockDomains(domains)
  LC.log.print(LC.log.DEBUG, "postfix.lockDomains() called")
  -- get configuration
  local cfg = LC.smtp.getConfig('postfix')
  if cfg == nil then
    LC.log.print(LC.log.ERR, "postfix.update_sender_access(): no configuration for 'postfix' available!?")
    return
  end

  local filename = cfg.configpath .. "/sender_access"

  local fhr, fhw, msg, line
  fhr, msg = io.open(filename, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", filename, "' for reading: ", msg)
    return false, "Can't open '" .. filename .. "' for reading: " .. msg
  end

  fhw, msg = io.open(filename .. ".tmp", "w")
  if fhw == nil then
    LC.log.print(LC.log.ERR, "Can't open '", filename .. ".tmp", "' for writing: ", msg)
    fhr:close()
    return false, "Can't open '" .. filename .. ".tmp" .. "' for writing: " .. msg
  end
  -- adjust owner & permissions - only Postfix needs to read this file:
  LC.fs.setperm(filename .. ".tmp", 640, "root", "postfix")

  -- create list with search patterns
  local search = { }
  local i, s, p
  for i, s in ipairs(domains) do
    p = "^@" .. s .. "%s"
    p = string.gsub(p, "%.", "%%.")
    p = string.gsub(p, "%-", "%%-")
    search[s] = p
  end

  while true do
    line = fhr:read()
    if line == nil then break end
    for i, s in pairs(search) do
      if string.find(line, s) ~= nil then
        -- domain found, so remove pattern from list
        search[i] = nil
        break
      end
    end
    fhw:write(line, "\n")
  end

  fhr:close()

  -- append all domains left in list to the file:
  local modified = false
  for i, s in pairs(search) do
    fhw:write("@", i, "\t500 No outbound mails allowed - please contact your administrator.\n")
    modified = true
  end

  fhw:close()

  if modified then
    LC.fs.rename(filename .. ".tmp", filename)
    -- update map file
    LC.timeout.set('postfix.update_sender_access', 10, 60)
  end

  return true

end

-- ---------------------------------------------------------------------------
-- update_lclogparse
--
-- Create/update lclogparse configuration
-- ---------------------------------------------------------------------------
function update_lclogparse(cfg)
  if cfg == nil or cfg.logfile == nil or (string.sub(cfg.logfile, 1, 9) ~= 'journald:' and not LC.fs.is_file(cfg.logfile)) then return end
  local rules
  if string.sub(cfg.logfile, 1, 9) ~= 'journald:' then
    rules = [=[	# Postfix
	NEW:	-I-f	^.* postfix/([^/]+/)?smtp.*: ([A-Za-z0-9]+): (client=[^,]+), sasl_method=[^,]+, sasl_username=([^ ]+)
	NEW:	-IE	^.* postfix/([^/]+/)?smtp.*: ([A-Za-z0-9]+): (client=[^,]+)$
	UPDATE:	IX	^.* postfix/.*: ([A-Za-z0-9]+): from=(<>)
	UPDATE:	IS	^.* postfix/.*: ([A-Za-z0-9]+): from=<[^>]+>, size=(\d+)
	UPDATE:	IO---	^.* postfix/.*: ([A-Za-z0-9]+): (to)=<([^>]+)>, relay=(?!dovecot).*, status=(sent|bounced)
	UPDATE:	IOt--	^.* postfix/.*: ([A-Za-z0-9]+): (to)=<[^>]+>, orig_to=<([^>]+)>, relay=(?!dovecot).*, status=(sent|bounced)
	UPDATE:	IOt	^.* postfix/.*: ([A-Za-z0-9]+): (to)=<([^>]+)>, relay=dovecot
	UPDATE:	IOt	^.* postfix/.*: ([A-Za-z0-9]+): (to)=<[^>]+>, orig_to=<([^>]+)>, relay=dovecot
	DONE:	I	^.* postfix/.*: ([A-Za-z0-9]+): removed
	OUTPUT-FILE:	]=] .. LC.liveconfig.localstatedir .. [=[/lib/liveconfig/smtp.stats
]=]
  else
    -- journald has all log data without timestamp and service name
    rules = [=[ # Postfix
	NEW:	I-f	^([A-Za-z0-9]+): (client=[^,]+), sasl_method=[^,]+, sasl_username=([^ ]+)
	NEW:	IE	^([A-Za-z0-9]+): (client=[^,]+)$
	UPDATE:	IX	^([A-Za-z0-9]+): from=(<>)
	UPDATE:	IS	^([A-Za-z0-9]+): from=<[^>]+>, size=(\d+)
	UPDATE:	IO---	^([A-Za-z0-9]+): (to)=<([^>]+)>, relay=(?!dovecot).*, status=(sent|bounced)
	UPDATE:	IOt--	^([A-Za-z0-9]+): (to)=<[^>]+>, orig_to=<([^>]+)>, relay=(?!dovecot).*, status=(sent|bounced)
	UPDATE:	IOt	^([A-Za-z0-9]+): (to)=<([^>]+)>, relay=dovecot
	UPDATE:	IOt	^([A-Za-z0-9]+): (to)=<[^>]+>, orig_to=<([^>]+)>, relay=dovecot
	DONE:	I	^([A-Za-z0-9]+): removed
	OUTPUT-FILE:	]=] .. LC.liveconfig.localstatedir .. [=[/lib/liveconfig/smtp.stats
]=]
  end
  LC.liveconfig.update_lclogparse(cfg.logfile, rules)
end

-- ---------------------------------------------------------------------------
-- lcpolicyd_enable
-- ---------------------------------------------------------------------------
function lcpolicyd_enable(cfg)
  -- install service (systemd)
  if LC.fs.is_dir("/lib/systemd/system") and not LC.fs.is_file("/lib/systemd/system/lcpolicyd.service") then
    LC.exec("cp --preserve=all " .. LC.liveconfig.libexecdir .. "/lcpolicyd.service /lib/systemd/system/")
    if LC.fs.is_file("/bin/systemctl") then
      LC.exec("/bin/systemctl enable lcpolicyd")
    end
  end

  -- install init script
  if not LC.fs.is_file("/etc/init.d/lcpolicyd") then
    LC.exec("cp " .. LC.liveconfig.libexecdir .. "/lcpolicyd.init /etc/init.d/lcpolicyd")
    LC.fs.setperm("/etc/init.d/lcpolicyd", "0700", "root", "root")

    if LC.fs.is_file("/usr/sbin/update-rc.d") then
      LC.exec("/usr/sbin/update-rc.d lcpolicyd defaults")
    elseif LC.fs.is_file("/sbin/chkconfig") then
      LC.exec("/sbin/chkconfig --add lcpolicyd")
    elseif LC.fs.is_file("/sbin/rc-update") then
      LC.exec("/sbin/rc-update add lcpolicyd default")
    else
      LC.log.print(LC.log.ERR, "Can't install init script 'lcpolicyd' - no supported tool found")
    end

  end

  -- start lcpolicyd if not running
  if LC.fs.is_file("/sbin/service") then
    LC.exec("/sbin/service lcpolicyd start")
  else
    LC.exec("/etc/init.d/lcpolicyd start")
  end
end

-- ---------------------------------------------------------------------------
-- lcpolicyd_disable
-- ---------------------------------------------------------------------------
function lcpolicyd_disable(cfg)

  -- disable lcpolicyd (chkconfig)
  if LC.fs.is_file("/etc/init.d/lcpolicyd") then

    -- quit lcpolicyd
    if LC.fs.is_file("/sbin/service") then
      LC.exec("/sbin/service lcpolicyd stop")
    else
      LC.exec("/etc/init.d/lcpolicyd stop")
    end

    if LC.fs.is_file("/lib/systemd/system/lcpolicyd") then
      -- disable service (systemd)
      if LC.fs.is_file("/bin/systemctl") then
        LC.exec("/bin/systemctl disable lcpolicyd")
      end
      -- remove service file
      os.remove("/lib/systemd/system/lcpolicyd")
    end

    if LC.fs.is_file("/usr/sbin/update-rc.d") then
      LC.exec("/usr/sbin/update-rc.d -f lcpolicyd remove")
    elseif LC.fs.is_file("/sbin/chkconfig") then
      LC.exec("/sbin/chkconfig --del lcpolicyd")
    elseif LC.fs.is_file("/sbin/rc-update") then
      LC.exec("/sbin/rc-update del lcpolicyd default")
    else
      LC.log.print(LC.log.ERR, "Can't uninstall init script 'lcpolicyd' - no supported tool found")
    end
    os.remove("/etc/init.d/lcpolicyd")
  end
end

-- ---------------------------------------------------------------------------
-- editlock()
-- ---------------------------------------------------------------------------
function editlock(cfg, data)

  -- lock mode = 0: unlock
  --             1: lock (allow receiving, but not sending)
  --             2: disable (neither receiving nor sending allowed)

  local fhr, fhw, msg, line

  if data.accounts ~= nil then
    -- edit /etc/postfix/sasl_access (add/remove all SASL accounts)
    local sasl_access_file = cfg["configpath"] .. "/sasl_access"
    fhr, msg = io.open(sasl_access_file, "r")
    if fhr == nil then
      LC.log.print(LC.log.ERR, "Can't open '", sasl_access_file, "' for reading: ", msg)
      return false, "Can't open '" .. sasl_access_file.. "' for reading: " .. msg
    end

    fhw, msg = io.open(sasl_access_file .. ".tmp", "w")
    if fhw == nil then
      fhr:close()
      LC.log.print(LC.log.ERR, "Can't open '", sasl_access_file, ".tmp' for writing: ", msg)
      return false, "Can't open '" .. sasl_access_file .. ".tmp' for writing: " .. msg
    end
    -- adjust owner & permissions - only Postfix needs to read this file:
    LC.fs.setperm(sasl_access_file .. ".tmp", 640, "root", "postfix")

    while true do
      line = fhr:read()
      if line == nil then break end
      local ws = line:find('%s')
      if ws == nil or not LC.liveconfig.inTable(line:sub(1, ws-1), data.accounts) then
        fhw:write(line, "\n")
      end
    end

    fhr:close()

    if data.mode > 0 then
      -- add accounts to "sasl_access" table to forbid sending e-mails
      for _,line in pairs(data.accounts) do
        fhw:write(line, " REJECT ", TXT_SASL_REJECT, "\n")
      end
    end

    fhw:close()

    LC.fs.rename(sasl_access_file .. ".tmp", sasl_access_file)

    -- update map file
    LC.exec("/usr/sbin/postmap " .. sasl_access_file)
  end

  if data.domains ~= nil then
    -- edit /etc/postfix/recipient_access (add/remove all domains)
    local recipient_access_file = cfg["configpath"] .. "/recipient_access"
    fhr, msg = io.open(recipient_access_file, "r")
    if fhr == nil then
      LC.log.print(LC.log.ERR, "Can't open '", recipient_access_file, "' for reading: ", msg)
      return false, "Can't open '" .. recipient_access_file.. "' for reading: " .. msg
    end

    fhw, msg = io.open(recipient_access_file .. ".tmp", "w")
    if fhw == nil then
      fhr:close()
      LC.log.print(LC.log.ERR, "Can't open '", recipient_access_file, ".tmp' for writing: ", msg)
      return false, "Can't open '" .. recipient_access_file .. ".tmp' for writing: " .. msg
    end
    -- adjust owner & permissions - only Postfix needs to read this file:
    LC.fs.setperm(recipient_access_file .. ".tmp", 640, "root", "postfix")

    while true do
      line = fhr:read()
      if line == nil then break end
      local ws = line:find('%s')
      if ws == nil or not LC.liveconfig.inTable(line:sub(1, ws-1), data.domains) then
        LC.log.print(LC.log.DEBUG, "LINE not found: '", line:sub(1, ws-1), "'")
        fhw:write(line, "\n")
      end
    end

    fhr:close()

    if data.mode == 2 then
      -- add domains to "recipient_access" table to reject e-mails:
      for _,line in pairs(data.domains) do
        fhw:write(line, " 521 ", TXT_DOMAIN_DISABLED, "\n")
      end
    end

    fhw:close()

    LC.fs.rename(recipient_access_file .. ".tmp", recipient_access_file)

    -- update map file
    LC.exec("/usr/sbin/postmap " .. recipient_access_file)

  end

  return true

end

-- ---------------------------------------------------------------------------
-- updateDomain()
-- ---------------------------------------------------------------------------
function updateDomain(cfg, data)

  local fhr, fhw, msg, line

  if data.addNoExt ~= nil or data.delNoExt ~= nil then
    -- edit /etc/postfix/sender_local (add/remove domains)
    local senderfile = cfg["configpath"] .. "/sender_local"

    if not LC.fs.is_file(senderfile) then
      fhw, msg = io.open(senderfile, "w")
      if fhw == nil then
        LC.log.print(LC.log.ERR, "Can't open '", senderfile, "' for writing: ", msg)
        return false, "Can't open '" .. senderfile .. "' for writing: " .. msg
      end
      -- adjust owner & permissions - only Postfix needs to read this file:
      LC.fs.setperm(senderfile, 640, "root", "postfix")
      fhw:close()
    end

    fhr, msg = io.open(senderfile, "r")
    if fhr == nil then
      LC.log.print(LC.log.ERR, "Can't open '", senderfile, "' for reading: ", msg)
      return false, "Can't open '" .. senderfile.. "' for reading: " .. msg
    end

    fhw, msg = io.open(senderfile .. ".tmp", "w")
    if fhw == nil then
      fhr:close()
      LC.log.print(LC.log.ERR, "Can't open '", senderfile, ".tmp' for writing: ", msg)
      return false, "Can't open '" .. senderfile .. ".tmp' for writing: " .. msg
    end
    -- adjust owner & permissions - only Postfix needs to read this file:
    LC.fs.setperm(senderfile .. ".tmp", 640, "root", "postfix")

    while true do
      line = fhr:read()
      if line == nil then break end
      local ws = line:find('%s')
      if ws == nil or data.delNoExt == nil or not LC.liveconfig.inTable(line:sub(1, ws-1), data.delNoExt) then
        fhw:write(line, "\n")
      end
    end

    fhr:close()

    if data.addNoExt then
      -- add domains to "sender_local" table to reject e-mails:
      for _,line in pairs(data.addNoExt) do
        fhw:write(line, " 521 ", TXT_DOMAIN_EXT_REJECTED, "\n")
      end
    end

    fhw:close()

    LC.fs.rename(senderfile .. ".tmp", senderfile)

    -- update map file
    LC.exec("/usr/sbin/postmap " .. senderfile)

  end

  return true

end

-- <EOF>----------------------------------------------------------------------
