--  _    _          ___           __ _     (R)
-- | |  (_)_ _____ / __|___ _ _  / _(_)__ _
-- | |__| \ V / -_) (__/ _ \ ' \|  _| / _` |
-- |____|_|\_/\___|\___\___/_||_|_| |_\__, |
--                                    |___/
-- Copyright (c) 2009-2023 LiveConfig GmbH.
-- ---------------------------------------------------------------------------
-- common/lua/web.lua
-- Web server management
-- ---------------------------------------------------------------------------

-- Imported functions
local LC = LC
local LCS = LCS
local type = type
local io = io
local os = os
local string = string
local require = require
local pairs = pairs
local tonumber = tonumber
local table = table

-- Module declaration
module("web")

-- Exported functions
-- getWebRoot()
-- load()
-- detect()
-- getConfig()
-- install()
-- uninstall()
-- configure()
-- addAccount()
-- deleteAccount()
-- updateAccount()
-- configureVHost()
-- updatePasswd()
-- addPHP()
-- getPHP()
-- cleanupAccesslogMap()
-- updateMapFile
-- reloadMapFile
-- reloadFPM
-- addValidation()
-- delValidation()
-- php2int()

-- Exported variables
-- HTDOCS_PATH
-- PHPDEFAULT
-- PHPCLI

-- define the path name for webspace (/var/www/<SUBSCRIPTION>/<HTDOCS_PATH/)
if HTDOCS_PATH == nil then
  HTDOCS_PATH = 'htdocs'
end

-- default PHP version (used when no explicit version is defined)
-- "nil": use systems' default PHP
PHPDEFAULT = nil

-- default PHP CLI binary (e.g. for application installer)
if PHPCLI == nil then
  PHPCLI = '/usr/bin/php'
end

-- global PHP FPM pool configuration overrides
if FPMCONFIG == nil then
  FPMCONFIG = { }
end

-- optionally allow php function "opcache_get_status" with PHP-FPM
-- (disabled by default for security reasons)
if FPM_ALLOW_OPCACHE_STATUS == nil then
  FPM_ALLOW_OPCACHE_STATUS = false
end

-- ---------------------------------------------------------------------------

-- module variables
local modlist = { }

-- array to manage additional PHP versions
local phplist = { }
local phpDefaultCode = nil

-- forward declarations
local updateLogrotate
local updateMapFile

-- set default file owner
LC.fs.setDefaultGroup("fpm-pool.conf", "root")

-- ---------------------------------------------------------------------------
-- getWebRoot()
-- Return web root directory (without trailing slash!!!)
-- ---------------------------------------------------------------------------
function getWebRoot()
  if LC.distribution.family == "Debian" then
    return "/var/www"
  elseif LC.distribution.family == "SunOS" then
    return "/var/apache2/2.2/htdocs"
  elseif LC.distribution.family == "RedHat" then
    return "/var/www"
  elseif LC.distribution.family == "SUSE" then
    return "/srv/www"
  elseif LC.distribution.family == "BSD" then
    return "/usr/local/www/apache22/data"
  elseif LC.distribution.family == "Gentoo" then
    return "/var/www"
  end

  -- return default path
  return "/var/www"
end

-- ---------------------------------------------------------------------------
-- load(modname)
--
-- Load "driver" module for a specific web server
-- ---------------------------------------------------------------------------
function load(mod)
  modlist[mod] = require(mod)
end

-- ---------------------------------------------------------------------------
-- detect()
--
-- Detect installed web server packages
-- ---------------------------------------------------------------------------
function detect()
  local data = {}
  local mod

  -- check if web server software was already detected; avoid double-detection
  -- also across multiple threads
  LC.mutex.lock("web.detect")

  if LCS.web ~= nil then
    if LCS.web.data ~= nil then
      -- web server software already detected; load values from global LCS storage
      LC.mutex.unlock("web.detect")
      return LCS.web.data
    end
  end

  -- run detection handler for all installed driver modules
  LC.log.print(LC.log.DEBUG, "running web.detect()")
  for mod in pairs(modlist) do
    local m = modlist[mod].detect()
    if m then
      data[#data+1] = m
      -- now check for configuration revision (if already managed by LiveConfig)
      if m.statusfile and LC.fs.is_file(m.statusfile) then
        local r = LC.liveconfig.readStatus(m.statusfile)
        if r and r.revision then
          -- this module is in use
          data[#data].revision = r.revision
          -- call init function (if existing)
          if modlist[mod].init then
            modlist[mod].init()
          end
        end
      end
    end
  end

  LCS.web = {
    generated = true,
    data      = data -- LCS.web.data = (local)data
  }

  -- detect PHP version
  local php = { }
  if LC.fs.is_file("/usr/bin/php-cgi") then
    php.cgi = "/usr/bin/php-cgi"
  elseif LC.fs.is_file("/usr/bin/php5-cgi") then
    php.cgi = "/usr/bin/php5-cgi"
  elseif LC.fs.is_file("/usr/bin/php-cgi7.0") then
    php.cgi = "/usr/bin/php-cgi7.0"
  elseif LC.fs.is_file("/usr/bin/php-cgi7.2") then
    php.cgi = "/usr/bin/php-cgi7.2"
  elseif LC.fs.is_file("/usr/bin/php-cgi7.3") then
    php.cgi = "/usr/bin/php-cgi7.3"
  elseif LC.fs.is_file("/usr/bin/php-cgi7.4") then
    php.cgi = "/usr/bin/php-cgi7.4"
  elseif LC.fs.is_file("/usr/bin/php-cgi8.0") then
    php.cgi = "/usr/bin/php-cgi8.0"
  elseif LC.fs.is_file("/usr/bin/php-cgi8.1") then
    php.cgi = "/usr/bin/php-cgi8.1"
  elseif LC.fs.is_file("/usr/bin/php-cgi8.2") then
    php.cgi = "/usr/bin/php-cgi8.2"
  end

  local fpmbin, fpmpool, fpmsvc, fpmsock
  if LC.distribution.family == "Debian" then
    local pkg, v = LC.distribution.hasPackage('php5-fpm', 'php7.0-fpm', 'php7.2-fpm', 'php7.3-fpm', 'php7.4-fpm', 'php8.0-fpm', 'php8.1-fpm', 'php8.2-fpm')
    if pkg == "php5-fpm" then
      fpmbin  = "/usr/sbin/php5-fpm"
      fpmpool = "/etc/php5/fpm/pool.d"
      fpmsvc  = "php5-fpm"
      fpmsock = "/var/run/php5-fpm"
    elseif pkg == "php7.0-fpm" then
      fpmbin  = "/usr/sbin/php-fpm7.0"
      fpmpool ="/etc/php/7.0/fpm/pool.d"
      fpmsvc  = "php7.0-fpm"
      fpmsock = "/var/run/php7.0-fpm"
    elseif pkg == "php7.2-fpm" then
      fpmbin  = "/usr/sbin/php-fpm7.2"
      fpmpool ="/etc/php/7.2/fpm/pool.d"
      fpmsvc  = "php7.2-fpm"
      fpmsock = "/var/run/php7.2-fpm"
    elseif pkg == "php7.3-fpm" then
      fpmbin  = "/usr/sbin/php-fpm7.3"
      fpmpool ="/etc/php/7.3/fpm/pool.d"
      fpmsvc  = "php7.3-fpm"
      fpmsock = "/var/run/php7.3-fpm"
    elseif pkg == "php7.4-fpm" then
      fpmbin  = "/usr/sbin/php-fpm7.4"
      fpmpool ="/etc/php/7.4/fpm/pool.d"
      fpmsvc  = "php7.4-fpm"
      fpmsock = "/var/run/php7.4-fpm"
    elseif pkg == "php8.0-fpm" then
      fpmbin  = "/usr/sbin/php-fpm8.0"
      fpmpool ="/etc/php/8.0/fpm/pool.d"
      fpmsvc  = "php8.0-fpm"
      fpmsock = "/var/run/php8.0-fpm"
    elseif pkg == "php8.1-fpm" then
      fpmbin  = "/usr/sbin/php-fpm8.1"
      fpmpool ="/etc/php/8.1/fpm/pool.d"
      fpmsvc  = "php8.1-fpm"
      fpmsock = "/var/run/php8.1-fpm"
    elseif pkg == "php8.2-fpm" then
      fpmbin  = "/usr/sbin/php-fpm8.2"
      fpmpool ="/etc/php/8.2/fpm/pool.d"
      fpmsvc  = "php8.2-fpm"
      fpmsock = "/var/run/php8.2-fpm"
    end
  elseif LC.distribution.family == "RedHat" then
    local pkg, v = LC.distribution.hasPackage('php-fpm')
    if pkg == "php-fpm" then
      fpmbin  = "/usr/sbin/php-fpm"
      if LC.fs.is_dir("/etc/php-fpm.d") then
        fpmpool = "/etc/php-fpm.d"
      else
        fpmpool = "/etc/php/fpm/pool.d"
      end
      fpmsvc  = "php-fpm"
      fpmsock = "/var/run/php-fpm"
    end
  elseif LC.distribution.family == "SUSE" then
    local pkg, v = LC.distribution.hasPackage('php5-fpm', 'php7-fpm')
    -- UNTESTED! openSUSE doesn't configure ANYTHING for FPM :-(
    if pkg == "php5-fpm" then
      fpmbin  = "/usr/sbin/php-fpm"
      fpmpool = "/etc/php5/fpm/pool.d"
      fpmsvc  = "php-fpm"
      fpmsock = "/var/run/php-fpm"
    elseif pkg == "php7-fpm" then
      fpmbin  = "/usr/sbin/php-fpm"
      fpmpool = "/etc/php7/fpm/pool.d"
      fpmsvc  = "php-fpm"
      fpmsock = "/var/run/php-fpm"
    end
  end

  if fpmbin ~= nil then
    php.fpm = {
      ['bin']     = fpmbin,
      ['start']   = "service " .. fpmsvc .. " start",
      ['stop']    = "service " .. fpmsvc .. " stop",
      ['reload']  = "service " .. fpmsvc .. " reload",
      ['pool']    = fpmpool,
      ['service'] = fpmsvc,
      ['sockets'] = fpmsock
    }
  end
  if php.cgi ~= nil or php.fpm ~= nil then
    addPHP(php)
  end

  if not LC.fs.is_file("/usr/bin/php") then
    LC.log.print(LC.log.WARNING, "Warning: /usr/bin/php not found (have you installed the php-cli package?)")
  end

  -- find out default PHP version
  -- PHPDEFAULT: user-configurable default PHP version code
  -- phpDefaultCode: first PHP registered (detected) with code==nil gets this code
  --                 ("php5" or "php7")
  local p
  local found = false
  for p in pairs(phplist) do
    if PHPDEFAULT ~= nil and phplist[p].code == PHPDEFAULT then
      -- found explicit default version
      found = true
      phpDefaultCode = PHPDEFAULT
      phplist[p].prio = 0
      LC.log.print(LC.log.INFO, "setting PHP default version to '", PHPDEFAULT, "'")
    elseif PHPDEFAULT == nil and phplist[p].code == phpDefaultCode then
      -- default version found
      found = true
      phplist[p].prio = 0
    else
      -- reset "priority" value for all other PHP versions
      phplist[p].prio = 1
    end
  end
  if not found and PHPDEFAULT ~= nil then
    LC.log.print(LC.log.ERR, "selected PHP default version '", PHPDEFAULT, "' not found! Using '", phpDefaultCode, "'...")
    for p in pairs(phplist) do
      if phplist[p].code == phpDefaultCode then phplist[p].prio = 0 end
    end
    -- continue anyway...
  end

  LCS.web.phpVersions = phplist
  LCS.web.phpDefaultCode = phpDefaultCode

  LC.mutex.unlock("web.detect")

  return data
end

-- ---------------------------------------------------------------------------
-- getConfig()
--
-- Return configuration table for specified web server
-- (eg. "apache" or "lighttpd")
-- ---------------------------------------------------------------------------
function getConfig(name)
  local w = detect()
  local i
  for i in pairs(w) do
    if w[i]['type'] == name then
      return w[i]
    end
  end
end

-- ---------------------------------------------------------------------------
-- install()
--
-- Start management of web server
-- ---------------------------------------------------------------------------
function install(config, opts)
  if config == nil or config.type == nil then
    LC.log.print(LC.log.ERR, "web.install(): no configuration submitted")
    return false, "web.install(): no configuration submitted"
  end
  if modlist[config.type] == nil then
    LC.log.print(LC.log.ERR, "No module for web server '", config.type, "' found")
    return false, "No module for web server '" .. config.type .. "' found"
  end

  -- adjust permissions of web root directory
  LC.fs.setperm(getWebRoot(), 751, "root", "root")

  -- create access.log map file
  updateMapFile(nil, {}, nil)

  -- run install() function from web server module
  return modlist[config.type].install(config, opts)

end

-- ---------------------------------------------------------------------------
-- uninstall()
--
-- Stop management of web server
-- ---------------------------------------------------------------------------
function uninstall(config, opts)
  if config == nil or config.type == nil then
    return false
  end
  if modlist[config.type] == nil then
    LC.log.print(LC.log.ERR, "No module for web server '", config.type, "' found")
    return false
  end

  -- run uninstall() function from web server module
  return modlist[config.type].uninstall(config, opts)
end

-- ---------------------------------------------------------------------------
-- configure()
--
-- Configure web server
-- ---------------------------------------------------------------------------
function configure(opts)

  LC.log.print(LC.log.DEBUG, "LC.web.configure()")

  local res, msg
  -- check options
  if opts == nil or opts.server == nil then
    return false, "No web server specified"
  end

  -- get configuration
  local cfg = getConfig(opts.server)
  if cfg == nil then
    return false, "Web server '" .. opts.server .. "' not found"
  end

  local statusfile = cfg["statusfile"]
  if opts and opts.prefix then
    statusfile = opts.prefix .. statusfile
  end

  -- core configuration (LiveConfig) already installed?
  if not LC.fs.is_file(statusfile) then
    -- install LiveConfig configuration
    res, msg = install(cfg, opts)
    if res == false then
      return false, msg
    end
    local r = LC.liveconfig.readStatus(statusfile)
    if r and r.revision then
      -- managed sucessfully. :)
      cfg.revision = r.revision
    end
  end

  -- update core configuration & restart service
  res, msg = modlist[cfg.type].configure(cfg, opts)
  if not res then
    return false, msg
  end

  return true
end

-- ---------------------------------------------------------------------------
local function getWebUser()
  -- get group name under which web server runs
  local wwwuser  = "www-data"
  local wwwgroup = "www-data"
  if LC.distribution.family == "Gentoo" or LC.distribution.family == "RedHat" then
    wwwuser  = "apache"
    wwwgroup = "apache"
  elseif LC.distribution.family == "SUSE" then
    wwwuser  = "wwwrun"
    wwwgroup = "www"
  elseif LC.distribution.family == "BSD" then
    wwwuser  = "www"
    wwwgroup = "www"
  end
  return wwwuser, wwwgroup
end

-- ---------------------------------------------------------------------------
-- addAccount()
--
-- Add a new webspace account
-- ---------------------------------------------------------------------------
function addAccount(name, quota, shell, password)

  -- get group name under which web server runs
  local wwwuser, wwwgroup = getWebUser()

  -- build home directory name
  local home = getWebRoot() .. "/" .. name

  LC.mutex.lock("web.configure")

  -- create group
  if not LC.sys.group_exists(name) then
    LC.log.print(LC.log.DEBUG, "Adding group '" .. name .. "'")
    if LC.users.addGroup(name) == false then
      -- group creation failed
      LC.mutex.unlock("web.configure")
      return false
    end
  end

  -- create user
  if not LC.sys.user_exists(name) then
    LC.log.print(LC.log.DEBUG, "Adding user '" .. name .. "'")
    if shell ~= "sh" and shell ~= "scponly" then shell="nologin" end
    if not LC.users.addUser(name, name, home, shell) then
      -- user creation failed
      LC.mutex.unlock("web.configure")
      return false
    end
  end

  -- set new password (if given)
  if password ~= nil then
    LC.users.chPasswd(name, password)
  end

  LC.log.print(LC.log.DEBUG, "home: ", home)

  -- create home directory (if not existing)
  if not LC.fs.is_dir(home) then
    LC.log.print(LC.log.DEBUG, "creating home: ", home)
    LC.fs.mkdir_rec(home)

    -- adjust SElinux permission type
    if LC.distribution.family == "RedHat" then
      os.execute("chcon -t user_home_dir_t \"" .. home .. "\"")
    end
  end

  -- adjust permissions of home directory
  -- the home dir itself is owned by root:root, so the customer is *not*
  -- allowed to create or modify any files here.
  LC.fs.setperm(home, 755, "root", "root")

  -- create 'htdocs' directory (if not existing), and adjust permissions
  if not LC.fs.is_dir(home .. "/" .. HTDOCS_PATH) then
    LC.fs.mkdir(home .. "/" .. HTDOCS_PATH)

    -- adjust SElinux permission type
    if LC.distribution.family == "RedHat" then
      os.execute("chcon -t user_home_t \"" .. home .. "/" .. HTDOCS_PATH .. "\"")
    end
  end
  LC.fs.setperm(home .. "/" .. HTDOCS_PATH, 750, name, wwwgroup)

  if LC.distribution.family == "RedHat" then
    -- only on RedHat/CentOS: prepare "apps" directory and adjust SELinux permissions!
    if not LC.fs.is_dir(home .. "/apps") then
      LC.fs.mkdir(home .. "/apps")
      LC.fs.setperm(home .. "/apps", 750, name, wwwgroup)
      os.execute("chcon -t user_home_t \"" .. home .. "/apps\"")
    end
  end

  -- create 'logs' directory (if not existing), and adjust permissions
  if not LC.fs.is_dir(home .. "/logs") then
    LC.fs.mkdir(home .. "/logs")
  end
  LC.fs.setperm(home .. "/logs", 750, wwwuser, name)
  -- adjust SElinux permission type
  if LC.distribution.family == "RedHat" then
    os.execute("chcon -t httpd_log_t \"" .. home .. "/logs\"")
  end

  -- create 'logs/priv' directory (if not existing), and adjust permissions
  if not LC.fs.is_dir(home .. "/logs/priv") then
    LC.fs.mkdir(home .. "/logs/priv")
  end
  LC.fs.setperm(home .. "/logs/priv", 770, wwwuser, name)
  -- adjust SElinux permission type
  if LC.distribution.family == "RedHat" then
    os.execute("chcon -t httpd_log_t \"" .. home .. "/logs/priv\"")
  end

  -- create 'priv' directory (if not existing), and adjust permissions
  if not LC.fs.is_dir(home .. "/priv") then
    LC.fs.mkdir(home .. "/priv")
  end
  LC.fs.setperm(home .. "/priv", 750, name, name)

  -- create 'tmp' directory (if not existing), and adjust permissions
  if not LC.fs.is_dir(home .. "/tmp") then
    LC.fs.mkdir(home .. "/tmp")
  end
  LC.fs.setperm(home .. "/tmp", 770, name, wwwgroup)

  -- set quota (if defined)
  if quota ~= nil and quota > 0 then
    LC.fs.setGroupQuota(name, home, quota)
  end

  LC.mutex.unlock("web.configure")

  return true, home

end

-- ---------------------------------------------------------------------------
-- deleteAccount()
--
-- Delete a webspace account
-- ---------------------------------------------------------------------------
function deleteAccount(name, opts)

  LC.mutex.lock("web.configure")

  LC.log.print(LC.log.DEBUG, "LC.web.deleteAccount()")

  -- remove all configuration files
  local i
  for i in pairs(LCS.web.data) do
    local cfg = LCS.web.data[i]
    if cfg.revision ~= nil then
      modlist[cfg.type].deleteAccount(cfg, name)
    end -- if cfg.revision...
  end

  -- convert to lowercase (user/group/path are all lowercase)
  local lname = name:lower()

  -- build home directory name
  local home = getWebRoot() .. "/" .. lname

  local user = lname
  if (opts and opts.user) then user = opts.user end
  if (opts and opts.path) then home = opts.path end

  -- remove user's home directory
  if home ~= nil and LC.fs.is_dir(home) and home ~= "/" and lname ~= "" then
    LC.log.print(LC.log.INFO, "Deleting directory '" .. home .. "' ...")

    -- remove "immutable" flags:
    os.execute("chattr -f -i " .. home .. "/conf/*")
    os.execute("chattr -f -i " .. home .. "/conf/*/php.ini")
    os.execute("chattr -f -i " .. home .. "/conf/*/php-fcgi-starter")

    os.execute("rm -rf " .. home)
    LC.log.print(LC.log.INFO, "Deleting directory '" .. home .. "': DONE")
  end

  -- delete user
  if LC.sys.user_exists(user) then
    LC.log.print(LC.log.DEBUG, "Deleting user '" .. user .. "'")
    if not LC.users.delUser(user) then
      -- deleting the user failed
      LC.mutex.unlock("web.configure")
      return false
    end
  end

  -- delete group
  if LC.sys.group_exists(user) then
    LC.log.print(LC.log.DEBUG, "Deleting group '" .. user .. "'")
    if not LC.users.delGroup(user) then
      -- deleting the group failed
      LC.mutex.unlock("web.configure")
      return false
    end
  end

  -- remove all domains from accesslog.map
  updateMapFile(nil, {}, home .. "/logs/access.log")

  -- remove from logrotate configuration
  local opts = {
    path = home,
    del  = true
  }
  updateLogrotate(opts)

  -- remove PHP-FPM configuration
  local p
  local fpm_reload_required = false
  for p in pairs(LCS.web.phpVersions) do
    local php = LCS.web.phpVersions[p]
    if php.fpm ~= nil then
      local fname = php.fpm.pool .. "/" .. name .. ".conf"
      if LC.fs.is_file(fname) then
        os.remove(fname)
        php.fpm.modified = true
        fpm_reload_required = true
      end
    end
  end

  -- remove web statistics configuration
  if LC.fs.is_file("/etc/awstats/liveconfig/awstats." .. name .. ".conf") then
    os.remove("/etc/awstats/liveconfig/awstats." .. name .. ".conf")
  end
  if LC.fs.is_file("/etc/webalizer/" .. name .. ".conf") then
    os.remove("/etc/webalizer/" .. name .. ".conf")
  end

  if fpm_reload_required then
    LC.timeout.set('web.reloadFPM', 10, 60)
  end

  LC.mutex.unlock("web.configure")

  return true

end

-- ---------------------------------------------------------------------------
-- updateAccount()
--
-- Update an existing webspace account
-- ---------------------------------------------------------------------------
function updateAccount(name, data)
  local ret = true

  -- convert to lowercase (user/group/path are all lowercase)
  local lname = name:lower()

  if data.password ~= nil then
    ret = LC.users.chPasswd(lname, data.password)
  end

  if data.shell ~= nil then
    ret = LC.users.chShell(lname, data.shell)
  end

  if data.quota ~= nil then
    -- set quota (if defined)
    local home = getWebRoot() .. "/" .. lname
    LC.fs.setGroupQuota(lname, home, data.quota)
  end

  return ret

end

-- ---------------------------------------------------------------------------
-- updateLogrotate()
--
-- Update logrotate configuration in /etc/logrotate.d/liveconfig-vhosts
-- (declared "local" in head!)
-- ---------------------------------------------------------------------------
function updateLogrotate(opts)
  if not LC.fs.is_dir("/etc/logrotate.d") then
    -- obviously no logrotate installed - abort...
    return
  end

  local fhr, fhw, msg
  -- open input file
  if LC.fs.is_file("/etc/logrotate.d/liveconfig-vhosts") then
    fhr, msg = io.open("/etc/logrotate.d/liveconfig-vhosts", "r")
    if fhr == nil then
      LC.log.print(LC.log.ERR, "Can't open '/etc/logrotate.d/liveconfig-vhosts' for reading: ", msg)
      return false, "Can't open '/etc/logrotate.d/liveconfig-vhosts' for reading: " .. msg
    end
  end

  -- open temporary output file
  fhw, msg = io.open("/etc/logrotate.d/liveconfig-vhosts.tmp", "w")
  if fhw == nil then
    LC.log.print(LC.log.ERR, "Can't open '/etc/logrotate.d/liveconfig-vhosts.tmp' for writing: ", msg)
    if fhr ~= nil then fhr:close() end
    return false, "Can't open '/etc/logrotate.d/liveconfig-vhosts.tmp' for writing: " .. msg
  end

  -- write header
  LC.liveconfig.writeHeader(fhw)
  fhw:write("# DO NOT MODIFY - ANY CHANGES WILL BE OVERWRITTEN!\n")
  fhw:write("# Last updated at: ", os.date("%Y-%m-%d %H:%M:%S %Z"), "\n")
  fhw:write("# ----------------------------------------------------------------------------\n\n")

  -- copy line by line (except the entry to be updated)
  local line
  local found = false
  local empty = true
  local search = opts.path
  local search = "^" .. opts.path .. "/logs/"
  search = string.gsub(search, "%%", "%%%%")
  search = string.gsub(search, "%.", "%%.")
  search = string.gsub(search, "%-", "%%-")
  while fhr ~= nil do
    line = fhr:read()
    if line == nil then break end

    if line == "" then
      -- dont copy multiple empty lines
      if empty then
        line = nil
      else
        empty = true
      end
    elseif string.find(line, "^#") then
      -- don't copy comment lines
      line = nil
    else
      empty = false
    end

    if line ~= nil then
      if not found then
        if string.find(line, search) then
          -- found entry!
          found = true
        else
          local v = string.match(line, "^(/.*)/logs/")
          if v and not LC.fs.is_dir(v) then
            -- found another logrotate entry; remove if customer directory doesn't exist any more:
            LC.log.print(LC.log.INFO, "Removing orphaned entry '", v, "/logs/*' from logrotate configuration")
            found = true
          end
        end
      else
        if string.find(line, "^}") then
          -- end of found section
          found = false
          line = nil
        end
      end
      if not found and line ~= nil then
        fhw:write(line, "\n")
      end
    end
  end

  if not opts.del then
    -- add new/updated entry
    -- first: configuration options for Apache/NGINX log files:
    fhw:write(opts.path .. "/logs/*.log {\n")
    if opts.logrotate ~= nil then
      if opts.logrotate.size ~= nil then
        fhw:write("    size ", opts.logrotate.size, "M\n")
      elseif opts.logrotate.interval == 1 then
        fhw:write("    daily\n")
      elseif opts.logrotate.interval == 2 then
        fhw:write("    weekly\n")
      else
        fhw:write("    monthly\n")
      end
      if opts.logrotate.compress then
        fhw:write("    compress\n    delaycompress\n")
      else
        fhw:write("    nocompress\n")
      end
      if opts.logrotate.keep ~= nil then
        fhw:write("    rotate ", opts.logrotate.keep, "\n")
      elseif opts.logrotate.maxage ~= nil then
        fhw:write("    maxage ", opts.logrotate.maxage, "\n")
        -- use same value for "rotate", as default value from /etc/logrotate.conf
        -- would otherwise keep only last 4 logs
        fhw:write("    rotate ", opts.logrotate.maxage, "\n")
      else
        fhw:write("    maxage 100\n")
      end
    else
      -- use default values
      fhw:write("    monthly\n",
                "    compress\n",
                "    delaycompress\n")
    end
    fhw:write([[
    nocreate
    missingok
    notifempty
    sharedscripts
    postrotate
      /usr/bin/killall -HUP lclogsplit || true
    endscript
}

]])

    -- second: options for "private" log files, to be rotated with user permissions!
    fhw:write(opts.path .. "/logs/priv/*.log {\n")
    fhw:write("    su root ", opts.user, "\n")
    if opts.logrotate ~= nil then
      if opts.logrotate.size ~= nil then
        fhw:write("    size ", opts.logrotate.size, "M\n")
      elseif opts.logrotate.interval == 1 then
        fhw:write("    daily\n")
      elseif opts.logrotate.interval == 2 then
        fhw:write("    weekly\n")
      else
        fhw:write("    monthly\n")
      end
      if opts.logrotate.compress then
        fhw:write("    compress\n    delaycompress\n")
      else
        fhw:write("    nocompress\n")
      end
      if opts.logrotate.keep ~= nil then
        fhw:write("    rotate ", opts.logrotate.keep, "\n")
      elseif opts.logrotate.maxage ~= nil then
        fhw:write("    maxage ", opts.logrotate.maxage, "\n")
        -- use same value for "rotate", as default value from /etc/logrotate.conf
        -- would otherwise keep only last 4 logs
        fhw:write("    rotate ", opts.logrotate.maxage, "\n")
      else
        fhw:write("    maxage 100\n")
      end
    else
      -- use default values
      fhw:write("    monthly\n",
                "    compress\n",
                "    delaycompress\n")
    end
    fhw:write([[
    nocreate
    missingok
    notifempty
}

]])

  end

  -- close (temporary) config file
  LC.liveconfig.writeFooter(fhw)
  fhw:close()

  -- close input file
  if fhr ~= nil then
    fhr:close()
  end

  -- rename map file
  LC.fs.rename("/etc/logrotate.d/liveconfig-vhosts.tmp", "/etc/logrotate.d/liveconfig-vhosts")

end

-- ---------------------------------------------------------------------------
-- getPhpIniReplaced()
--
-- get custom php.ini settings for this subcription into table,
-- replace placeholder values
-- ---------------------------------------------------------------------------
local function getPhpIniReplaced(opts, pv, code)
  local replace = {
    ["php"] = { }
  }
  if opts.phpini ~= nil then
    -- load php.ini settings into replacement table
    for i in pairs(opts.phpini) do
      -- "piTab" contains a table (array!) with all values for the given variable.
      -- if only 1 element exists, create pseudo-table around it.
      -- this way we can parse "simple" variable settings as well as arrays (eg. for "extension=")
      local piTab = opts.phpini[i]
      if (#piTab == 0) then
        piTab = { }
        piTab[1] = opts.phpini[i]
      end
      local j
      for j=1, #piTab do
        -- look for ad-hoc replacements:
        local value = piTab[j].value
        if value == nil then value = '' end
        value = string.gsub(value, "%%HOME%%", opts.path)
        value = string.gsub(value, "%%PHP%%", code)
        value = string.gsub(value, "%%USER%%", opts.user)
        local v_min = piTab[j].min  -- minimum supported PHP version (as integer)
        local v_max = piTab[j].max  -- supported up to this PHP version (not included)
        if pv == nil or ((v_min == nil or pv >= v_min) and (v_max == nil or pv < v_max)) then
          -- look for a dot in php.ini setting
          local group = "php"
          local dot = string.find(i, ".", 1, true)
          if dot ~= nil then
            -- use prefix as group name
            group = string.sub(i, 1, dot-1)
          end
          if replace[group] == nil then replace[group] = { } end
          if replace[group][i] == nil then
            -- store simple value
            replace[group][i] = value
          elseif type(replace[group][i]) == "table" then
            -- append value to array
            replace[group][i][#replace[group][i]+1] = value
          else
            -- convert existing value to table, then append to table
            local arr = { }
            arr[1] = replace[group][i]
            arr[2] = value
            replace[group][i]=arr
          end
        end -- PHP version check
      end -- "piTab" loop
    end
  end
  return replace
end

-- ---------------------------------------------------------------------------
-- updatePhpIni()
--
-- Create/update customized php.ini files per subscription:
-- ---------------------------------------------------------------------------
local function updatePhpIni(opts)
  local inputfile

  -- get group name under which web server runs
  local wwwuser, wwwgroup = getWebUser()

  -- iterate through all detected PHP versions
  local phpidx
  for phpidx in pairs(LCS.web.phpVersions) do
    local php = LCS.web.phpVersions[phpidx]
    local inputfile = php["ini"]
    local code = php["code"]
    local pv = php["versionInt"]
    local i, fhr, fhw, msg

    -- create destination path (if not existing)
    if LC.fs.is_dir(opts.path .. "/conf") == nil then
      LC.fs.mkdir(opts.path .. "/conf")
      LC.fs.setperm(opts.path .. "/conf", 750, wwwuser, opts.group)
    end
    if LC.fs.is_dir(opts.path .. "/conf/" .. code) == nil then
      LC.fs.mkdir(opts.path .. "/conf/" .. code)
      LC.fs.setperm(opts.path .. "/conf/" .. code, 555, opts.user, opts.group)
    end

    os.execute("chattr -i " .. opts.path .. "/conf/" .. code)

    -- open temporary output file
    fhw, msg = io.open(opts.path .. "/conf/" .. code .. "/php.ini.tmp", "w")
    if fhw == nil then
      LC.log.print(LC.log.ERR, "Can't open '", opts.path .. "/conf/" .. code .. "/php.ini.tmp' for writing: ", msg)
      os.execute("chattr +i " .. opts.path .. "/conf/" .. code)
      return false, "Can't open '", opts.path .. "/conf/" .. code .. "/php.ini.tmp' for writing: " .. msg
    end

    LC.liveconfig.writeHeader(fhw, ';')
    fhw:write("; DO NOT MODIFY - ANY CHANGES WILL BE OVERWRITTEN!\n")
    fhw:write("; Last updated at: ", os.date("%Y-%m-%d %H:%M:%S %Z"), "\n")
    fhw:write("; ----------------------------------------------------------------------------\n\n")

    -- replacement table with custom php.ini settings:
    local replace = getPhpIniReplaced(opts, pv, code)
    replace['FPM'] = nil  -- discard pseudo section "FPM"

    -- create opcache directory if not existing and if within %HOME%/
    if replace['opcache'] ~= nil and replace['opcache']['opcache.file_cache'] ~= nil and replace['opcache']['opcache.file_cache'] ~= '' then
      local cachedir = string.gsub(replace['opcache']['opcache.file_cache'], '^"(.*)"$', "%1")
      local search = "^" .. opts.path .. "/"
      search = string.gsub(search, "%%", "%%%%")
      search = string.gsub(search, "%.", "%%.")
      search = string.gsub(search, "%-", "%%-")
      if LC.fs.is_dir(cachedir) == nil and string.find(cachedir, search) == 1 then
        LC.fs.mkdir_rec(cachedir, 700, opts.user, opts.group)
      end
    end

    -- open default php.ini file
    fhr = nil
    if inputfile ~= nil then
      -- open input file
      fhr, msg = io.open(inputfile, "r")
      if fhr == nil then
        LC.log.print(LC.log.ERR, "Can't open '", inputfile, "' for reading: ", msg)
      end
    end

    if fhr ~= nil then
      -- parse & copy php.ini line by line
      local line
      local section
      while fhr ~= nil do
        line = fhr:read()
        if line == nil then break end

        if string.find(line, "^%s*;") or string.find(line, "^%s*$") then
          -- don't copy comment lines
          line = nil
        elseif string.find(line, "^%s*%[.*%]") then
          -- found (new) section
          -- write all not-yet written directives from previous section:
          if section ~= nil and replace[section] ~= nil then
            for i in pairs(replace[section]) do
              if type(replace[section][i]) == "table" then
                local j
                for j=1, #replace[section][i] do
                  fhw:write(i, " = ", replace[section][i][j], "\n")
                end
              else
                fhw:write(i, " = ", replace[section][i], "\n")
              end
              replace[section][i] = nil
            end
            replace[section] = nil
          end

          -- now switch to new section
          section = string.lower(string.match(line, "^%s*%[(.*)%]"))
        else
          i = string.match(line, "^%s*(.-)%s*=")
          if i ~= nil and replace[section] ~= nil and replace[section][i] ~= nil then
            if type(replace[section][i]) == "table" then
              local j
              line = ""
              for j=1, #replace[section][i] do
                if j > 1 then line = line .. "\n" end
                line = line .. i .. " = " .. replace[section][i][j]
              end
            else
              line = i .. " = " .. replace[section][i]
            end
            replace[section][i] = nil
          end
        end

        if line ~= nil then
          -- replace placeholders:
          line = string.gsub(line, "##LC_WEBROOT##", opts.path)

          fhw:write(line, "\n")
        end
      end

      -- close input file
      if fhr ~= nil then
        fhr:close()
      end
    else
      fhw:write("; No default php.ini file available for ", php.bin, "\n\n")
    end

    -- append all sections not written yet:
    for section in pairs(replace) do
      fhw:write("[", section, "]\n")
      for i in pairs(replace[section]) do
        if type(replace[section][i]) == "table" then
          local j
          for j=1, #replace[section][i] do
            fhw:write(i, "=", replace[section][i][j], "\n")
          end
        else
          fhw:write(i, "=", replace[section][i], "\n")
        end
      end
    end

    if opts.hasPHP == 1 then
      -- disable APC with suPHP
      fhw:write("; disabling APC with suPHP:\n")
      fhw:write("apc.enabled=off\n")
    end

    -- close (temporary) php.ini file
    LC.liveconfig.writeFooter(fhw, ';')
    fhw:close()

    -- adjust owner and permissions
    LC.fs.setperm(opts.path .. "/conf/" .. code .. "/php.ini.tmp", 644, "root", opts.group)
    -- adjust SElinux permission type
    if LC.distribution.family == "RedHat" then
      os.execute("chcon -t etc_t \"" .. opts.path .. "/conf/" .. code .. "/php.ini.tmp\"")
    end
    if LC.fs.is_file(opts.path .. "/conf/" .. code .. "/php.ini") then
      -- remove "immutable" flag:
      os.execute("chattr -i " .. opts.path .. "/conf/" .. code .. "/php.ini")
    end
    -- rename temporary php.ini file
    LC.fs.rename(opts.path .. "/conf/" .. code .. "/php.ini.tmp", opts.path .. "/conf/" .. code .. "/php.ini")
    -- set "immutable" flag:
    os.execute("chattr +i " .. opts.path .. "/conf/" .. code .. "/php.ini")
    os.execute("chattr +i " .. opts.path .. "/conf/" .. code)

  end

end

local function sortedPairs (t, f)
    local a = {}
    for n in pairs(t) do table.insert(a, n) end
    table.sort(a, f)
    local i = 0      -- iterator variable
    local iter = function ()   -- iterator function
        i = i + 1
        if a[i] == nil then return nil
        else return a[i], t[a[i]]
        end
    end
    return iter
end

-- ---------------------------------------------------------------------------
-- configureVHost()
--
-- Configure virtual host
-- ---------------------------------------------------------------------------
function configureVHost(opts)

  local res, msg, i

  LC.mutex.lock("web.configure")

  LC.log.print(LC.log.DEBUG, "LC.web.configureVHost()")

  -- check options
  if opts == nil or opts.name == nil then
    LC.mutex.unlock("web.configure")
    return false, "No or invalid options specified"
  end

  if opts.user == nil then opts.user = opts.name end

  -- update individual php.ini files:
  if opts.hasPHP then
    updatePhpIni(opts)
    if opts.hasPHP == 4 then
      -- update PHP-FPM pool configurations:

      -- find out which PHP versions are actually in use
      local phpversions = {}
      if opts.vhosts ~= nil then
        for i in pairs(opts.vhosts) do
          local vhost = opts.vhosts[i]
          local d
          for d in pairs(vhost.domains) do
            if vhost.domains[d].type == 1 or vhost.domains[d].type == 5 or vhost.domains[d].type == 7 then
              -- type=1: normal subdomain, type=5: application, type=7: website builder. enable PHP in either case:
              local pv = vhost.domains[d].php
              if pv == nil then pv = LC.web.getPHPDefaultCode() end
              phpversions[pv] = true
            end
          end
        end
      end

      -- remove unused pool configurations:
      local pv
      for pv in pairs(LCS.web.phpVersions) do
        local php = LCS.web.phpVersions[pv]
        -- if php.fpm ~= nil and (phpversions[php.code] == nil or (php.code == LC.web.getPHPDefaultCode() and phpversions[''] == nil)) then
        if php.fpm ~= nil and phpversions[php.code] == nil then
          local fname = php.fpm.pool .. "/" .. opts.name .. ".conf"
          if LC.fs.is_file(fname) then
            LC.log.print(LC.log.DEBUG, opts.name, ": removing FPM configuration " .. fname)
            os.remove(fname)
            php.fpm.modified = true
          end
        end
      end

      -- create corresponding pool configuration files:
      local pv
      for pv in pairs(phpversions) do
        local p
        local found = false
        for p in pairs(LCS.web.phpVersions) do
          if LCS.web.phpVersions[p].code == pv or (pv == "" and LCS.web.phpVersions[p].code == LC.web.getPHPDefaultCode()) then
            -- create FPM config file:
            found = true
            local php = LCS.web.phpVersions[p]
            LC.log.print(LC.log.DEBUG, opts.name, ": configuring PHP-FPM with ", php.code)
            if php.fpm == nil then
              LC.log.print(LC.log.ERR, "PHP version ", i, " configured as FPM for subscription ", opts.name, " but no PHP-FPM found for this version!")
              break
            end
            local fname = php.fpm.pool .. "/" .. opts.name .. ".conf"
            local fh, msg = io.open(fname .. ".tmp", "w")
            if fh == nil then
              LC.log.print(LC.log.ERR, "Can't open '", fname .. ".tmp", "' for writing: ", msg)
              break
            end
            local wwwuser, wwwgroup = getWebUser()
            LC.fs.setperm(fname .. ".tmp", 640, "root", LC.fs.getDefaultGroup("fpm-pool.conf"))
            fh:write("[", opts.name, "]\n")

            -- create pool configuration with default values
            local pcfg = {
              -- socket filename must be completely lower-case for Apache compatibility!
              ['listen']        = php.fpm.sockets .. '/' .. opts.name:lower() .. '.sock',
              ['listen.owner']  = wwwuser,
              ['listen.group']  = wwwgroup,
              ['listen.mode']   = '0600',
              ['pm']            = 'ondemand',
              ['pm.max_children']         = 8,
              ['pm.process_idle_timeout'] = 30,
              ['pm.max_requests']         = 1000,
              ['slowlog']       = opts.path .. '/logs/' .. php.code .. '-fpm.slow.log',
              ['request_slowlog_timeout'] = '10s',
              ['chdir']         = opts.path,
              ['env[TMP]']      = opts.path .. '/tmp',
              ['env[TEMP]']     = opts.path .. '/tmp',
              ['env[TMPDIR]']   = opts.path .. '/tmp'
            }

            -- get table with custom php.ini settings:
            local replace = getPhpIniReplaced(opts, php.versionInt, php.code)
            local s

            if replace['FPM'] ~= nil then
              -- merge with FPM settings from LiveConfig
              for i,s in pairs(replace['FPM']) do
                i = string.gsub(i, "^FPM%.", "")
                pcfg[i]=s
              end
            end

            -- discard FPM-specific settings from php.ini (for later processing)
            replace['FPM'] = nil

            -- merge with global pool configuration overrides:
            for i,s in pairs(FPMCONFIG) do pcfg[i]=s end

            -- merge with local pool configuration overrides:
            if LC.fs.is_file(opts.path .. "/conf/fpm.conf") then
              local t, err = LC.fs.parse_ini(opts.path .. "/conf/fpm.conf")
              if t == nil then
                -- error while parsing overrides:
                LC.log.print(LC.log.ERR, "Error while parsing '", opts.path .. "/conf/fpm.conf': ", err)
              else
                local key, value
                for key, value in pairs(t) do
                  pcfg[key] = value
                end
              end
            end

            -- (re)set some critical values:
            pcfg['prefix']  = opts.path
            pcfg['user']    = opts.user
            pcfg['group']   = opts.user

            -- now sort pool configuration and write to config file:
            local key, value
            for key, value in sortedPairs(pcfg) do
              fh:write(key, " = ", value, "\n")
            end

            fh:write("\n")

            -- insert php.ini settings for this PHP version
            -- other than in function "updatePhpIni()", we don't need to read the whole default php.ini and merge its contents,
            -- because the default php.ini is read by FPM in this case, so we only store our override values from LiveConfig:

            if replace['php'] ~= nil and replace['php']['open_basedir'] ~= nil then
              -- remove access to '/tmp' and '/var/tmp/' from open_basedir in FPM configurations for security reasons!
              replace['php']['open_basedir'] = LC.util.remove_path(replace['php']['open_basedir'], ':', '/tmp')
              replace['php']['open_basedir'] = LC.util.remove_path(replace['php']['open_basedir'], ':', '/var/tmp')
            end

            -- force-add "opcache_get_status" to disable_functions
            if not FPM_ALLOW_OPCACHE_STATUS then
              if replace['php'] == nil then replace['php'] = {} end
              if replace['php']['disable_functions'] == nil then replace['php']['disable_functions'] = '' end
              replace['php']['disable_functions'] = LC.liveconfig.listAppend(replace['php']['disable_functions'], "opcache_get_status")
              if opts.phpini['disable_functions'] == nil then
                opts.phpini['disable_functions'] = { ["o"] = false }
              end
            end

            local section
            for section in sortedPairs(replace) do
              fh:write("; [", section, "]\n")
              for i in sortedPairs(replace[section]) do
                if type(replace[section][i]) == "table" then
                  local j
                  for j=1, #replace[section][i] do
                    fh:write("php_")
                    if opts.phpini[i].o == nil or opts.phpini[i].o == false then
                      fh:write("admin_")
                    end
                    if opts.phpini[i].t and opts.phpini[i].t == 'b' then
                      fh:write("flag")
                    else
                      fh:write("value")
                    end
                    fh:write("[", i, "] = ", replace[section][i][j], "\n")
                  end
                else
                  fh:write("php_")
                  if opts.phpini[i].o == nil or opts.phpini[i].o == false then
                    fh:write("admin_")
                  end
                  if opts.phpini[i].t and opts.phpini[i].t == 'b' then
                    fh:write("flag")
                  else
                    fh:write("value")
                  end
                  fh:write("[", i, "] = ", replace[section][i], "\n")
                end
              end
            end

            fh:close()
            LC.fs.rename(fname .. ".tmp", fname)

            -- create opcache directory if not existing and if within %HOME%/
            if replace['opcache'] ~= nil and replace['opcache']['opcache.file_cache'] ~= nil and replace['opcache']['opcache.file_cache'] ~= '' then
              local search = "^" .. opts.path .. "/"
              search = string.gsub(search, "%%", "%%%%")
              search = string.gsub(search, "%.", "%%.")
              search = string.gsub(search, "%-", "%%-")
              if LC.fs.is_dir(replace['opcache']['opcache.file_cache']) == nil and string.find(replace['opcache']['opcache.file_cache'], search) == 2 then
                LC.fs.mkdir_rec(replace['opcache']['opcache.file_cache'], 700, opts.user, opts.group)
              end
            end

            -- reload corresponding PHP-FPM process
            php.fpm.modified = true

            break
          end
        end
        if not found then
          LC.log.print(LC.log.ERR, "PHP version ", i, " configured as FPM for subscription ", opts.name, " but no PHP found for this version!")
        end
      end
    else
      -- php mode is *NOT* FPM -> delete FPM configuration file(s)
      -- create corresponding pool configuration files:
      local p
      for p in pairs(LCS.web.phpVersions) do
        local php = LCS.web.phpVersions[p]
        if php.fpm ~= nil then
          local fname = php.fpm.pool .. "/" .. opts.name .. ".conf"
          if LC.fs.is_file(fname) then
            os.remove(fname)
            php.fpm.modified = true
          end
        end
      end
    end
    -- trigger config reload
    LC.timeout.set('web.reloadFPM', 10, 60)
  end

  -- iterate through all detected web servers:
  local cnames = {} -- canonical ServerName list (required for lclogsplit map entry!)
  for i in pairs(LCS.web.data) do
    local cfg = LCS.web.data[i]
    -- server: cfg.type (eg. 'apache')
    -- update vHost configuration (only if service is managed by LiveConfig!)
    if cfg.revision ~= nil then
      res, msg = modlist[cfg.type].configureVHost(cfg, opts, cnames)
      if not res then
        LC.mutex.unlock("web.configure")
        return false, msg
      end

      -- remove unused SSL certificate files
      if opts.ssl_del then
        for i in pairs(opts.ssl_del) do
          local filename = cfg.ssl_cert_dir .. "/" .. opts.ssl_del[i] .. ".crt"
          if LC.fs.is_file(filename) then
            LC.log.print(LC.log.INFO, "Deleting unused SSL certificate file: ", filename)
            os.remove(filename)
          end
          filename = cfg.ssl_chain_dir .. "/" .. opts.ssl_del[i] .. "-ca.crt"
          if LC.fs.is_file(filename) then
            LC.log.print(LC.log.INFO, "Deleting unused SSL certificate chain file: ", filename)
            os.remove(filename)
          end
          filename = cfg.ssl_key_dir .. "/" .. opts.ssl_del[i] .. ".key"
          if LC.fs.is_file(filename) then
            LC.log.print(LC.log.INFO, "Deleting unused SSL certificate key: ", filename)
            os.remove(filename)
          end
        end
      end

    end -- if cfg.revision...
  end

  -- update accesslog.map file (for lclogsplit)
  updateMapFile(opts, cnames, opts.path .. "/logs/access.log")

  -- update logrotate configuration:
  updateLogrotate(opts)

  -- check web statistics configuration
  if opts.webstats and opts.webstats.software > 0 then
    local wwwuser, wwwgroup = getWebUser()

    -- create stats path
    if not LC.fs.is_dir(opts.path .. "/stats") then
      LC.fs.mkdir(opts.path .. "/stats")
      LC.fs.setperm(opts.path .. "/stats", 750, wwwuser, opts.group)
      -- set SELinux permissions
      if LC.distribution.family == "RedHat" then
        LC.exec("chcon --type=user_home_t \"" .. opts.path .. "/stats\"")
      end
    end

    -- install placeholder page (until stats have been generated)
    if not LC.fs.is_file(opts.path .. "/stats/index.html") then
      os.execute("cp /usr/share/liveconfig/html/no-stats.html " .. opts.path .. "/stats/index.html")
      LC.fs.setperm(opts.path .. "/stats/index.html", 644, "root", "root")
    end

    -- create rotated log file dummy
    if not LC.fs.is_file(opts.path .. "/logs/access.log.1") then
      local fh, msg = io.open(opts.path .. "/logs/access.log.1", "w")
      if fh == nil then
        LC.log.print(LC.log.ERR, "Can't open '", opts.path .. "/logs/access.log.1", "' for writing: ", msg)
      else
        fh:write("0.0.0.1 - - [01/Jan/2011:00:00:00 +0100] \"OPTIONS * HTTP/1.0\" 200 - \"-\" \"Dummy entry\" 0 0\n")
        fh:close()
        LC.fs.setperm(opts.path .. "/logs/access.log.1", 644, "root", "root")
      end
    end

    -- create/update password file
    if opts.webstats.user and opts.webstats.password then
      local pfh, msg = io.open(opts.path .. "/stats/.htpasswd", "w")
      if pfh == nil then
        LC.log.print(LC.log.ERR, "Can't open '", opts.path .. "/stats/.htpasswd", "' for writing: ", msg)
      else
        pfh:write(opts.webstats.user .. ":" .. LC.crypt.crypt(opts.webstats.password) .. "\n")
        pfh:close()
        LC.fs.setperm(opts.path .. "/stats/.htpasswd", 440, wwwuser, opts.group)
      end
      -- set SELinux permissions
      if LC.distribution.family == "RedHat" then
        LC.exec("chcon --type=httpd_config_t \"" .. opts.path .. "/stats/.htpasswd\"")
      end
    end

  end

  -- ###TODO### move statistics configuration to own Lua file
  if LC.fs.is_file("/usr/bin/webalizer") then
    -- check if configuration has to be generated/deleted
    if opts.webstats and opts.webstats.software == 1 then
      -- most distributions don't install webalizer ready-to-use
      local filename
      if LC.distribution.family == "Debian" then
        filename = "/etc/webalizer/" .. opts.name .. ".conf"
      else
        if not LC.fs.is_dir("/etc/webalizer") then
          -- create directory
          LC.fs.mkdir("/etc/liveconfig/webalizer")
          LC.fs.setperm("/etc/liveconfig/webalizer", 755, "root", "root")
        end
        if not LC.fs.is_file("/etc/cron.daily/liveconfig.webalizer") then
          -- create webalizer cron job
          local fh, msg = io.open("/etc/cron.daily/liveconfig.webalizer", "w")
          if fh == nil then
            LC.log.print(LC.log.ERR, "Can't create '/etc/cron.daily/liveconfig.webalizer': ", msg)
          else
            fh:write("#!/bin/sh\n")
            LC.liveconfig.writeHeader(fh)
            fh:write("# Cron job for daily webalizer statistics\n")
            fh:write("# ----------------------------------------------------------------------------\n\n")
            fh:write([[
WEBALIZER=/usr/bin/webalizer
CONFDIR=/etc/liveconfig/webalizer

for i in ${CONFDIR}/*.conf; do

  # first try rotated log file:
  LOGFILE=`awk '$1 ~ /^LogFile$/ {print $2}' $i`;
  if [ -r "${LOGFILE}" ]; then
    ${WEBALIZER} -c ${i} -Q || continue;
  fi;

  # then check for non-rotated log file:
  NLOGFILE=`awk '$1 ~ /^LogFile$/ {gsub(/\.[0-9]+(\.gz)?/,""); print $2}' $i`;
  if [ "${LOGFILE}" != "${NLOGFILE}" -a -r "${NLOGFILE}" ]; then
    ${WEBALIZER} -c ${i} -Q ${NLOGFILE} || continue;
  fi;

done;

# done.
]])
            fh:close()
            LC.fs.setperm("/etc/cron.daily/liveconfig.webalizer", 755, "root", "root")
          end
        end
        filename = "/etc/liveconfig/webalizer/" .. opts.name .. ".conf"
      end

      -- create config file
      local tempfile = filename .. ".tmp"
      local fh, msg = io.open(tempfile, "w")
      if fh == nil then
        LC.log.print(LC.log.ERR, "Can't open '", tempfile, "' for writing: ", msg)
        LC.mutex.unlock("web.configure")
        return false, "Can't open '" .. tempfile .. "' for writing: " .. msg
      end
      LC.liveconfig.writeHeader(fh)
      fh:write("# Created at: ", os.date("%Y-%m-%d %H:%M:%S %Z"), "\n")
      fh:write("# ----------------------------------------------------------------------------\n\n")

      -- variable part:
      -- Webalizer is first run against a rotated log file, and then the current log file (to catch log rotation)
      fh:write("LogFile ", opts.path, "/logs/access.log.1\n")
      fh:write("OutputDir ", opts.path, "/stats\n")
      fh:write("HostName ", opts.name, "\n")

      -- static part:
      fh:write([[
ReportTitle Usage statistics for
Incremental     yes
IncrementalName webalizer.current
HistoryName     webalizer.hist
DNSCache        dns_cache.db
PageType        htm*
PageType        php*
PageType        cgi
PageType        pl
PageType        shtml

HTMLHead  <style type="text/css">
HTMLHead    body, p, th, td, font { font-family:Arial, Helvetica, Sans-serif;
HTMLHead      font-size:12px; }
HTMLHead    body { background:url(images/bg.png) top left repeat-x #F0F0F0; }
HTMLHead    hr { border:0px; border-top:1px solid #999999;
HTMLHead      border-bottom:solid 1px #F0F0F0; }
HTMLHead    center > hr { visibility:hidden; }
HTMLHead    h2 { color:#222222; font-size:18px; font-weight:bold;
HTMLHead      text-shadow: 0px 1px 2px #999999; }
HTMLHead    tr > th[BGCOLOR="#C0C0C0"] { text-shadow: 1px 1px 0px #999999;
HTMLHead      font-size:13px; }
HTMLHead  </style>
HTMLBody <BODY BGCOLOR="#F0F0F0" TEXT="#000000" LINK="#0000FF" VLINK="#FF0000">
HTMLTail <img width="99" height="17" alt="LiveConfig" src="images/logo.png">

HideURL         *.gif
HideURL         *.GIF
HideURL         *.jpg
HideURL         *.JPG
HideURL         *.png
HideURL         *.PNG
HideURL         *.swf
HideURL         *.js
HideURL         *.css
HideURL         robots.txt
HideURL         favicon.ico

IgnoreSite      0.0.0.1

SearchEngine    yahoo.com       p=
SearchEngine    altavista.com   q=
SearchEngine    eureka.com      q=
SearchEngine    lycos.com       query=
SearchEngine    hotbot.com      MT=
SearchEngine    msn.com         MT=
SearchEngine    infoseek.com    qt=
SearchEngine    webcrawler      searchText=
SearchEngine    excite          search=
SearchEngine    netscape.com    search=
SearchEngine    mamma.com       query=
SearchEngine    alltheweb.com   query=
SearchEngine    northernlight.com  qr=
SearchEngine    sensis.com.au   find=
SearchEngine    www.google.     q=

]])

      LC.liveconfig.writeFooter(fh)
      fh:close()

      -- rename config file (atomic)
      LC.fs.rename(tempfile, filename)

    else
      -- delete webalizer config file (if existing)
      os.remove("/etc/webalizer/" .. opts.name .. ".conf")
    end
  end

  if LC.fs.is_dir("/etc/awstats") then
    -- check if configuration has to be generated/deleted
    if opts.webstats and opts.webstats.software == 2 then
      -- check if /etc/awstats/liveconfig exists
      if not LC.fs.is_dir("/etc/awstats/liveconfig") then
        LC.fs.mkdir("/etc/awstats/liveconfig")
      end

      -- check if default configuration already exists
      if not LC.fs.is_file("/etc/awstats/liveconfig/awstats.conf.liveconfig") then
        local filename = "/etc/awstats/liveconfig/awstats.conf.liveconfig"
        local fh, msg = io.open(filename, "w")
        if fh == nil then
          LC.log.print(LC.log.ERR, "Can't open '", filename, "' for writing: ", msg)
          LC.mutex.unlock("web.configure")
          return false, "Can't open '" .. filename .. "' for writing: " .. msg
        end
        LC.liveconfig.writeHeader(fh)
        fh:write("# Created at: ", os.date("%Y-%m-%d %H:%M:%S %Z"), "\n")
        fh:write("# ----------------------------------------------------------------------------\n\n")

        fh:write([[
DirIcons="../icon"
StyleSheet="../awstats/awstats.css"
HTMLHeadSection="<div id="dLCNav"><iframe src="../awstats-nav.html" scrolling="no"></iframe></div>"
HTMLEndSection="<div class="dLiveConfig"><a href="http://www.liveconfig.com" title="LiveConfig Server Control Panel"><img src="../awstats/liveconfig.png" width="137" height="32" alt="powered by LiveConfig"/></a></div>"
SkipHosts="0.0.0.1"

LogType=W
LogFormat=1
LogSeparator=" "
HostAliases="localhost 127.0.0.1"
DNSLookup=0
DirCgi="cgi-bin"
AllowToUpdateStatsFromBrowser=0
AllowFullYearView=2

ShowMenu=1
ShowSummary=UVPHB
ShowMonthStats=UVPHB
ShowDaysOfMonthStats=VPHB
ShowDaysOfWeekStats=PHB
ShowHoursStats=PHB
ShowDomainsStats=PHB
ShowHostsStats=PHBL
ShowAuthenticatedUsers=0
ShowRobotsStats=HBL
ShowWormsStats=0
ShowEMailSenders=0
ShowEMailReceivers=0
ShowSessionsStats=1
ShowPagesStats=PBEX
ShowFileTypesStats=HB
ShowFileSizesStats=0
ShowDownloadsStats=HB
ShowOSStats=1
ShowBrowsersStats=1
ShowScreenSizeStats=0
ShowOriginStats=PH
ShowKeyphrasesStats=1
ShowKeywordsStats=1
ShowMiscStats=a
ShowHTTPErrorsStats=1
ShowSMTPErrorsStats=0
ShowClusterStats=0
ShowFlagLinks=""
ShowLinksOnUrl=1

LoadPlugin="hashfiles"

]])

        LC.liveconfig.writeFooter(fh)
        fh:close()
      end

      -- create config file
      local filename = "/etc/awstats/liveconfig/awstats." .. opts.name .. ".conf"
      local tempfile = filename .. ".tmp"
      local fh, msg = io.open(tempfile, "w")
      if fh == nil then
        LC.log.print(LC.log.ERR, "Can't open '", tempfile, "' for writing: ", msg)
        LC.mutex.unlock("web.configure")
        return false, "Can't open '" .. tempfile .. "' for writing: " .. msg
      end
      LC.liveconfig.writeHeader(fh)
      fh:write("# Created at: ", os.date("%Y-%m-%d %H:%M:%S %Z"), "\n")
      fh:write("# ----------------------------------------------------------------------------\n\n")

      fh:write("LogFile=\"", opts.path, "/logs/access.log\"\n")
      fh:write("SiteDomain=\"", opts.name, "\"\n")
      fh:write("DirData=\"", opts.path, "/stats\"\n")
      if opts.language ~= nil then
        fh:write("Lang=\"", opts.language, "\"\n")
      else
        fh:write("Lang=\"de\"\n")
      end

      fh:write("\nInclude \"/etc/awstats/liveconfig/awstats.conf.liveconfig\"\n\n")

      LC.liveconfig.writeFooter(fh)
      fh:close()

      -- rename config file (atomic)
      LC.fs.rename(tempfile, filename)

    else
      -- delete AWStats config file (if existing)
      os.remove("/etc/awstats/liveconfig/awstats." .. opts.name .. ".conf")
    end
  end

  LC.mutex.unlock("web.configure")
  return true

end

-- ---------------------------------------------------------------------------
-- updatePasswd()
--
-- Create/update
-- ---------------------------------------------------------------------------
function updatePasswd(opts)

  local fh, res, msg, user

  -- get group name under which web server runs
  local wwwuser, wwwgroup = getWebUser()

  LC.mutex.lock("web.updatePasswd")

  LC.log.print(LC.log.DEBUG, "LC.web.updatePasswd()")

  -- check options
  if opts == nil or opts.name == nil then
    LC.mutex.unlock("web.updatePasswd")
    return false, "No or invalid options specified"
  end

  if opts.users == nil then
    -- no users/passwords defined, delete .htpasswd if existing
    if LC.fs.is_file(opts.path .. "/conf/.htpasswd") then
      os.remove(opts.path .. "/conf/.htpasswd")
    end
    LC.mutex.unlock("web.updatePasswd")
    return true
  end

  -- create destination path (if not existing)
  if opts.path == nil then
    LC.mutex.unlock("web.updatePasswd")
    return false, "No webspace path defined"
  end
  if not LC.fs.is_dir(opts.path .. "/conf") then
    LC.fs.mkdir(opts.path .. "/conf")
    LC.fs.setperm(opts.path .. "/conf", 750, wwwuser, opts.group)
  end

  -- open temporary output file
  fh, msg = io.open(opts.path .. "/conf/.htpasswd.tmp", "w")
  if fh == nil then
    LC.mutex.unlock("web.updatePasswd")
    LC.log.print(LC.log.ERR, "Can't open '", opts.path .. "/conf/.htpasswd.tmp' for writing: ", msg)
    return false, "Can't open '", opts.path .. "/conf/.htpasswd.tmp' for writing: " .. msg
  end

  for user in pairs(opts.users) do
    fh:write(user, ":", opts.users[user], "\n")
  end

  fh:close()

  -- adjust owner and permissions
  LC.fs.setperm(opts.path .. "/conf/.htpasswd.tmp", 440, wwwuser, opts.group)

  -- rename temporary .htpasswd file
  LC.fs.rename(opts.path .. "/conf/.htpasswd.tmp", opts.path .. "/conf/.htpasswd")

  LC.mutex.unlock("web.updatePasswd")
  return true

end

-- ---------------------------------------------------------------------------
-- addPHP()
--
-- Register PHP version
-- code: string or table
-- ---------------------------------------------------------------------------
function addPHP(code, bin, version)
  local id
  local eoldate

  if type(code) == 'table' then
    -- Lua table with all PHP data (LiveConfig v2.6.0+)
    id = code.id
    if code.cgi == nil then
      LC.log.print(LC.log.ERR, "addPHP: missing parameter 'cgi' (FastCGI binary required!) ")
      return false
    end
    -- check CGI binary:
    bin = code.cgi
    if code.version ~= nil then version = code.version end
    eoldate = code['eol-date']
  else
    id = code
  end

  -- check that this binary exists:
  if not LC.fs.is_file(bin) then
    return false
  end

  -- priority: 0=default version, 1=additional version
  local prio = 1

  local handle = io.popen(bin .. " -v", "r")
  if handle == nil then
    -- can't execute PHP binary
    LC.log.print(LC.log.WARNING, "addPHP: can't execute '" .. bin .. "'")
    return false
  end

  local ret = handle:read("*a")
  handle:close()

  if ret == nil then
    LC.log.print(LC.log.WARNING, "addPHP: got no output from '" .. bin .. "'")
    return false
  end

  local v_major, v_minor, v_patch = string.match(ret, "^PHP (%d+)%.(%d+)%.(%d+)")
  if v_major == nil or v_minor == nil or v_patch == nil then
    LC.log.print(LC.log.ERR, "PHP version detection failed for " .. bin .. " (try to run with -v and check for errors!)")
    return false
  end
  local v_int = tonumber(v_major) * 10000 + tonumber(v_minor) * 100 + tonumber(v_patch)
  local api = 0
  if string.match(ret, "%(cgi%-fcgi%)") then
    api = 1
  elseif string.match(ret, "%(fpm%-fcgi%)") then
    api = 2
  elseif string.match(ret, "%(cli%)") then
    LC.log.print(LC.log.ERR, "PHP command line interface (CLI) can't be used with addPHP() - please use CGI/FCGI/FPM SAPI (" .. bin .. ")")
    return false
  else
    LC.log.print(LC.log.ERR, "Unknown or unsupported PHP SAPI for " .. bin)
    return false
  end

  -- if detected binary is for CGI/FCGI API, check if we also have a FPM binary:
  local fpm
  while api ~= 2 and type(code) == 'table' and type(code.fpm) == 'table' do
    LC.log.print(LC.log.DEBUG, "Checking FPM binary...")
    if code.fpm.bin == nil or code.fpm.start == nil or code.fpm.stop == nil or code.fpm.reload == nil
      or code.fpm.pool == nil or code.fpm.sockets == nil then
      LC.log.print(LC.log.WARNING, "addPHP: incomplete FPM configuration for '" .. code.fpm.bin .. "' - skipping...")
      break
    end
    handle = io.popen(code.fpm.bin .. " -v", "r")
    if handle == nil then
      -- can't execute PHP FPM binary
      LC.log.print(LC.log.WARNING, "addPHP: can't execute FPM binary '" .. code.fpm.bin .. "'")
      break
    end

    -- read output
    ret = handle:read("*a")
    handle:close()

    if ret == nil then
      LC.log.print(LC.log.WARNING, "addPHP: got no output from FPM binary '" .. code.fpm.bin .. "'")
      break
    end

    local fpm_major, fpm_minor, fpm_patch = string.match(ret, "^PHP (%d+)%.(%d+)%.(%d+)")
    if fpm_major == nil or fpm_minor == nil or fpm_patch == nil then
      LC.log.print(LC.log.ERR, "PHP version detection failed for " .. code.fpm.bin .. " (try to run with -v and check for errors!)")
      break
    end
    local fpm_int = tonumber(fpm_major) * 10000 + tonumber(fpm_minor) * 100 + tonumber(fpm_patch)
    if not string.match(ret, "%(fpm%-fcgi%)") then
      LC.log.print(LC.log.ERR, "addPHP: no FPM support detected: " .. code.fpm.bin)
      break
    end

    fpm = code.fpm

    -- bitmask: 1=CGI, 2=FPM
    api = LC.bits.bor(api, 2)

    break
  end

  local inifile = nil
  if v_int >= 50204 then
    -- get location of default php.ini file
    handle = io.popen("echo '<?php echo php_ini_loaded_file(); ?>' | " .. bin .. " -q", "r")
    ret = handle:read("*a")
    handle:close()
    if LC.fs.is_file(ret) then
      inifile = ret
    else
      LC.log.print(LC.log.ERR, "Can't get default php.ini location from " .. bin)
    end
  end
  if id == nil then
    -- default PHP version
    id = "php" .. v_major
    if phpDefaultCode == nil then phpDefaultCode = id end
  end

  -- check if this version (code!) is already configured:
  local p
  for p in pairs(phplist) do
    if phplist[p].code == id then
      LC.log.print(LC.log.WARNING, "PHP version '" .. id .. "' already registered")
      return
    end
    if phplist[p].bin == bin then
      LC.log.print(LC.log.WARNING, "PHP binary '" .. bin .. "' already registered with id '" .. phplist[p].code .. "' - ignoring previous registration...")
      -- for backward compatibility with "old" PHP registration from custom.lua:
      phplist[p] = nil
    end
  end

  -- if no explicit version is set, use detected version
  if version == nil then
    version = v_major .. "." .. v_minor .. "." .. v_patch
  end

  LC.log.print(LC.log.DEBUG, "adding PHP '" .. id .. "'")
  phplist[#phplist+1] = {
    ["code"]        = id,
    ["bin"]         = bin,
    ["fpm"]         = fpm,
    ["version"]     = version,
    ["versionInt"]  = v_int,
    ["ini"]         = inifile,
    ["prio"]        = prio,
    ["api"]         = api,
    ["eol-date"]    = eoldate
  }

end

-- ---------------------------------------------------------------------------
-- getPHP()
--
-- Get list with available PHP versions
-- ---------------------------------------------------------------------------
function getPHP()
  return LCS.web.phpVersions
end

-- ---------------------------------------------------------------------------
-- getPHPDefaultCode()
--
-- Get default PHP version code (eg. "php5" or "php7")
-- ---------------------------------------------------------------------------
function getPHPDefaultCode()
  return LCS.web.phpDefaultCode
end

-- ---------------------------------------------------------------------------
-- cleanupAccesslogMap()
--
-- Remove duplicate entries from accesslog.map file
-- ---------------------------------------------------------------------------
function cleanupAccesslogMap(filename)
  local fh, errmsg = io.open(filename, "r")
  if fh == nil then
    LC.log.print(LC.log.INFO, "Can't open '" .. filename .. "' for reading: ", errmsg)
    return false
  end

  local oldvalues = { }
  local line

  while fh ~= nil do
    line = fh:read()
    if line == nil then break end
    if not line:match("^#") then
      local k,v = string.match(line, "^([^ ]*) (.*)")
      if k ~= nil and v ~= nil then
        oldvalues[k] = v
      end
    end
  end
  fh:close()

  fh, errmsg = io.open(filename .. ".tmp", "w")
  if fh == nil then
    LC.log.print(LC.log.INFO, "Can't open '" .. filename .. ".tmp' for writing: ", errmsg)
    return false
  end

  LC.liveconfig.writeHeader(fh)

  local key
  for key in pairs(oldvalues) do
    fh:write(key, " ", oldvalues[key], "\n")
  end

  LC.liveconfig.writeFooter(fh)

  fh:close()

  -- rename map file
  LC.fs.rename(filename .. "tmp", filename)

  return true

end

-- ---------------------------------------------------------------------------
-- updateMapFile(opts, domains, logname)
--
-- Update, add or delete an entry in accesslog.map file (for lclogsplit)
-- ---------------------------------------------------------------------------
function updateMapFile(opts, domains, logname)
  local i
  local domainlist = { }

  local fhr, fhw, msg

  -- create log map file if not yet existing:
  local mapName = LC.liveconfig.localstatedir .. "/lib/" .. LC.liveconfig.project_name .. "/accesslog.map"
  if not LC.fs.is_file(mapName) then
    fh, msg = io.open(mapName, "w")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't create '", mapName, "': ", msg)
      return false, "Can't create '" .. mapName .. "': " .. msg
    end
    -- write header
    LC.liveconfig.writeHeader(fh)
    if LC.distribution.family == "RedHat" then
      fh:write("default /var/log/httpd/access_log\n")
    else
      fh:write("default /var/log/apache2/access.log\n")
    end
    LC.liveconfig.writeFooter(fh)
    fh:close()
  end

  if logname == nil then return end

  -- copy every entry from "domains" table as key into "domainlist" table
  -- (for faster access to this table when checking matched domains)
  for i=1, #domains do
    domainlist[domains[i]] = true
  end

  local search = " " .. logname
  search = string.gsub(search, "%%", "%%%%")
  search = string.gsub(search, "%.", "%%.")
  search = string.gsub(search, "%-", "%%-")

  fhr, msg = io.open(mapName, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", mapName, "' for reading: ", msg)
    return false, "Can't open '" .. mapName .. "' for reading: " .. msg
  end

  fhw, msg = io.open(mapName .. ".tmp", "w")
  if fhw == nil then
    LC.log.print(LC.log.ERR, "Can't open '", mapName, ".tmp' for writing: ", msg)
    fhr:close()
    return false, "Can't open '" .. mapName .. ".tmp' for writing: " .. msg
  end

  local line
  while true do
    line = fhr:read()
    if line == nil then break end
    if string.find(line, search) then
      -- remove this entry from map file
      line = nil
    end
    if line ~= nil then
      fhw:write(line, "\n")
    end
  end

  for i in pairs(domainlist) do
    -- entry not found, so add new one:
    local f4 = opts.filterv4
    if f4 == nil then f4 = 0 end
    local f6 = opts.filterv6
    if f6 == nil then f6 = 0 end
    fhw:write(i, " ", logname, " ", f4, "/", f6, "\n")
  end

  fhw:close()
  fhr:close()

  -- rename map file
  LC.fs.rename(mapName .. ".tmp", mapName)

  -- send (delayed) HUP signal to lclogsplit to reload map
  if LC.fs.is_file(LC.liveconfig.runstatedir .. "/lclogsplit.pid") then
    LC.timeout.set('web.reloadMapFile', 10, 60)
  end

end

-- ---------------------------------------------------------------------------
-- reloadMapFile()
--
-- Send HUP signal to standalone lclogsplit process to reload map file
-- ---------------------------------------------------------------------------
function reloadMapFile()
  local st, msg = LC.sys.kill_pid(LC.liveconfig.runstatedir .. '/lclogsplit.pid', 1)
  if not st then
    LC.log.print(LC.log.ERR, "Error while sending SIGHUP to lclogsplit: ", msg)
  end
end

-- ---------------------------------------------------------------------------
-- reloadFPM()
--
-- Reload FPM configuration
-- ---------------------------------------------------------------------------
function reloadFPM()
  local p
  for p in pairs(LCS.web.phpVersions) do
    local php = LCS.web.phpVersions[p]
    if php.fpm ~= nil and php.fpm.modified ~= nil and php.fpm.modified then
      -- FPM for this version requires reload
      -- if FPM package is provided by distribution, eventually some settings must be tuned...

      -- get number of pool configuration files
      local files, errmsg, errno = LC.fs.glob(php.fpm.pool .. "/*.conf")
      if files == nil then
        LC.log.print(LC.log.ERR, "glob(" .. php.fpm.pool .. "/*.conf) failed: " .. errmsg .. " (" .. errno .. ")")
      elseif #files == 0 then
        -- no more pools configured -> stop daemon
        LC.log.print(LC.log.DEBUG, "Stopping FPM daemon for ", php.code)
        os.execute(php.fpm.stop)
      else -- #files > 0
        files, errmsg, errno = LC.fs.glob(php.fpm.sockets .. "/*.sock")
        if not LC.fs.is_dir(php.fpm.sockets) or (files ~= nil and #files == 0) then
          -- if this FPM package is provided by the distribution, some settings need to be tweaked...
          if php.fpm.service ~= nil and not LC.fs.is_dir(php.fpm.sockets) and LC.fs.is_dir("/etc/systemd/system") and not LC.fs.is_dir("/etc/systemd/system/" .. php.fpm.service .. ".service.d") then
            -- create systemd drop-in:
            LC.fs.mkdir("/etc/systemd/system/" .. php.fpm.service .. ".service.d")
            local svcfile = "/etc/systemd/system/" .. php.fpm.service .. ".service.d/liveconfig.conf"
            LC.fs.setperm(svcfile, 644, "root", "root")
            local fhw, msg = io.open(svcfile, "w")
            if fhw == nil then
              LC.log.print(LC.log.ERR, "Can't open '", svcfile, "' for writing: ", msg)
            else
              fhw:write("[Unit]\n")
              fhw:write("ConditionPathExistsGlob=" .. php.fpm.pool .. "/*.conf\n")
              fhw:write("\n")
              fhw:write("[Service]\n")
              fhw:write("RuntimeDirectory=" .. php.fpm.service .. "\n")
              fhw:close()
              os.execute("systemctl daemon-reload")
            end
          end
          LC.log.print(LC.log.DEBUG, "Stopping FPM daemon for ", php.code)
          os.execute(php.fpm.stop)
          LC.log.print(LC.log.DEBUG, "Starting FPM daemon for ", php.code)
          os.execute(php.fpm.start)
        else
          LC.log.print(LC.log.DEBUG, "Reloading FPM daemon for ", php.code)
          os.execute(php.fpm.reload)
        end
      end

      -- reeset reload flag:
      php.fpm.modified = false

    end
  end

end

-- ---------------------------------------------------------------------------
-- addValidation()
--
-- Add token for domain validation
-- ---------------------------------------------------------------------------
function addValidation(domainname, pathname, filename, content)
  local wwwuser, wwwgroup = getWebUser()

  if pathname:match("%.%./") or pathname:match("%.%.$") or filename:match("/") then
    return false, "Invalid characters in path or file name"
  end

  local basedir = getWebRoot() .. "/.well-known"
  if not LC.fs.is_dir(basedir) then
    LC.fs.mkdir(basedir)
    LC.fs.setperm(basedir, 750, wwwuser, wwwgroup)
  end

  basedir = basedir .. "/htdocs"
  if not LC.fs.is_dir(basedir) then
    LC.fs.mkdir(basedir)
    LC.fs.setperm(basedir, 750, wwwuser, wwwgroup)
  end

  -- strip "/.well-known" from path name
  if pathname:match("^/%.well%-known/") then pathname = pathname:sub(13) end

  basedir = basedir .. pathname
  if not LC.fs.is_dir(basedir) then
    LC.fs.mkdir(basedir)
    LC.fs.setperm(basedir, 750, wwwuser, wwwgroup)
  end

  local tokenfile = basedir .. '/' .. domainname:lower() .. "." .. filename
  local tk, msg = io.open(tokenfile .. ".tmp", "w")
  if tk == nil then
    LC.log.print(LC.log.ERR, "Can't open '", tokenfile, ".tmp' for writing: ", msg)
  else
    tk:write(content)
    tk:flush()
    tk:close()
    LC.fs.setperm(tokenfile .. ".tmp", "0640", wwwuser, wwwgroup)
    -- rename file (for atomic creation)
    LC.fs.rename(tokenfile .. ".tmp", tokenfile)
    LC.log.print(LC.log.INFO, "Created SSL/TLS domain-validation file '", tokenfile, "'")
  end

  return true

end

-- ---------------------------------------------------------------------------
-- delValidation()
--
-- Remove token for domain validation
-- ---------------------------------------------------------------------------
function delValidation(domainname, pathname, filename)
  local wwwuser, wwwgroup = getWebUser()

  if pathname:match("%.%./") or pathname:match("%.%.$") or filename:match("/") then
    return false, "Invalid characters in path or file name"
  end

  local basedir = getWebRoot() .. "/.well-known/htdocs"

  -- strip "/.well-known" from path name
  if pathname:match("^/%.well%-known/") then pathname = pathname:sub(13) end

  local tokenfile = basedir .. pathname .. '/' .. domainname:lower() .. "." .. filename
  if LC.fs.is_file(tokenfile) then
    LC.log.print(LC.log.DEBUG, "Removing SSL/TLS domain-validation file '", tokenfile, "'")
    os.remove(tokenfile)
  end

  return true

end

-- ---------------------------------------------------------------------------
-- php2int()
--
-- Convert php.ini byte/size value into number
-- ---------------------------------------------------------------------------
function php2int(str)
  if str == nil then return end
  local val, unit = string.match(str, "(%d+)([KMG]?)")
  val = tonumber(val)
  if unit == "K" then
    val = val * 1024
  elseif unit == "M" then
    val = val * 1048576
  elseif unit == "G" then
    val = val * 1073741824
  end
  return val
end

-- <EOF>----------------------------------------------------------------------
